=== modified file 'drizzled/plugin/event_observer.cc'
--- drizzled/plugin/event_observer.cc	2010-10-18 18:24:29 +0000
+++ drizzled/plugin/event_observer.cc	2010-10-20 20:41:00 +0000
@@ -99,7 +99,7 @@
     {
       for_each(event_observer_lists.begin(),
                event_observer_lists.end(),
-               DeletePtr());
+               SafeDeletePtr());
 			event_observer_lists.clear();
     }
 
@@ -405,6 +405,9 @@
   {
     EventObserverList::ObserverMap *eventObservers;
 
+    if (observerList == NULL)
+      return false; // Nobody was interested in the event. :(
+
     eventObservers = observerList->getObservers(event);
 
     if (eventObservers == NULL)
@@ -452,7 +455,7 @@
 
   /*==========================================================*/
   /* Static meathods called by drizzle to notify interested plugins 
-   * of a schema event,
+   * of a schema event.
  */
   bool EventObserver::beforeDropTable(Session &session, const drizzled::TableIdentifier &table)
   {
@@ -492,11 +495,13 @@
 
   /*==========================================================*/
   /* Static meathods called by drizzle to notify interested plugins 
-   * of a table event,
+   * of a table event.
+   *
+   * A quick test is done first to see if there are any interested observers.
  */
   bool EventObserver::beforeInsertRecord(Table &table, unsigned char *buf)
   {
-    if (all_event_plugins.empty())
+    if (all_event_plugins.empty() || !TableEventData::hasEvents(table))
       return false;
 
     BeforeInsertRecordEventData eventData(*(table.in_use), table, buf);
@@ -505,7 +510,7 @@
 
   bool EventObserver::afterInsertRecord(Table &table, const unsigned char *buf, int err)
   {
-    if (all_event_plugins.empty())
+    if (all_event_plugins.empty() || !TableEventData::hasEvents(table))
       return false;
 
     AfterInsertRecordEventData eventData(*(table.in_use), table, buf, err);
@@ -514,7 +519,7 @@
 
   bool EventObserver::beforeDeleteRecord(Table &table, const unsigned char *buf)
   {
-    if (all_event_plugins.empty())
+    if (all_event_plugins.empty() || !TableEventData::hasEvents(table))
       return false;
 
     BeforeDeleteRecordEventData eventData(*(table.in_use), table, buf);
@@ -523,7 +528,7 @@
 
   bool EventObserver::afterDeleteRecord(Table &table, const unsigned char *buf, int err)
   {
-    if (all_event_plugins.empty())
+    if (all_event_plugins.empty() || !TableEventData::hasEvents(table))
       return false;
 
     AfterDeleteRecordEventData eventData(*(table.in_use), table, buf, err);
@@ -532,7 +537,7 @@
 
   bool EventObserver::beforeUpdateRecord(Table &table, const unsigned char *old_data, unsigned char *new_data)
   {
-    if (all_event_plugins.empty())
+    if (all_event_plugins.empty() || !TableEventData::hasEvents(table))
       return false;
 
     BeforeUpdateRecordEventData eventData(*(table.in_use), table, old_data, new_data);
@@ -541,7 +546,7 @@
 
   bool EventObserver::afterUpdateRecord(Table &table, const unsigned char *old_data, unsigned char *new_data, int err)
   {
-    if (all_event_plugins.empty())
+    if (all_event_plugins.empty() || !TableEventData::hasEvents(table))
       return false;
 
     AfterUpdateRecordEventData eventData(*(table.in_use), table, old_data, new_data, err);
@@ -550,11 +555,13 @@
 
   /*==========================================================*/
   /* Static meathods called by drizzle to notify interested plugins 
-   * of a session event,
- */
+   * of a session event.
+   *
+   * A quick test is done first to see if there are any interested observers.
+*/
   bool EventObserver::beforeCreateDatabase(Session &session, const std::string &db)
   {
-    if (all_event_plugins.empty())
+    if (all_event_plugins.empty() || !SessionEventData::hasEvents(session))
       return false;
 
     BeforeCreateDatabaseEventData eventData(session, db);
@@ -563,7 +570,7 @@
 
   bool EventObserver::afterCreateDatabase(Session &session, const std::string &db, int err)
   {
-    if (all_event_plugins.empty())
+    if (all_event_plugins.empty() || !SessionEventData::hasEvents(session))
       return false;
 
     AfterCreateDatabaseEventData eventData(session, db, err);
@@ -572,7 +579,7 @@
 
   bool EventObserver::beforeDropDatabase(Session &session, const std::string &db)
   {
-    if (all_event_plugins.empty())
+    if (all_event_plugins.empty() || !SessionEventData::hasEvents(session))
       return false;
 
     BeforeDropDatabaseEventData eventData(session, db);
@@ -581,7 +588,7 @@
 
   bool EventObserver::afterDropDatabase(Session &session, const std::string &db, int err)
   {
-    if (all_event_plugins.empty())
+    if (all_event_plugins.empty() || !SessionEventData::hasEvents(session))
       return false;
 
     AfterDropDatabaseEventData eventData(session, db, err);
@@ -590,7 +597,7 @@
 
   bool EventObserver::connectSession(Session &session)
   {
-    if (all_event_plugins.empty())
+    if (all_event_plugins.empty() || !SessionEventData::hasEvents(session))
       return false;
 
     ConnectSessionEventData eventData(session);
@@ -599,7 +606,7 @@
 
   bool EventObserver::disconnectSession(Session &session)
   {
-    if (all_event_plugins.empty())
+    if (all_event_plugins.empty() || !SessionEventData::hasEvents(session))
       return false;
 
     DisconnectSessionEventData eventData(session);
@@ -608,7 +615,7 @@
 
   bool EventObserver::beforeStatement(Session &session)
   {
-    if (all_event_plugins.empty())
+    if (all_event_plugins.empty() || !SessionEventData::hasEvents(session))
       return false;
 
     BeforeStatementEventData eventData(session);
@@ -617,7 +624,7 @@
 
   bool EventObserver::afterStatement(Session &session)
   {
-    if (all_event_plugins.empty())
+    if (all_event_plugins.empty() || !SessionEventData::hasEvents(session))
       return false;
 
     AfterStatementEventData eventData(session);

=== modified file 'drizzled/plugin/event_observer.h'
--- drizzled/plugin/event_observer.h	2010-10-20 17:02:21 +0000
+++ drizzled/plugin/event_observer.h	2010-10-22 16:47:38 +0000
@@ -44,6 +44,7 @@
 #define DRIZZLED_PLUGIN_EVENT_OBSERVER_H
 
 #include "drizzled/plugin/plugin.h"
+#include "drizzled/session.h"
 
 #include <string>
 
@@ -292,6 +293,8 @@
 
   // Call all the event observers that are registered for this event.
   virtual bool callEventObservers();
+  
+  static bool hasEvents(Session &session) { return (session.getSessionObservers() != NULL);}
 };
 
 //-----
@@ -311,6 +314,7 @@
 
   // Call all the event observers that are registered for this event.
   virtual bool callEventObservers();
+  
 };
 
 //-----
@@ -330,6 +334,8 @@
 
   // Call all the event observers that are registered for this event.
   virtual bool callEventObservers();
+  
+  static bool hasEvents(Table &table) { return (table.getMutableShare()->getTableObservers() != NULL);}
 };
 
 //-----

=== modified file 'drizzled/util/functors.h'
--- drizzled/util/functors.h	2010-10-18 18:24:29 +0000
+++ drizzled/util/functors.h	2010-10-20 20:41:00 +0000
@@ -34,6 +34,17 @@
   }
 };
 
+class SafeDeletePtr
+{
+public:
+  template<typename T>
+  inline void operator()(const T *ptr) const
+  {
+    if (ptr)
+			delete ptr;
+  }
+};
+
 } /* namespace drizzled */
 
 #endif /* DRIZZLED_UTIL_FUNCTORS_H */

=== modified file 'm4/pandora_optimize.m4'
--- m4/pandora_optimize.m4	2010-08-20 17:54:51 +0000
+++ m4/pandora_optimize.m4	2010-10-18 21:21:20 +0000
@@ -22,7 +22,7 @@
     dnl with using AC_CC_STD_C99 above
     CC="${CC} -std=gnu99"
 
-    AM_CPPFLAGS="-ggdb3 ${AM_CPPFLAGS}"
+    AM_CPPFLAGS="-g ${AM_CPPFLAGS}"
 
     DEBUG_CFLAGS="-O0"
     DEBUG_CXXFLAGS="-O0"

=== modified file 'plugin/hello_events/hello_events.cc'
--- plugin/hello_events/hello_events.cc	2010-10-15 21:33:19 +0000
+++ plugin/hello_events/hello_events.cc	2010-10-20 20:41:00 +0000
@@ -25,9 +25,9 @@
  * This plugin is an example events plugin that just prints some info for
  * the events that it is tracking. 
  *  
-set global hello_events_enable = ON;
-set global hello_events_watch_databases = "x";   
-set global hello_events_watch_tables = "x,y";
+set global hello_events1_enable = ON;
+set global hello_events1_watch_databases = "x";   
+set global hello_events1_watch_tables = "x,y";
 
  */
 
@@ -190,8 +190,16 @@
 HelloEvents::~HelloEvents()
 {
   /* These are strdup'd in option processing */
-  free(sysvar_table_list);
-  free(sysvar_db_list);
+  if (sysvar_table_list) {
+	free(sysvar_table_list);
+	sysvar_table_list = NULL;
+  }
+  
+  if (sysvar_db_list) {
+	free(sysvar_db_list);
+	sysvar_db_list = NULL;
+  }
+  
 }
 
 //==================================

=== modified file 'plugin/innobase/CMakeLists.txt'
--- plugin/innobase/CMakeLists.txt	2009-08-12 06:25:19 +0000
+++ plugin/innobase/CMakeLists.txt	2010-10-18 21:21:20 +0000
@@ -11,7 +11,7 @@
 # 
 # You should have received a copy of the GNU General Public License
 # along with this program; if not, write to the Free Software
-# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02111-1307  USA
 
 # This is the CMakeLists for InnoDB Plugin
 

=== modified file 'plugin/innobase/pars/pars0grm.y'
--- plugin/innobase/pars/pars0grm.y	2010-09-11 21:17:22 +0000
+++ plugin/innobase/pars/pars0grm.y	2010-10-18 21:21:20 +0000
@@ -12,7 +12,7 @@
 
 You should have received a copy of the GNU General Public License along with
 this program; if not, write to the Free Software Foundation, Inc., 59 Temple
-Place, Suite 330, Boston, MA 02111-1307 USA
+Place, Suite 330, Boston, MA 02110-1301 USA
 
 *****************************************************************************/
 

=== modified file 'plugin/innobase/pars/pars0lex.l'
--- plugin/innobase/pars/pars0lex.l	2010-10-12 07:33:36 +0000
+++ plugin/innobase/pars/pars0lex.l	2010-10-20 20:41:00 +0000
@@ -12,7 +12,7 @@
 
 You should have received a copy of the GNU General Public License along with
 this program; if not, write to the Free Software Foundation, Inc., 59 Temple
-Place, Suite 330, Boston, MA 02111-1307 USA
+Place, Suite 330, Boston, MA 02110-1301 USA
 
 *****************************************************************************/
 

=== modified file 'plugin/innobase/plugin.ini'
--- plugin/innobase/plugin.ini	2010-10-19 21:58:26 +0000
+++ plugin/innobase/plugin.ini	2010-10-20 20:41:00 +0000
@@ -11,7 +11,7 @@
 # 
 # You should have received a copy of the GNU General Public License along with
 # this program; if not, write to the Free Software Foundation, Inc., 59 Temple
-# Place, Suite 330, Boston, MA 02111-1307 USA
+# Place, Suite 330, Boston, MA 02110-1301 USA
 #
 
 [plugin]

=== modified file 'plugin/pbms/plugin.ini'
--- plugin/pbms/plugin.ini	2010-08-17 20:57:43 +0000
+++ plugin/pbms/plugin.ini	2010-10-18 21:21:20 +0000
@@ -11,7 +11,7 @@
 # 
 # You should have received a copy of the GNU General Public License along with
 # this program; if not, write to the Free Software Foundation, Inc., 59 Temple
-# Place, Suite 330, Boston, MA 02111-1307 USA
+# Place, Suite 330, Boston, MA 02110-1301 USA
 #
 
 [plugin]
@@ -74,6 +74,7 @@
 		src/cslib/CSTime.cc
 		src/cslib/CSUTF8.cc
 		src/cslib/CSXML.cc
+		src/cslib/CSSys_unix.cc
 
 headers=src/backup_ms.h
 		src/cloud_ms.h
@@ -104,6 +105,7 @@
 		src/cslib/CSTime.h
 		src/cslib/CSUTF8.h
 		src/cslib/CSXML.h
+		src/cslib/CSSys.h
 		src/database_ms.h
 		src/defs_ms.h
 		src/discover_ms.h

=== modified file 'plugin/pbms/src/TransTest.cc'
--- plugin/pbms/src/TransTest.cc	2010-10-02 21:15:42 +0000
+++ plugin/pbms/src/TransTest.cc	2010-10-18 21:21:20 +0000
@@ -523,15 +523,15 @@
 	if (log_row_cnt == ref_row_cnt) {
 		for ( i = 0; i < log_row_cnt; i++) {
 			l_record = mysql_fetch_row(l_results);
-			tab_id = atol(l_getUpdateRecord());
+			tab_id = atol(l_record[1]);
 			r_record = mysql_fetch_row(r_results[tab_id-1]);		
-			if ((atol(l_getInsertRecord()) != atol(r_getInsertRecord())) ||
-				(atol(l_getUpdateRecord()) != atol(r_getUpdateRecord())) ||
+			if ((atol(l_record[0]) != atol(r_record[0])) ||
+				(atol(l_record[1]) != atol(r_record[1])) ||
 				(atol(l_record[2]) != atol(r_record[2]))) {
 				
 				printf("verify_database() Failed: in row %d, tab_id %d\n", i+1, tab_id);
-				printf("field 1:  %d =? %d\n", atol(l_getInsertRecord()), atol(r_getInsertRecord()));
-				printf("field 2:  %d =? %d\n", atol(l_getUpdateRecord()), atol(r_getUpdateRecord()));
+				printf("field 1:  %d =? %d\n", atol(l_record[0]), atol(r_record[0]));
+				printf("field 2:  %d =? %d\n", atol(l_record[1]), atol(r_record[1]));
 				printf("field 3:  %d =? %d\n", atol(l_record[2]), atol(r_record[2]));
 				goto done;
 			}
@@ -716,7 +716,7 @@
 						mysql_data_seek(results, rand()%cnt);
 						record = mysql_fetch_row(results);
 							
-						blob_ref_id = atol(getInsertRecord());
+						blob_ref_id = atol(record[0]);
 						blob_id = atol(record[2]);
 						
 						sprintf(stmt, "DELETE FROM "REF_TABLE" WHERE blob_ref = %"PRIu64" AND blob_id = %"PRIu64"", tab_id, blob_ref_id, blob_id); 

=== modified file 'plugin/pbms/src/backup_ms.cc'
--- plugin/pbms/src/backup_ms.cc	2010-10-02 21:15:42 +0000
+++ plugin/pbms/src/backup_ms.cc	2010-10-18 21:21:20 +0000
@@ -127,10 +127,40 @@
 }
 
 //-------------------------------
+class StartDumpCleanUp : public CSRefObject {
+	bool do_cleanup;
+	uint32_t ref_id;
+
+	public:
+	
+	StartDumpCleanUp(): CSRefObject(),
+		do_cleanup(false){}
+		
+	~StartDumpCleanUp() 
+	{
+		if (do_cleanup) {
+			MSBackupInfo::gBackupInfo->remove(ref_id);
+		}
+	}
+	
+	void setCleanUp(uint32_t id)
+	{
+		ref_id = id;
+		do_cleanup = true;
+	}
+	
+	void cancelCleanUp()
+	{
+		do_cleanup = false;
+	}
+	
+};
+
 MSBackupInfo *MSBackupInfo::startDump(MSDatabase *db, uint32_t cloud_ref, uint32_t backup_no)
 {
 	MSBackupInfo *info;
 	uint32_t ref_id;
+	StartDumpCleanUp *cleanup;
 	
 	enter_();
 	push_(db);
@@ -146,17 +176,22 @@
 
 	pop_(info);
 	unlock_(gBackupInfo);
-	pop_(db);
-	
-	try_(a) {
-		MSBackupTable::saveTable(db);
-	}
-	catch_(a);
-	gBackupInfo->remove(ref_id);
-	info->release();
-	throw_();
-	
-	cont_(a);
+	push_(info);
+	
+	// Create a cleanup object to handle cleanup
+	// after a possible exception.
+	new_(cleanup, StartDumpCleanUp());
+	push_(cleanup);
+	cleanup->setCleanUp(ref_id);
+	
+	MSBackupTable::saveTable(RETAIN(db));
+	
+	cleanup->cancelCleanUp();
+	release_(cleanup);
+	
+	pop_(info);
+	release_(db);
+
 	return_(info);
 }
 //-------------------------------
@@ -220,111 +255,143 @@
 	return_(bu);
 }
 
+//-------------------------------
+class StartBackupCleanUp : public CSRefObject {
+	bool do_cleanup;
+	MSBackup *backup;
+
+	public:
+	
+	StartBackupCleanUp(): CSRefObject(),
+		do_cleanup(false){}
+		
+	~StartBackupCleanUp() 
+	{
+		if (do_cleanup) {
+			backup->completeBackup();
+		}
+	}
+	
+	void setCleanUp(MSBackup *bup)
+	{
+		backup = bup;
+		do_cleanup = true;
+	}
+	
+	void cancelCleanUp()
+	{
+		do_cleanup = false;
+	}
+	
+};
+
 void MSBackup::startBackup(MSDatabase *src_db)
 {
 	CSSyncVector	*repo_list;
 	bool			compacting = false;
 	MSRepository	*repo;
+	StartBackupCleanUp *cleanup;
 	enter_();
 
-	try_(a) {
-		bu_SourceDatabase = src_db;
-		repo_list = bu_SourceDatabase->getRepositoryList();
-		// Suspend the compactor before locking the list.
-		bu_Compactor = bu_SourceDatabase->getCompactorThread();
-		if (bu_Compactor) {
-			bu_Compactor->retain();
-			bu_Compactor->suspend();
-		}
-
-		// Build the list of repositories to be backed up.
-		lock_(repo_list);
-
-		new_(bu_BackupList, CSVector(repo_list->size()));
-		for (uint32_t i = 0; i<repo_list->size(); i++) {
-			if ((repo = (MSRepository *) repo_list->get(i))) {
-				if (!repo->isRemovingFP && !repo->mustBeDeleted) {
-					bu_BackupList->add(RETAIN(repo));
-					if (repo->initBackup() == REPO_COMPACTING) 
-						compacting = true; 
-					
-					if (!repo->myRepoHeadSize) {
-						/* The file has not yet been opened, so the
-						 * garbage count will not be known!
-						 */
-						MSRepoFile *repo_file;
-
-						//repo->retain();
-						//unlock_(myRepostoryList);
-						//push_(repo);
-						repo_file = repo->openRepoFile();
-						repo_file->release();
-						//release_(repo);
-						//lock_(myRepostoryList);
-						//goto retry;
-					}
-					
-					bu_size += repo->myRepoFileSize; 
-
+	// Create a cleanup object to handle cleanup
+	// after a possible exception.
+	new_(cleanup, StartBackupCleanUp());
+	push_(cleanup);
+	cleanup->setCleanUp(this);
+
+	bu_SourceDatabase = src_db;
+	repo_list = bu_SourceDatabase->getRepositoryList();
+	// Suspend the compactor before locking the list.
+	bu_Compactor = bu_SourceDatabase->getCompactorThread();
+	if (bu_Compactor) {
+		bu_Compactor->retain();
+		bu_Compactor->suspend();
+	}
+
+	// Build the list of repositories to be backed up.
+	lock_(repo_list);
+
+	new_(bu_BackupList, CSVector(repo_list->size()));
+	for (uint32_t i = 0; i<repo_list->size(); i++) {
+		if ((repo = (MSRepository *) repo_list->get(i))) {
+			if (!repo->isRemovingFP && !repo->mustBeDeleted) {
+				bu_BackupList->add(RETAIN(repo));
+				if (repo->initBackup() == REPO_COMPACTING) 
+					compacting = true; 
+				
+				if (!repo->myRepoHeadSize) {
+					/* The file has not yet been opened, so the
+					 * garbage count will not be known!
+					 */
+					MSRepoFile *repo_file;
+
+					//repo->retain();
+					//unlock_(myRepostoryList);
+					//push_(repo);
+					repo_file = repo->openRepoFile();
+					repo_file->release();
+					//release_(repo);
+					//lock_(myRepostoryList);
+					//goto retry;
 				}
+				
+				bu_size += repo->myRepoFileSize; 
+
 			}
 		}
-		
-		// Copy the table list to the backup database:
-		uint32_t		next_tab = 0;
-		MSTable		*tab;
-		while ((tab = bu_SourceDatabase->getNextTable(&next_tab))) {
-			push_(tab);
-			bu_Database->addTable(tab->myTableID, tab->myTableName->getCString(), 0, false);
-			release_(tab);
-		}
-		unlock_(repo_list);
-		
-		// Copy over any physical PBMS system tables.
-		PBMSSystemTables::transferSystemTables(RETAIN(bu_Database), RETAIN(bu_SourceDatabase));
-
-		// Load the system tables into the backup database. This will
-		// initialize the database for cloud storage if required.
-		PBMSSystemTables::loadSystemTables(RETAIN(bu_Database));
-		
-		// Set the cloud backup info.
-		bu_Database->myBlobCloud->cl_setBackupInfo(RETAIN(bu_info));
-		
-		
-		// Set the backup number in the pbms_variable tabe. (This is a hidden value.)
-		// This value is used in case a drag and drop restore was done. When a data base is
-		// first loaded this value is checked and if it is not zero then the backup record
-		// will be read and any used to recover any BLOBs.
-		// 
-		char value[20];
-		snprintf(value, 20, "%"PRIu32"", bu_info->getBackupRefId());
-		MSVariableTable::setVariable(RETAIN(bu_Database), BACKUP_NUMBER_VAR, value);
-		
-		// Once the repositories are locked the compactor can be restarted
-		// unless it is in the process of compacting a repository that is
-		// being backed up.
-		if (bu_Compactor && !compacting) {	
-			bu_Compactor->resume();		
-			bu_Compactor->release();		
-			bu_Compactor = NULL;		
-		}
-		
-		// Suspend the transaction writer while the backup is running.
-		MSTransactionManager::suspend(true);
-		bu_TransactionManagerSuspended = true;
-		
-		// Start the backup daemon thread.
-		bu_ID = bu_start_time = time(NULL);
-		start();
-	}
-	catch_(a) {
-		completeBackup();
-		throw_();
-	}
-	cont_(a);
-	
+	}
+	
+	// Copy the table list to the backup database:
+	uint32_t		next_tab = 0;
+	MSTable		*tab;
+	while ((tab = bu_SourceDatabase->getNextTable(&next_tab))) {
+		push_(tab);
+		bu_Database->addTable(tab->myTableID, tab->myTableName->getCString(), 0, false);
+		release_(tab);
+	}
+	unlock_(repo_list);
+	
+	// Copy over any physical PBMS system tables.
+	PBMSSystemTables::transferSystemTables(RETAIN(bu_Database), RETAIN(bu_SourceDatabase));
+
+	// Load the system tables into the backup database. This will
+	// initialize the database for cloud storage if required.
+	PBMSSystemTables::loadSystemTables(RETAIN(bu_Database));
+	
+	// Set the cloud backup info.
+	bu_Database->myBlobCloud->cl_setBackupInfo(RETAIN(bu_info));
+	
+	
+	// Set the backup number in the pbms_variable tabe. (This is a hidden value.)
+	// This value is used in case a drag and drop restore was done. When a data base is
+	// first loaded this value is checked and if it is not zero then the backup record
+	// will be read and any used to recover any BLOBs.
+	// 
+	char value[20];
+	snprintf(value, 20, "%"PRIu32"", bu_info->getBackupRefId());
+	MSVariableTable::setVariable(RETAIN(bu_Database), BACKUP_NUMBER_VAR, value);
+	
+	// Once the repositories are locked the compactor can be restarted
+	// unless it is in the process of compacting a repository that is
+	// being backed up.
+	if (bu_Compactor && !compacting) {	
+		bu_Compactor->resume();		
+		bu_Compactor->release();		
+		bu_Compactor = NULL;		
+	}
+	
+	// Suspend the transaction writer while the backup is running.
+	MSTransactionManager::suspend(true);
+	bu_TransactionManagerSuspended = true;
+	
+	// Start the backup daemon thread.
+	bu_ID = bu_start_time = time(NULL);
+	start();
+	
+	cleanup->cancelCleanUp();
+	release_(cleanup);
+
 	exit_();
-
 }
 
 void MSBackup::completeBackup()
@@ -379,46 +446,46 @@
 
 bool MSBackup::doWork()
 {
-	CSMutex				*my_lock;
-	MSRepository		*src_repo, *dst_repo;
-	MSRepoFile			*src_file, *dst_file;
-	off64_t				src_offset, prev_offset;
-	uint16_t				head_size;
-	uint64_t				blob_size, blob_data_size;
-	CSStringBuffer		*head;
-	MSRepoPointersRec	ptr;
-	uint32_t				table_ref_count;
-	uint32_t				blob_ref_count;
-	int					ref_count;
-	size_t				ref_size;
-	uint32_t				auth_code;
-	uint32_t				tab_id;
-	uint64_t				blob_id;
-	MSOpenTable			*otab;
-	uint32_t				src_repo_id;
-	uint8_t				status;
-	uint8_t				blob_storage_type;
-	uint16_t				tab_index;
-	uint32_t				mod_time;
-	char				transferBuffer[MS_BACKUP_BUFFER_SIZE];
-	CloudKeyRec			cloud_key;
-
-	
 	enter_();
-	bu_BackupRunning = true;
-	bu_State = BU_RUNNING; 
-
-/*
-	// For testing:
-	{
-		int blockit = 0;
-		myWaitTime = 5 * 1000;  // Time in milli-seconds
-		while (blockit)
-			return_(true);
-	}
-*/
-	
 	try_(a) {
+		CSMutex				*my_lock;
+		MSRepository		*src_repo, *dst_repo;
+		MSRepoFile			*src_file, *dst_file;
+		off64_t				src_offset, prev_offset;
+		uint16_t				head_size;
+		uint64_t				blob_size, blob_data_size;
+		CSStringBuffer		*head;
+		MSRepoPointersRec	ptr;
+		uint32_t				table_ref_count;
+		uint32_t				blob_ref_count;
+		int					ref_count;
+		size_t				ref_size;
+		uint32_t				auth_code;
+		uint32_t				tab_id;
+		uint64_t				blob_id;
+		MSOpenTable			*otab;
+		uint32_t				src_repo_id;
+		uint8_t				status;
+		uint8_t				blob_storage_type;
+		uint16_t				tab_index;
+		uint32_t				mod_time;
+		char				transferBuffer[MS_BACKUP_BUFFER_SIZE];
+		CloudKeyRec			cloud_key;
+
+	
+		bu_BackupRunning = true;
+		bu_State = BU_RUNNING; 
+
+	/*
+		// For testing:
+		{
+			int blockit = 0;
+			myWaitTime = 5 * 1000;  // Time in milli-seconds
+			while (blockit)
+				return_(true);
+		}
+	*/
+	
 		new_(head, CSStringBuffer(100));
 		push_(head);
 
@@ -565,7 +632,7 @@
 						if (blob_storage_type == MS_CLOUD_STORAGE) { 
 							bu_Database->myBlobCloud->cl_backupBLOB(&cloud_key);
 						} else
-							CSFile::transfer(dst_file, dst_offset + head_size, src_file, src_offset + head_size, blob_size, transferBuffer, MS_BACKUP_BUFFER_SIZE);
+							CSFile::transfer(RETAIN(dst_file), dst_offset + head_size, RETAIN(src_file), src_offset + head_size, blob_size, transferBuffer, MS_BACKUP_BUFFER_SIZE);
 					
 						/* Update the references: */
 						ptr.rp_chars = head->getBuffer(0) + src_repo->myRepoBlobHeadSize;

=== modified file 'plugin/pbms/src/backup_ms.h'
--- plugin/pbms/src/backup_ms.h	2010-10-02 21:15:42 +0000
+++ plugin/pbms/src/backup_ms.h	2010-10-18 21:21:20 +0000
@@ -34,6 +34,9 @@
 class MSDatabase;
 
 class MSBackupInfo : public CSRefObject {
+	friend class StartDumpCleanUp;
+	friend class InsertRowCleanUp;
+	
 	private:
 	static uint32_t	gMaxInfoRef;
 	static CSSyncSparseArray *gBackupInfo;
@@ -155,6 +158,7 @@
 	
 	static MSBackup* newMSBackup(MSBackupInfo *backup_info);
 
+	friend class StartBackupCleanUp;
 private:
 	void completeBackup();
 	

=== modified file 'plugin/pbms/src/cloud_ms.cc'
--- plugin/pbms/src/cloud_ms.cc	2010-10-02 21:15:42 +0000
+++ plugin/pbms/src/cloud_ms.cc	2010-10-18 21:21:20 +0000
@@ -269,43 +269,40 @@
 	release_(src_objectKey);
 	push_(list);
 	
+	
 	// Go through the list copying the keys.
-	try_(a) {
-		dst_cloudRef = src_cloudRef;
-		dst_cloud = src_cloud;
-		dst_cloud->retain();
+	dst_cloudRef = src_cloudRef;
+	dst_cloud = src_cloud;
+	dst_cloud->retain();
 	
-		while ((key = (CSString*)(list->take(0))) ) {
-			push_(key);
-
-			// The source key name must be parsed to get its
-			// destination cloud reference. The destination for
-			// the BLOBs may not all be in the same cloud. 
-			CloudObjectKey::parseObjectKey(key->getCString(), &cloudKey);
-			
-			// Reset the destination cloud if required.
-			if (cloudKey.cloud_ref != dst_cloudRef) {
-				if (dst_cloud) {
-					dst_cloud->release();
-					dst_cloud = NULL;
-				}
-				dst_cloudRef = 	cloudKey.cloud_ref;
-				dst_cloud = MSCloudInfo::getCloudInfo(dst_cloudRef);
+	push_ref_(dst_cloud); // Push a reference to dst_cloud so that what ever it references will be released.
+
+	while ((key = (CSString*)(list->take(0))) ) {
+		push_(key);
+
+		// The source key name must be parsed to get its
+		// destination cloud reference. The destination for
+		// the BLOBs may not all be in the same cloud. 
+		CloudObjectKey::parseObjectKey(key->getCString(), &cloudKey);
+		
+		// Reset the destination cloud if required.
+		if (cloudKey.cloud_ref != dst_cloudRef) {
+			if (dst_cloud) {
+				dst_cloud->release();
+				dst_cloud = NULL;
 			}
+			dst_cloudRef = 	cloudKey.cloud_ref;
+			dst_cloud = MSCloudInfo::getCloudInfo(dst_cloudRef);
+		}
 
-			// Copy the BLOB to the recovered database.
-			dst_objectKey->setObjectKey(&cloudKey);
-			src_cloud->copy(RETAIN(dst_cloud), dst_objectKey->getCString(), key->getCString());
-			release_(key);
-			
-		}
+		// Copy the BLOB to the recovered database.
+		dst_objectKey->setObjectKey(&cloudKey);
+		src_cloud->copy(RETAIN(dst_cloud), dst_objectKey->getCString(), key->getCString());
+		release_(key);
+		
 	}
 	
-	finally_(a) {
-		if (dst_cloud)
-			dst_cloud->release();
-	}
-	finally_end_block(a);
+	release_(dst_cloud);
 	
 	blob_recovery_no = 0;
 	release_(list);

=== modified file 'plugin/pbms/src/compactor_ms.cc'
--- plugin/pbms/src/compactor_ms.cc	2010-10-07 01:20:26 +0000
+++ plugin/pbms/src/compactor_ms.cc	2010-10-18 21:21:20 +0000
@@ -277,7 +277,7 @@
 				dst_file->write(head->getBuffer(0), dst_offset, head_size);
 
 				/* We have an engine reference, copy the BLOB over: */
-				CSFile::transfer(dst_file, dst_offset + head_size, src_file, src_offset + head_size, blob_size, iCompactBuffer, MS_COMPACTOR_BUFFER_SIZE);
+				CSFile::transfer(RETAIN(dst_file), dst_offset + head_size, RETAIN(src_file), src_offset + head_size, blob_size, iCompactBuffer, MS_COMPACTOR_BUFFER_SIZE);
 
 #ifdef HAVE_ALIAS_SUPPORT
 				/* If the BLOB has an alias update the alias index. */

=== modified file 'plugin/pbms/src/connection_handler_ms.cc'
--- plugin/pbms/src/connection_handler_ms.cc	2010-10-07 01:20:26 +0000
+++ plugin/pbms/src/connection_handler_ms.cc	2010-10-18 21:21:20 +0000
@@ -27,22 +27,21 @@
  *
  */
 
-#include "config.h"
-
-#include "connection_handler_ms.h"
-#include "network_ms.h"
-#include "open_table_ms.h"
-#include "engine_ms.h"
-#include "version_ms.h"
-
 #include "cslib/CSConfig.h"
 #include <inttypes.h>
 
+#include "defs_ms.h"
+
 #include "cslib/CSGlobal.h"
 #include "cslib/CSSocket.h"
 #include "cslib/CSStrUtil.h"
 #include "cslib/CSHTTPStream.h"
 
+#include "connection_handler_ms.h"
+#include "network_ms.h"
+#include "open_table_ms.h"
+#include "engine_ms.h"
+#include "version_ms.h"
 
 //#include "mysql_ms.h"
 
@@ -451,7 +450,10 @@
 	new_(metadata, CSStringBuffer(80));
 	push_(metadata);
 	
-	blob_len = iInputStream->getContentLength();
+	 if (! iInputStream->getContentLength(&blob_len)) {
+		CSException::throwException(CS_CONTEXT, CS_ERR_MISSING_HTTP_HEADER, "Missing content lenght header");
+	 }
+	 
 	
 	// Collect the meta data.
 	for (uint32_t i = 0; i < iInputStream->numHeaders(); i++) {

=== modified file 'plugin/pbms/src/connection_handler_ms.h'
--- plugin/pbms/src/connection_handler_ms.h	2010-10-07 01:20:26 +0000
+++ plugin/pbms/src/connection_handler_ms.h	2010-10-18 21:21:20 +0000
@@ -30,12 +30,11 @@
 #ifndef __CONNECTIONHANDLER_MS_H__
 #define __CONNECTIONHANDLER_MS_H__
 
-#include "engine_ms.h"
-
 #include "cslib/CSDefs.h"
 #include "cslib/CSThread.h"
 #include "cslib/CSHTTPStream.h"
 
+#include "engine_ms.h"
 
 class MSConnectionHandler : public CSDaemon {
 public:

=== modified file 'plugin/pbms/src/cslib/CSConfig.h'
--- plugin/pbms/src/cslib/CSConfig.h	2010-10-02 21:15:42 +0000
+++ plugin/pbms/src/cslib/CSConfig.h	2010-10-18 21:21:20 +0000
@@ -43,18 +43,16 @@
 #include "mysql_priv.h"
 #endif
 
-/* Note: mysql_priv.h messes with new, which caused a crash. */
-#ifdef new
-#undef new
-#endif
+#else  // defined(MYSQL_SERVER) ||  defined(DRIZZLED)
 
+#if defined(WIN32) ||  defined(WIN64) 
+#include "win_config.h"
 #else
 #include "config.h"
-#endif	//defined(MYSQL_SERVER) ||  defined(DRIZZLED)
-
-#include <sys/types.h>
-#include <inttypes.h>
-#include <unistd.h>
+#endif
+
+#endif // defined(MYSQL_SERVER) ||  defined(DRIZZLED)
+
 
 /*
  * This enables everything that GNU can do. The macro is actually
@@ -64,42 +62,48 @@
 #define _GNU_SOURCE
 #endif
 
-#ifndef NODEBUG
-#ifdef DEBUG
-#ifndef DBUG_OFF
-#ifndef DEBUG
-#define DEBUG
-#endif
-#endif
-#endif
+/*
+ * Make sure we use the thread safe version of the library.
+ */
+#ifndef _THREAD_SAFE // Seems to be defined by some Drizzle header
+#define _THREAD_SAFE
+#endif
+
+/*
+ * This causes things to be defined like stuff in inttypes.h
+ * which is used in printf()
+ */
+#ifndef __STDC_FORMAT_MACROS
+#define __STDC_FORMAT_MACROS
 #endif
 
 /*
  * What operating system are we on?
  */
-#ifdef __darwin__
+#ifdef __APPLE__
 #define OS_MACINTOSH
 #endif
 
-#if defined(MSDOS) || defined(__WIN__) || defined(_WIN64)
+#if defined(MSDOS) || defined(__WIN__) || defined(_WIN64) || defined(WIN32)
 #define OS_WINDOWS
 #endif
 
+#ifdef __FreeBSD__
+#define OS_FREEBSD
+#endif
+
+#ifdef __NetBSD__
+#define OS_NETBSD
+#endif
+
 #ifdef __sun
 #define OS_SOLARIS
 #endif
 
-
-#ifdef OS_WINDOWS
-#ifdef _DEBUG
-#ifndef NODEBUG
-#ifndef DBUG_OFF
-#ifndef DEBUG
+#if defined(_DEBUG) || defined(DBUG_ON)
+#if !defined(NODEBUG) && !defined(DBUG_OFF) && !defined(DEBUG)
 #define DEBUG
 #endif
 #endif
-#endif
-#endif
-#endif
 
 #endif

=== modified file 'plugin/pbms/src/cslib/CSDefs.h'
--- plugin/pbms/src/cslib/CSDefs.h	2010-10-02 21:15:42 +0000
+++ plugin/pbms/src/cslib/CSDefs.h	2010-10-20 20:21:41 +0000
@@ -35,6 +35,18 @@
 // Use standard portable data types
 #include <stdint.h>
 
+/* Those compilers that support the function
+ * macro (in particular the "pretty" function
+ * macro must be defined here.
+ */
+#ifdef OS_WINDOWS
+#define __FUNC__				__FUNCTION__
+#elif defined(OS_SOLARIS)
+#define __FUNC__				"__func__"
+#else
+#define __FUNC__				__PRETTY_FUNCTION__
+#endif
+
 /*
  * An unsigned integer, 1 byte long:
  */
@@ -47,7 +59,7 @@
  */
 #define s_char			unsigned char
 
-/* PBMS assumes that off_t is 8 bytes so to ensure this always use  off64_t*/
+/* We assumes that off_t is 8 bytes so to ensure this always use  off64_t*/
 #define off64_t			uint64_t
 
 
@@ -71,6 +83,14 @@
 #define CS_DIR_DELIM		"\\"
 #define IS_DIR_CHAR(ch)		((ch) == CS_DIR_CHAR || (ch) == '/')
 
+#ifndef PATH_MAX
+#define PATH_MAX		MAX_PATH
+#endif
+
+#ifndef NAME_MAX
+#define NAME_MAX		MAX_PATH
+#endif
+
 #else
 
 #define CS_DEFAULT_EOL		"\n"
@@ -84,20 +104,6 @@
 #define CS_RELEASE_STACK_SIZE	200
 #define CS_JUMP_STACK_SIZE		20
 
-/* Fixed length types */
-
-#ifdef CS_WIN
-
-#ifndef PATH_MAX
-#define PATH_MAX		MAX_PATH
-#endif
-
-#ifndef NAME_MAX
-#define NAME_MAX		MAX_PATH
-#endif
-
-#endif // CS_WIN
-
 /* C string display width sizes including space for a null terminator and possible sign. */
 #define CS_WIDTH_INT_8	5
 #define CS_WIDTH_INT_16	7
@@ -299,5 +305,11 @@
 		char *rec_chars;
 		CSIntPtr int_val;
 } CSDiskData;
+
+#define CHECKSUM_VALUE_SIZE			16
+typedef struct {
+	u_char val[CHECKSUM_VALUE_SIZE];
+} Md5Digest;
+
 	
 #endif

=== modified file 'plugin/pbms/src/cslib/CSDirectory.cc'
--- plugin/pbms/src/cslib/CSDirectory.cc	2010-10-02 21:15:42 +0000
+++ plugin/pbms/src/cslib/CSDirectory.cc	2010-10-18 21:21:20 +0000
@@ -29,9 +29,6 @@
 #include "CSConfig.h"
 
 #include <string.h>
-#include <sys/time.h>
-#include <sys/stat.h>
-#include <limits.h>
 
 #include "CSStrUtil.h"
 #include "CSPath.h"
@@ -43,18 +40,6 @@
  * CORE SYSTEM DIRECTORY
  */
 
-CSDirectory::~CSDirectory()
-{
-	enter_();
-	close();
-	if (iPath)
-		iPath->release();
-	if (iEntry)
-		cs_free(iEntry);
-		
-	exit_();
-}
-
 void CSDirectory::print(CSOutputStream *out)
 {
 	char	buffer[500];
@@ -80,75 +65,14 @@
 	}
 }
 
-/*
- * ---------------------------------------------------------------
- * UNIX DIRECTORY
- */
-
-/*
- * The filter may contain one '*' as wildcard.
- */
-void CSDirectory::open()
-{
-	enter_();
-	if (!(iDir = opendir(iPath->getCString())))
-		CSException::throwFileError(CS_CONTEXT, iPath->getCString(), errno);
-	exit_();
-}
-
-void CSDirectory::close()
-{
-	enter_();
-	if (iDir) {
-		closedir(iDir);
-		iDir = NULL;
-	}
-	exit_();
-}
-
-bool CSDirectory::next()
-{
-	int				err;
-	struct dirent	*result;
-
-	enter_();
-	for (;;) {
-		err = readdir_r(iDir, &iEntry->entry, &result);
-		self->interrupted();
-		if (err)
-			CSException::throwFileError(CS_CONTEXT, iPath->getCString(), err);
-		if (!result)
-			break;
-		/* Filter out '.' and '..': */
-		if (iEntry->entry.d_name[0] == '.') {
-			if (iEntry->entry.d_name[1] == '.') {
-				if (iEntry->entry.d_name[2] == '\0')
-					continue;
-			}
-			else {
-				if (iEntry->entry.d_name[1] == '\0')
-					continue;
-			}
-		}
-		break;
-	}
-	return_(result ? true : false);
-}
-
-void CSDirectory::getFilePath(char *path, size_t size)
-{
-	cs_strcpy(size, path, iPath->getCString());
-	cs_add_dir_char(size, path);
-	cs_strcat(size, path, iEntry->entry.d_name);
-}
-
 void CSDirectory::deleteEntry()
 {
 	char path[PATH_MAX];
 
 	enter_();
-	getFilePath(path, PATH_MAX);
-
+	
+	getEntryPath(path, PATH_MAX);
+	
 	CSPath *cs_path = CSPath::newPath(path);
 	push_(cs_path);
 	cs_path->removeFile();
@@ -159,79 +83,71 @@
 
 const char *CSDirectory::name()
 {
-	return iEntry->entry.d_name;
+	return entryName();
 }
 
 bool CSDirectory::isFile()
 {
-#ifdef OS_SOLARIS
+	return entryIsFile();
+}
+
+off64_t CSDirectory::getSize()
+{
 	char path[PATH_MAX];
-	struct stat sb;
-
-	getFilePath(path, PATH_MAX);
-
-	if (stat(path, &sb) == -1) {
-		CSException::throwFileError(CS_CONTEXT, path, errno);
-		return false; // Never reached.
-	}
-
-	if ( sb.st_mode & S_IFDIR )
-		return false;
-#else
-	if (iEntry->entry.d_type & DT_DIR)
-		return false;
-#endif
-	return true;
+
+	getEntryPath(path, PATH_MAX);
+	
+	return CSPath::getSize(path);
 }
 
 void CSDirectory::info(bool *is_dir, off64_t *size, CSTime *mod_time)
 {
 	char path[PATH_MAX];
 
+	getEntryPath(path, PATH_MAX);
+	
+	CSPath::info(path, is_dir, size, mod_time);
+}
+
+bool CSDirectory::exists()
+{
+	CSPath *path;
+	bool yup;
+
 	enter_();
-	getFilePath(path, PATH_MAX);
-
-	CSPath *cs_path = CSPath::newPath(path);
-	push_(cs_path);
-	cs_path->info(is_dir, size, mod_time);
-	release_(cs_path);
-
-	exit_();
+	path = CSPath::newPath(RETAIN(sd_path));
+	push_(path);
+	yup = path->exists();
+	release_(path);
+	return_(yup);
 }
 
 CSDirectory *CSDirectory::newDirectory(CSString *path)
 {
 	CSDirectory *dir;
-	size_t size;
-	enter_();
-	push_(path);
-	
-#ifdef OS_SOLARIS
-	size = pathconf(path->getCString(), _PC_NAME_MAX) + sizeof(struct dirent)  + 1;
-#else
-	size = sizeof(struct dirent);
-#endif
 
 	if (!(dir = new CSDirectory())) {
+		path->release();
 		CSException::throwOSError(CS_CONTEXT, ENOMEM);
 	}
-	pop_(path);
-	dir->iPath = path;
-	push_(dir);
-	
-	dir->iEntry = (union var_dirent *) cs_malloc(size);
-	pop_(dir);
-	
+	dir->sd_path = path;
+	return dir;
+}
+
+CSDirectory *CSDirectory::newDirectory(CSPath *path)
+{
+	CSDirectory *dir;
+	enter_();
+	
+	push_(path);
+	dir = newDirectory(RETAIN(path->getString()));
+	release_(path);
 	return_(dir);
 }
 
-CSDirectory *CSDirectory::newDirectory(CSPath *path)
+CSDirectory *CSDirectory::newDirectory(const char *path)
 {
-	CSString *str = path->getString();
-	str->retain();
-	path->release();
-	
-	return newDirectory(str);
+	return newDirectory(CSString::newString(path));
 }
 
 

=== modified file 'plugin/pbms/src/cslib/CSDirectory.h'
--- plugin/pbms/src/cslib/CSDirectory.h	2010-10-02 21:15:42 +0000
+++ plugin/pbms/src/cslib/CSDirectory.h	2010-10-18 21:21:20 +0000
@@ -29,56 +29,35 @@
 #ifndef __CSDIRECTORY_H__
 #define __CSDIRECTORY_H__
 
-#include <dirent.h>
-
 #include "CSDefs.h"
+#include "CSSys.h"
 #include "CSPath.h"
 #include "CSTime.h"
 #include "CSObject.h"
 #include "CSStream.h"
 
-using namespace std;
-
-class CSDirectory : public CSObject {
+class CSDirectory : public CSSysDir, public CSObject  {
 public:
-	CSDirectory(): iEntry(NULL), iDir(NULL), iPath(NULL) { }
-
-	virtual ~CSDirectory();
-
-	virtual void open();
-
-	virtual void close();
-
-	virtual bool next();
-
-	virtual const char *name();
-
-	virtual bool isFile() ;
-	
-	virtual void deleteEntry() ;
-
-	virtual void info(bool *is_dir, off64_t *size, CSTime *mod_time);
-
-	virtual void print(CSOutputStream *out);
+	const char *name();
+
+	bool isFile();
+	
+	off64_t getSize();
+	
+	void deleteEntry();
+
+	void info(bool *is_dir, off64_t *size, CSTime *mod_time);
+
+	bool exists();
+
+	void print(CSOutputStream *out);
 
 	friend class TDDirectory;
 
-	static CSDirectory *newDirectory(CSPath *);
-	static CSDirectory *newDirectory(CSString *);
+	static CSDirectory *newDirectory(CSPath *path);
+	static CSDirectory *newDirectory(CSString *path);
+	static CSDirectory *newDirectory(const char *path);
 
-private:
-	/* Solaris requires od_entry.d_name member to have size at least as returned
-	 * by pathconf() function on per-directory basis. As a result 'struct dirent'
-	 * cannot be staticly alloacted. 
-	 */
-	union var_dirent {
-		struct dirent entry;
-		char space[1];
-	} *iEntry;
-	
-	DIR *iDir;
-	CSString *iPath;
-	void getFilePath(char *path, size_t size);
 };
 
 

=== modified file 'plugin/pbms/src/cslib/CSEncode.cc'
--- plugin/pbms/src/cslib/CSEncode.cc	2010-10-02 21:15:42 +0000
+++ plugin/pbms/src/cslib/CSEncode.cc	2010-10-18 21:21:20 +0000
@@ -28,7 +28,17 @@
 
 #include "CSEncode.h"
 
-static const u_char base64Map[64] = {
+static const u_char base64URLMap[64] = {
+	'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 
+	'I', 'J', 'K', 'L', 'M', 'N', 'O', 'P', 
+	'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X', 
+	'Y', 'Z', 'a', 'b', 'c', 'd', 'e', 'f', 
+	'g', 'h', 'i', 'j', 'k', 'l', 'm', 'n', 
+	'o', 'p', 'q', 'r', 's', 't', 'u', 'v', 
+	'w', 'x', 'y', 'z', '0', '1', '2', '3', 
+	'4', '5', '6', '7', '8', '9', '-', '_'
+};
+static const u_char base64STDMap[64] = {
 	'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 
 	'I', 'J', 'K', 'L', 'M', 'N', 'O', 'P', 
 	'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X', 
@@ -40,26 +50,26 @@
 };
  
 static const u_char decodeBase64Map[128] = {
-	0XBF, 0XFF, 0XF2, 0X2C, 0XBF, 0XFF, 0XF2, 0X24, 
-	0X0, 0X0, 0X2E, 0X4C, 0X0, 0X0, 0X0, 0X3, 
-	0X0, 0X0, 0X0, 0X3, 0XBF, 0XFF, 0XF1, 0X40, 
-	0X0, 0X0, 0X0, 0X50, 0X8F, 0XE1, 0X56, 0XD8, 
-	0X0, 0X30, 0X6, 0XEC, 0X0, 0X30, 0X6, 0XF0, 
-	0X0, 0X0, 0X2E, 0X3E, 0X0, 0X0, 0X0, 0X3F, 
+	0XFF, 0XFF, 0XFF, 0XFF, 0XFF, 0XFF, 0XFF, 0XFF, 
+	0XFF, 0XFF, 0XFF, 0XFF, 0XFF, 0XFF, 0XFF, 0XFF, 
+	0XFF, 0XFF, 0XFF, 0XFF, 0XFF, 0XFF, 0XFF, 0XFF, 
+	0XFF, 0XFF, 0XFF, 0XFF, 0XFF, 0XFF, 0XFF, 0XFF, 
+	0XFF, 0XFF, 0XFF, 0XFF, 0XFF, 0XFF, 0XFF, 0XFF, 
+	0XFF, 0XFF, 0XFF, 0X3E, 0XFF, 0X3E, 0XFF, 0X3F, 
 	0X34, 0X35, 0X36, 0X37, 0X38, 0X39, 0X3A, 0X3B, 
-	0X3C, 0X3D, 0X2C, 0X68, 0X90, 0X0, 0X17, 0XD8, 
-	0XBF, 0X0, 0X1, 0X2, 0X3, 0X4, 0X5, 0X6, 
-	0X7, 0X8, 0X9, 0XA, 0XB, 0XC, 0XD, 0XE, 
-	0XF, 0X10, 0X11, 0X12, 0X13, 0X14, 0X15, 0X16, 
-	0X17, 0X18, 0X19, 0X80, 0X0, 0X0, 0X0, 0X0, 
-	0X0, 0X1A, 0X1B, 0X1C, 0X1D, 0X1E, 0X1F, 0X20, 
+	0X3C, 0X3D, 0XFF, 0XFF, 0XFF, 0XFF, 0XFF, 0XFF, 
+	0XFF, 0X00, 0X01, 0X02, 0X03, 0X04, 0X05, 0X06, 
+	0X07, 0X08, 0X09, 0X0A, 0X0B, 0X0C, 0X0D, 0X0E, 
+	0X0F, 0X10, 0X11, 0X12, 0X13, 0X14, 0X15, 0X16, 
+	0X17, 0X18, 0X19, 0XFF, 0XFF, 0XFF, 0XFF, 0X3F, 
+	0XFF, 0X1A, 0X1B, 0X1C, 0X1D, 0X1E, 0X1F, 0X20, 
 	0X21, 0X22, 0X23, 0X24, 0X25, 0X26, 0X27, 0X28, 
 	0X29, 0X2A, 0X2B, 0X2C, 0X2D, 0X2E, 0X2F, 0X30, 
-	0X31, 0X32, 0X33, 0X0, 0X90, 0X1B, 0XBA, 0X48, 
+	0X31, 0X32, 0X33, 0XFF, 0XFF, 0XFF, 0XFF, 0XFF, 
 };
 
-
-static bool base64Encoded(const u_char *data, size_t len)
+//------------------
+static bool base64Encoded(const u_char *data, size_t len, const u_char *vc)
 {
 	if (len % 4)
 		return false;
@@ -72,21 +82,33 @@
 		return false;
 		
 	for (; i ; i--, data++) {
-		if (*data > 63)
+		if (((*data < 'A') || (*data > 'Z')) &&
+			((*data < 'a') || (*data > 'z')) &&
+			((*data < '0') || (*data > '9')) &&
+			((*data != vc[0]) && (*data != vc[1])))
 			return false;
 	}
 	
 	return true; // Actually this isn't so much 'true' as 'maybe'
 }
 
-char *base64Encode(const void *data, size_t len)
+//------------------
+static char *genericBase64Encode(const void *data, size_t len, char *encode_buffer, size_t encode_buffer_len, const u_char base64Map[])
 {
 	u_char *wptr, *rptr, *encoding;
 	size_t size;
 	enter_();
 
-	size = ((len + 2) / 3) * 4;
-	encoding = (u_char *) cs_malloc(size +1);
+	size = ((len + 2) / 3) * 4 +1;
+	if ((encode_buffer != NULL) && (encode_buffer_len < size)) 
+		CSException::throwException(CS_CONTEXT, CS_ERR_GENERIC_ERROR, "Base64 encode buffer is too small.");
+	
+	if (encode_buffer)
+		encoding = (u_char *) encode_buffer;
+	else
+		encoding = (u_char *) cs_malloc(size);
+		
+	size--;
 	encoding[size] = 0;
 	
 	wptr= encoding;
@@ -119,26 +141,30 @@
 	return_((char*)encoding);
 }
 
-void *base64Decode(const char *data, size_t len)
+//------------------
+static void *genericBase64Decode(const char *data, size_t len, void *decode_buffer, size_t decode_buffer_len, const u_char base64Map[])
 {
-	uint32_t tail;
 	u_char *wptr, *rptr, *decoding;
 	enter_();
 
 	rptr = (u_char *) data;
 
-	if (!base64Encoded(rptr, len)) 
+	if (!base64Encoded(rptr, len, base64Map +62)) 
 		CSException::throwException(CS_CONTEXT, CS_ERR_GENERIC_ERROR, "String was not Base64 encoded.");
 	
+	if ((decode_buffer != NULL) && (decode_buffer_len < ((len/ 4) * 3))) 
+		CSException::throwException(CS_CONTEXT, CS_ERR_GENERIC_ERROR, "Base64 decoded buffer is too small.");
+	
+	if (decode_buffer)
+		decoding = (u_char *) decode_buffer;
+	else
 	decoding = (u_char *) cs_malloc((len/ 4) * 3);
 	
 	wptr= decoding;
 	
-	tail = len; 
-	while (rptr[tail-1] == '=') 
-		tail--; 
+	while (rptr[len-1] == '=') 
+		len--; 
 
-	len -= tail;
 	while (len/4) {		
 		wptr[0] = ( ((decodeBase64Map[rptr[0]] << 2) & 0xFF) | ((decodeBase64Map[rptr[1]] >> 4) & 0x03) ); 
 		wptr[1] = ( ((decodeBase64Map[rptr[1]] << 4) & 0xFF) | ((decodeBase64Map[rptr[2]] >> 2) & 0x0F) ); 
@@ -158,6 +184,30 @@
 	return_(decoding);
 }
 
+//------------------
+char *base64Encode(const void *data, size_t len, char *encode_buffer, size_t encode_buffer_len)
+{
+	return genericBase64Encode(data, len, encode_buffer, encode_buffer_len, base64STDMap);
+}
+
+//------------------
+void *base64Decode(const char *data, size_t len, void *decode_buffer, size_t decode_buffer_len)
+{
+	return genericBase64Decode(data, len, decode_buffer, decode_buffer_len, base64STDMap);
+}
+
+//------------------
+char *base64UrlEncode(const void *data, size_t len, char *encode_buffer, size_t encode_buffer_len)
+{
+	return genericBase64Encode(data, len, encode_buffer, encode_buffer_len, base64URLMap);
+}
+
+//------------------
+void *base64UrlDecode(const char *data, size_t len, void *decode_buffer, size_t decode_buffer_len)
+{
+	return genericBase64Decode(data, len, decode_buffer, decode_buffer_len, base64URLMap);
+}
+
 #ifdef NOT_USED
 //-----------------------------------------------------
 void  hmac_md5(u_char *text, size_t text_len, u_char *key, size_t key_len, Md5Digest *digest)

=== modified file 'plugin/pbms/src/cslib/CSEncode.h'
--- plugin/pbms/src/cslib/CSEncode.h	2010-10-02 21:15:42 +0000
+++ plugin/pbms/src/cslib/CSEncode.h	2010-10-18 21:21:20 +0000
@@ -25,8 +25,14 @@
 #include "CSSha1.h"
 #include "CSMd5.h"
 
-char *base64Encode(const void *data, size_t len);
-void *base64Decode(const char *data, size_t len);
+// If the base64 buffer passed in is NULL then the function will allocate one the correct size for the data.
+// In this caser the caller must free the buffer using cs_free().
+// If a buffer is suppled but is too small en exception will be thrown.
+char *base64Encode(const void *data, size_t len, char *encode_buffer = NULL, size_t encode_buffer_len = 0);
+void *base64Decode(const char *data, size_t len, void *decode_buffer = NULL, size_t decode_buffer_len = 0);
+
+char *base64UrlEncode(const void *data, size_t len, char *encode_buffer = NULL, size_t encode_buffer_len = 0);
+void *base64UrlDecode(const char *data, size_t len, void *decode_buffer = NULL, size_t decode_buffer_len = 0);
 
 CSString *signature(const char *text, const char *key);
 CSString *urlEncode(CSString *src);

=== modified file 'plugin/pbms/src/cslib/CSException.cc'
--- plugin/pbms/src/cslib/CSException.cc	2010-10-02 21:15:42 +0000
+++ plugin/pbms/src/cslib/CSException.cc	2010-10-20 20:21:41 +0000
@@ -27,7 +27,12 @@
 
 #include "CSConfig.h"
 
+#ifdef OS_WINDOWS
+#define strsignal(s) NULL
+#else
 #include <sys/signal.h>
+#endif
+
 #include <limits.h>
 #include <string.h>
 
@@ -68,7 +73,9 @@
 	CSL.log(self, CSLog::Error, " ");
 	CSL.log(self, CSLog::Error, getMessage());
 	CSL.eol(self, CSLog::Error);
+#ifdef DUMP_STACK_TRACE
 	CSL.log(self, CSLog::Error, getStackTrace());
+#endif
 	CSL.unlock();
 }
 
@@ -81,10 +88,37 @@
 	CSL.log(self, CSLog::Error, " ");
 	CSL.log(self, CSLog::Error, getMessage());
 	CSL.eol(self, CSLog::Error);
+#ifdef DUMP_STACK_TRACE
 	CSL.log(self, CSLog::Error, getStackTrace());
+#endif
 	CSL.unlock();
 }
 
+void CSException::initException_va(const char *func, const char *file, int line, int err, const char *fmt, va_list ap)
+{
+
+	cs_format_context(CS_EXC_CONTEXT_SIZE, iContext, func, file, line);
+	iErrorCode = err;
+#ifdef OS_WINDOWS
+	vsprintf(iMessage, fmt, ap);
+#else
+	size_t len;
+	len = vsnprintf(iMessage, CS_EXC_MESSAGE_SIZE-1, fmt, ap);
+	if (len > CS_EXC_MESSAGE_SIZE-1)
+		len = CS_EXC_MESSAGE_SIZE-1;
+	iMessage[len] = 0;
+#endif
+}
+
+void CSException::initExceptionf(const char *func, const char *file, int line, int err, const char *fmt, ...)
+{
+	va_list ap;
+
+	va_start(ap, fmt);
+	initException_va(func, file, line, err, fmt, ap);
+	va_end(ap);
+}
+
 void CSException::initException(const char *func, const char *file, int line, int err, const char *message)
 {
 	cs_format_context(CS_EXC_CONTEXT_SIZE, iContext, func, file, line);
@@ -107,7 +141,8 @@
 {
 	cs_format_context(CS_EXC_CONTEXT_SIZE, iContext, func, file, line);
 	iErrorCode = CS_ERR_ASSERTION;
-	cs_strcpy(CS_EXC_MESSAGE_SIZE, iMessage, message);
+	cs_strcpy(CS_EXC_MESSAGE_SIZE, iMessage, "Assertion failed: ");
+	cs_strcat(CS_EXC_MESSAGE_SIZE, iMessage, message);
 }
 
 void CSException::getCoreError(uint32_t size, char *buffer, int err)
@@ -146,7 +181,7 @@
 	cs_format_context(CS_EXC_CONTEXT_SIZE, iContext, func, file, line);
 	iErrorCode = err;
 	getCoreError(CS_EXC_MESSAGE_SIZE, iMessage, err);
-	cs_replace_string(CS_EXC_MESSAGE_SIZE, iMessage, '%', item);
+	cs_replace_string(CS_EXC_MESSAGE_SIZE, iMessage, "%s", item);
 }
 
 void CSException::initOSError(const char *func, const char *file, int line, int err)
@@ -254,7 +289,7 @@
 	else {
 		CSException e;
 		
-		e.initException(func, file, line, err,message);
+		e.initException(func, file, line, err, message);
 		e.log(NULL, "*** Uncaught error");
 	}
 }
@@ -264,6 +299,27 @@
 	throwException(func, file, line, err, message, NULL);
 }
 
+void CSException::throwExceptionf(const char *func, const char *file, int line, int err, const char *fmt, ...)
+{
+	CSThread	*self;
+	va_list		ap;
+
+	va_start(ap, fmt);
+	if ((self = CSThread::getSelf())) {
+		self->myException.initException_va(func, file, line, err, fmt, ap);
+		va_end(ap);
+		self->myException.setStackTrace(self, NULL);
+		self->throwException();
+	}
+	else {
+		CSException e;
+		
+		e.initException_va(func, file, line, err, fmt, ap);
+		va_end(ap);
+		e.log(NULL, "*** Uncaught error");
+	}
+}
+
 void CSException::throwAssertion(const char *func, const char *file, int line, const char *message)
 {
 	CSThread *self;
@@ -271,7 +327,9 @@
 	if ((self = CSThread::getSelf())) {
 		self->myException.initAssertion(func, file, line, message);
 		self->myException.setStackTrace(self);
+		/* Not sure why we log the excpetion down here?!
 		self->logException();
+		*/
 		self->throwException();
 	}
 	else {
@@ -412,7 +470,7 @@
 void CSException::throwLastError(const char *func, const char *file, int line)
 {
 #ifdef OS_WINDOWS
-	throwOSError(func, file, line, (int) getLastError());
+	throwOSError(func, file, line, (int) GetLastError());
 #else
 	throwOSError(func, file, line, (int) errno);
 #endif

=== modified file 'plugin/pbms/src/cslib/CSException.h'
--- plugin/pbms/src/cslib/CSException.h	2010-10-02 21:15:42 +0000
+++ plugin/pbms/src/cslib/CSException.h	2010-10-18 21:21:20 +0000
@@ -30,13 +30,12 @@
 
 #include <errno.h>
 #include <limits.h>
+#include <stdarg.h>
 
 #include "CSDefs.h"
 #include "CSString.h"
 #include "CSObject.h"
 
-using namespace std;
-
 #define CS_ERR_ASSERTION						-14000
 #define CS_ERR_EOF								-14001
 #define CS_ERR_JUMP_OVERFLOW					-14002
@@ -56,6 +55,8 @@
 #define CS_ERR_CHECKSUM_ERROR					-14016
 #define CS_ERR_MISSING_HTTP_HEADER				-14017
 #define CS_ERR_OPERATION_NOT_SUPPORTED			-14018
+#define CS_ERR_BODY_INCOMPLETE					-14019
+#define CS_ERR_RECEIVE_TIMEOUT					-14020
 
 #define CS_EXC_CONTEXT_SIZE						300
 #define CS_EXC_MESSAGE_SIZE						(PATH_MAX + 300)
@@ -82,7 +83,9 @@
 	void log(CSThread *self);
 	void log(CSThread *self, const char *message);
 
+	void initException_va(const char *func, const char *file, int line, int err, const char *fmt, va_list ap);
 	void initException(CSException &exception);
+	void initExceptionf(const char *func, const char *file, int line, int err, const char *fmt, ...);
 	void initException(const char *func, const char *file, int line, int err, const char *message);
 	void initAssertion(const char *func, const char *file, int line, const char *message);
 	void getCoreError(uint32_t size, char *buffer, int err);
@@ -98,6 +101,7 @@
 
 	static void throwException(const char *func, const char *file, int line, int err, const char *message, const char *stack);
 	static void throwException(const char *func, const char *file, int line, int err, const char *message);
+	static void throwExceptionf(const char *func, const char *file, int line, int err, const char *fmt, ...);
 	static void throwAssertion(const char *func, const char *file, int line, const char *message);
 	static void throwCoreError(const char *func, const char *file, int line, int err);
 	static void throwCoreError(const char *func, const char *file, int line, int err, const char *item);

=== modified file 'plugin/pbms/src/cslib/CSFile.cc'
--- plugin/pbms/src/cslib/CSFile.cc	2010-10-02 21:15:42 +0000
+++ plugin/pbms/src/cslib/CSFile.cc	2010-10-20 20:21:41 +0000
@@ -28,30 +28,25 @@
 
 #include "CSConfig.h"
 
-#ifndef XT_WIN
+#ifdef OS_WINDOWS
+#include <sys/utime.h>
+#define utimes(f, s) _utime(f, s)
+#else
 #include <unistd.h>
 #include <dirent.h>
+#include <sys/time.h>
 #endif
 #include <stdio.h>
-#include <sys/stat.h>
-#include <fcntl.h>
+
 #include <errno.h>
 #include <string.h>
 
+#include "CSGlobal.h"
 #include "CSFile.h"
 #include "CSStream.h"
-#include "CSGlobal.h"
-
-
-static int unix_file_open(const char *path, int flags, mode_t mode)
-{
-	return open(path, flags, mode);
-}
-
-int unix_file_close(int fh)
-{
-	return close(fh);
-}
+#include "CSMd5.h"
+
+#define IS_MODE(m, s) ((m & s) == s)
 
 /*
  * ---------------------------------------------------------------
@@ -85,52 +80,70 @@
 	return CSFileInputStream::newStream(RETAIN(this), offset);
 }
 
+bool CSFile::try_CreateAndOpen(CSThread *self, int mode, bool retry)
+{
+	volatile bool rtc = true;
+	
+	try_(a) {
+		openFile(mode);
+		rtc = false; // success, do not try again.
+	}
+	catch_(a) {
+		if (retry || !isDirNotFound(&self->myException))
+			throw_();
+
+		/* Make sure the parent directory exists: */
+		CSPath	*dir = CSPath::newPath(RETAIN(myFilePath), "..");
+		push_(dir);
+		try_(b) {
+			dir->makePath();
+		}
+		catch_(b) { /* Two threads may try to create the directory at the same time. */
+			if (!isDirExists(&self->myException))
+				throw_();
+		}
+		cont_(b);
+
+		release_(dir);
+	}
+	cont_(a);
+	return rtc; // try again.
+}
+
 void CSFile::open(int mode)
 {
-	CSPath	*dir;
-	bool	retry = false;
-
-	CLOBBER_PROTECT(retry);
-
 	enter_();
 	if (mode & CREATE) {
-		/* Create and open the file: */
-		do {
-			try_(a) {
-				openFile(mode);
-				retry = false;
-			}
-			catch_(a) {
-				if (retry || !isDirNotFound(&self->myException))
-					throw_();
-
-				/* Make sure the parent directory exists: */
-				dir = CSPath::newPath(RETAIN(myFilePath), "..");
-				push_(dir);
-				try_(b) {
-					dir->makePath();
-				}
-				catch_(b) { /* Two threads may try to create the directory at the same time. */
-					if (!isDirExists(&self->myException))
-						throw_();
-				}
-				cont_(b);
-
-				release_(dir);
-				retry = true;
-			}
-			cont_(a);
-		}
-		while (retry);
+		bool retry = false;
+		while ((retry = try_CreateAndOpen(self, mode, retry)) == true){}		
 	}
 	else
 		openFile(mode);
 	exit_();
 }
 
+void CSFile::lock()
+{
+	if (!iLocked) {
+		sf_lock(IS_MODE(iMode, READONLY));
+	}
+	iLocked++;
+}
+
+void CSFile::unlock()
+{
+	iLocked--;
+	if (!iLocked) 
+		sf_unlock();
+}
+
 bool CSFile::transfer(CSFile *dst_file, off64_t dst_offset, CSFile *src_file, off64_t src_offset, off64_t size, char *buffer, size_t buffer_size)
 {
 	size_t tfer;
+	enter_();
+	
+	push_(dst_file);
+	push_(src_file);
 
 	while (size > 0) {
 		if (size > (off64_t) buffer_size)
@@ -138,18 +151,24 @@
 		else
 			tfer = (size_t) size;
 		if (!(tfer = src_file->read(buffer, src_offset, tfer, 0)))
-			return false;
+			break;
 		dst_file->write(buffer, dst_offset, tfer);
 		dst_offset += tfer;
 		src_offset += tfer;
 		size -= tfer;
 	}
-	return true;
+	
+	release_(src_file);
+	release_(dst_file);
+	return_(size == 0);
 }
 
 void CSFile::streamOut(CSOutputStream *dst_stream, off64_t src_offset, off64_t size, char *buffer, size_t buffer_size)
 {
 	size_t tfer;
+	enter_();
+	
+	push_(dst_stream);
 
 	while (size > 0) {
 		if (size > (off64_t) buffer_size)
@@ -163,57 +182,55 @@
 		src_offset += tfer;
 		size -= tfer;
 	}
+	
+	release_(dst_stream);
+	exit_();
 }
 
 #define CS_MASK				((S_IRUSR | S_IWUSR) | (S_IRGRP | S_IWGRP) | (S_IROTH))
 
+void CSFile::touch()
+{
+	// don't use futimes() here. On some platforms it will fail if the 
+	// file was opened readonly.
+	if (utimes(myFilePath->getCString(), NULL) == -1)
+		CSException::throwFileError(CS_CONTEXT, myFilePath->getCString(), errno);
+}
+
 void CSFile::close()
 {
-	if (iFH != -1) {
-		unix_file_close(iFH);
-		iFH = -1;
-	}
+	while (iLocked)
+		unlock();
+	sf_close();
 }
 
 off64_t CSFile::getEOF()
 {
-	off64_t eof;
-
-	if ((eof = lseek(iFH, 0, SEEK_END)) == (off64_t) -1)
-		CSException::throwFileError(CS_CONTEXT, myFilePath->getCString(), errno);
-
-     return eof;
+     return sf_getEOF();
 }
 
 void CSFile::setEOF(off64_t offset)
 {
-	if (ftruncate(iFH, offset) == -1)
-		CSException::throwFileError(CS_CONTEXT, myFilePath->getCString(), errno);
+	sf_setEOF(offset);
 }
 
 size_t CSFile::read(void *data, off64_t offset, size_t size, size_t min_size)
 {
-	ssize_t read_size;
+	size_t read_size;
 	
 	enter_();
-	read_size = pread(iFH, data, size, offset);
+	read_size = sf_pread(data, size, offset);
 	self->interrupted();
-	if (read_size ==  -1)
-		CSException::throwFileError(CS_CONTEXT, myFilePath->getCString(), errno);
-	if ((size_t) read_size < min_size)
+	if (read_size < min_size)
 		CSException::throwEOFError(CS_CONTEXT, myFilePath->getCString());
 	return_(read_size);
 }
 
 void CSFile::write(const void *data, off64_t offset, size_t size)
 {
-	size_t write_size;
-
 	enter_();
-    write_size = pwrite(iFH, (void *) data, size, offset);
+    sf_pwrite(data, size, offset);
 	self->interrupted();
-	if (write_size != size)
-		CSException::throwFileError(CS_CONTEXT, myFilePath->getCString(), errno);
 	exit_();
 }
 
@@ -223,7 +240,7 @@
 
 void CSFile::sync()
 {
-	fsync(iFH);
+	sf_sync();
 }
 
 CSFile *CSFile::newFile(CSPath *path)
@@ -246,26 +263,49 @@
 	return newFile(path);
 }
 
+CSFile *CSFile::newFile(const char *dir_str, const char *path_str)
+{
+	CSPath *path;
+
+	path = CSPath::newPath(dir_str, path_str);
+	return newFile(path);
+}
+
 void CSFile::openFile(int mode)
 {
-	int flags = 0;
-
-	if (mode & READONLY)
-		flags = O_RDONLY;
-	else
-		flags = O_RDWR;
-
-	if (mode & CREATE)
-		flags |= O_CREAT;
+	if (fs_isOpen() && (iMode != mode))
+		close();
+		
+	if (!fs_isOpen())
+		sf_open(myFilePath->getCString(), IS_MODE(mode, READONLY), IS_MODE(mode, CREATE));
 	
-	if ((iFH = unix_file_open(myFilePath->getCString(), flags, CS_MASK)) ==  -1)
-		CSException::throwFileError(CS_CONTEXT, myFilePath->getCString(), errno);
-
+	iMode = mode;
 	/* Does not make sense to truncate, and have READONLY! */
-	if ((mode & TRUNCATE) && !(mode & READONLY))
+	if (IS_MODE(mode, TRUNCATE) && !IS_MODE(mode, READONLY))
 		setEOF((off64_t) 0);
 }
 
+void CSFile::md5Digest(Md5Digest *digest)
+{
+	u_char buffer[1024];
+	off64_t offset = 0, size;
+	size_t len;
+	CSMd5 md5;
+	enter_();
+	
+	size = getEOF();
+	while (size) {
+		len = (size_t)((size < 1024)? size:1024);
+		len = read(buffer, offset, len, len);
+		offset +=len;
+		size -= len;
+		md5.md5_append(buffer, len);
+	}
+	md5.md5_get_digest(digest);
+	exit_();
+	
+}
+
 /*
  * ---------------------------------------------------------------
  * A READ BUFFERED FILE
@@ -312,7 +352,7 @@
 		iBufferDataLen = 0;
 	}
 	else if (offset < iFileBufferOffset + iBufferDataLen)
-		iBufferDataLen = offset - iFileBufferOffset;
+		iBufferDataLen = (size_t)(offset - iFileBufferOffset);
 }
 
 size_t CSReadBufferedFile::read(void *data, off64_t offset, size_t size, size_t min_size)
@@ -330,7 +370,7 @@
 			}
 			if (offset + size > iFileBufferOffset) {
 				// 2
-				tfer = offset + size - iFileBufferOffset;
+				tfer = (size_t)(offset + size - iFileBufferOffset);
 				memcpy((char *) data + (iFileBufferOffset - offset), iFileBuffer, tfer);
 				size -= tfer;
 			}
@@ -361,7 +401,7 @@
 		}
 		if (offset < iFileBufferOffset + iBufferDataLen) {
 			// 4 We assume we are reading front to back
-			tfer = iFileBufferOffset + iBufferDataLen - offset;
+			tfer = (size_t)(iFileBufferOffset + iBufferDataLen - offset);
 			memcpy(data, iFileBuffer + (offset - iFileBufferOffset), tfer);
 			data = (char *) data + tfer;
 			size -= tfer;
@@ -401,7 +441,7 @@
 			}
 			else if (offset + size > iFileBufferOffset) {
 				// 2
-				tfer = offset + size - iFileBufferOffset;
+				tfer = (size_t)(offset + size - iFileBufferOffset);
 				memcpy(iFileBuffer, (char *) data + (iFileBufferOffset - offset), tfer);
 			}
 		}
@@ -411,7 +451,7 @@
 		}
 		else if (offset < iFileBufferOffset + iBufferDataLen) {
 			// 4 We assume we are reading front to back
-			tfer = iFileBufferOffset + iBufferDataLen - offset;
+			tfer = (size_t)(iFileBufferOffset + iBufferDataLen - offset);
 			memcpy(iFileBuffer + (offset - iFileBufferOffset), data, tfer);
 		}
 		// else 5
@@ -441,60 +481,6 @@
 	myFile->openFile(mode);
 }
 
-CSFile *CSReadBufferedFile::newFile(CSFile *file)
-{
-	CSReadBufferedFile *f;
-
-	if (!(f = new CSReadBufferedFile())) {
-		file->release();
-		CSException::throwOSError(CS_CONTEXT, ENOMEM);
-	}
-	f->myFile = file;
-	f->myFilePath = file->myFilePath;
-	f->myFilePath->retain();
-	return (CSFile *) f;
-}
-
-/*
- * ---------------------------------------------------------------
- * A BUFFERED FILE
- */
-
-void CSBufferedFile::write(const void *data, off64_t offset, size_t size)
-{
-	if (iBufferDataLen > 0) {
-		if (offset < iFileBufferOffset && offset <= iFileBufferOffset + iBufferDataLen) {
-			size_t tfer;
-
-			tfer = iFileBufferOffset + SC_DEFAULT_FILE_BUFFER_SIZE - offset;
-			if (tfer >= size) {
-				memcpy(iFileBuffer + (offset - iFileBufferOffset), data, size);
-			}
-			size -= tfer;
-			data = (char *) data + tfer;
-		}
-		flush();
-	}
-
-	if (size < SC_DEFAULT_FILE_BUFFER_SIZE) {
-		iFileBufferOffset = offset;
-		iBufferDataLen = size;
-		memcpy(iFileBuffer, data, size);
-		iBufferDirty = true;
-	}
-	else
-		myFile->write(data, offset, size);
-}
-
-void CSBufferedFile::flush()
-{
-	if (iBufferDirty && iBufferDataLen) {
-		myFile->write(iFileBuffer, iFileBufferOffset, iBufferDataLen);
-		iBufferDirty = false;
-		
-	}
-	myFile->flush();
-}
 
 /*
  * ---------------------------------------------------------------

=== modified file 'plugin/pbms/src/cslib/CSFile.h'
--- plugin/pbms/src/cslib/CSFile.h	2010-10-02 21:15:42 +0000
+++ plugin/pbms/src/cslib/CSFile.h	2010-10-18 21:21:20 +0000
@@ -34,15 +34,12 @@
 #include "CSDefs.h"
 #include "CSPath.h"
 #include "CSException.h"
-
-using namespace std;
-
-extern int unix_file_close(int fh);
+#include "CSSys.h"
 
 class CSOutputStream;
 class CSInputStream;
 
-class CSFile : public CSRefObject {
+class CSFile : public CSSysFile, public CSRefObject {
 public:
 	CSPath *myFilePath;
 
@@ -51,7 +48,7 @@
 	static const int CREATE = 2;	// Create if it does not exist
 	static const int TRUNCATE = 4;	// After open, set EOF to zero	
 
-	CSFile(): myFilePath(NULL), iFH(-1) { }
+	CSFile(): myFilePath(NULL), iMode(-1), iLocked(0) { }
 
 	virtual ~CSFile(); 
 
@@ -67,12 +64,24 @@
 	 */
 	virtual void open(int mode);
 
+	/* Lock the file. The file will be unlocked
+	 * when closed.
+	 */
+	virtual void lock();
+
+	virtual void unlock();
+
 	/*
 	 * Close the file.
 	 */
 	virtual void close();
 
 	/*
+	 * Calculate the Md5 digest for the file.
+	 */
+	void md5Digest(Md5Digest *digest);
+
+	/*
 	 * Move the current position to
 	 * the end of the file.
 	 */
@@ -101,6 +110,9 @@
 	/* Flush the OS buffers: */
 	virtual void sync() ;
 
+	/* Resets access and modification times of the file: */
+	virtual void touch();
+
 	/*
 	 * Return a platform specific prefered 
 	 * line ending for text files.
@@ -109,16 +121,20 @@
 
 	virtual const char *getPathString() { return myFilePath->getCString(); }
 
+	bool exists() { return myFilePath->exists(); }
+
 	friend class CSReadBufferedFile;
-	friend class CSBufferedFile;
 
 private:
-	int		iFH;
+	int		iMode;
+	int		iLocked;
 
 	virtual void openFile(int mode);
+	bool try_CreateAndOpen(CSThread *self, int mode, bool retry);
 
 public:
 	void streamOut(CSOutputStream *dst_stream, off64_t src_offset, off64_t size, char *buffer, size_t buffer_size);
+	void streamIn(CSInputStream *src_stream, off64_t dst_offset, off64_t size, char *buffer, size_t buffer_size);
 
 	static bool isDirNotFound(CSException *e) { return e->getErrorCode() == ENOENT; }
 	static bool isDirExists(CSException *e) { return e->getErrorCode() == EEXIST; }
@@ -128,62 +144,56 @@
 	static CSFile *newFile(CSPath *path);
 
 	static CSFile *newFile(const char *path);
+
+	static CSFile *newFile(const char *dir_str, const char *path_str);
 };
 
+
+// This stuff needs to be retought.
+
 #ifdef DEBUG
 #define SC_DEFAULT_FILE_BUFFER_SIZE			127
 #else
 #define SC_DEFAULT_FILE_BUFFER_SIZE			(64 * 1024)
 #endif
 
-class CSReadBufferedFile : public CSFile {
+class CSReadBufferedFile : public CSRefObject {
 public:
+
+	CSReadBufferedFile();
+
+	~CSReadBufferedFile();
+	
+	void setFile(CSFile	*file) {myFile = file;}
+
+	const char *getPathString() { return myFile->getPathString(); }
+	void open(int mode) {myFile->open(mode); }
+
+	void close();
+
+	off64_t getEOF();
+
+	void setEOF(off64_t offset);
+
+	size_t read(void *data, off64_t offset, size_t size, size_t min_size);
+
+	void write(const void *data, off64_t offset, size_t size);
+
+	void flush();
+
+	void sync();
+
+	const char *getEOL();
+
+private:
 	CSFile	*myFile;
 
-	CSReadBufferedFile();
-
-	virtual ~CSReadBufferedFile();
-
-	virtual void close();
-
-	virtual off64_t getEOF();
-
-	virtual void setEOF(off64_t offset);
-
-	virtual size_t read(void *data, off64_t offset, size_t size, size_t min_size);
-
-	virtual void write(const void *data, off64_t offset, size_t size);
-
-	virtual void flush();
-
-	virtual void sync();
-
-	virtual const char *getEOL();
-
-	friend class CSBufferedFile;
-
-private:
 	char	iFileBuffer[SC_DEFAULT_FILE_BUFFER_SIZE];
 	off64_t	iFileBufferOffset;
 	size_t	iBufferDataLen;
 
 	virtual void openFile(int mode);
-public:
-	static CSFile *newFile(CSFile *file);
-};
-
-class CSBufferedFile : public CSReadBufferedFile {
-public:
-	CSBufferedFile(): CSReadBufferedFile(), iBufferDirty(false) { }
-
-	virtual ~CSBufferedFile() { };
-
-	virtual void write(const void *data, off64_t offset, size_t size);
-
-	virtual void flush();
-
-private:
-	bool	iBufferDirty;
-};
+};
+
 
 #endif

=== modified file 'plugin/pbms/src/cslib/CSGlobal.h'
--- plugin/pbms/src/cslib/CSGlobal.h	2010-10-02 21:15:42 +0000
+++ plugin/pbms/src/cslib/CSGlobal.h	2010-10-20 20:21:41 +0000
@@ -31,20 +31,9 @@
 #define __CSGLOBAL_H__
 
 #include "CSDefs.h"
+#include "CSObject.h"
 #include "CSThread.h"
 
-/* Those compilers that support the function
- * macro (in particular the "pretty" function
- * macro must be defined here.
- */
-#ifdef OS_WINDOWS
-#define __FUNC__				__FUNCTION__
-#elif defined(OS_SOLARIS)
-#define __FUNC__				"__func__"
-#else
-#define __FUNC__				__PRETTY_FUNCTION__
-#endif
-
 
 /* This is the call context: */
 #define CS_CONTEXT			__FUNC__, __FILE__, __LINE__
@@ -60,6 +49,11 @@
 #define ASSERT(x)
 #endif
 
+#ifdef DEBUG
+#define retain()			retain(__FUNC__, __FILE__, __LINE__)
+#define release()			release(__FUNC__, __FILE__, __LINE__)
+#endif
+
 #define new_(v, t)			do { v = new t; if (!v) CSException::throwOSError(CS_CONTEXT, ENOMEM); } while (0)
 
 /*
@@ -74,7 +68,14 @@
  *
  * It also sets up the current thread pointer 'self'.
  */
+#ifdef DEBUG
+#define STACK_CHECK    CSReleasePtr self_reltop = self->relTop
+#else
+#define STACK_CHECK   
+#endif
+
 #define inner_()			int			cs_frame = self->callTop++; \
+							STACK_CHECK ; \
 							do { \
 								if (cs_frame< CS_CALL_STACK_SIZE) { \
 									self->callStack[cs_frame].cs_func = __FUNC__; \
@@ -83,7 +84,8 @@
 								} \
 							} while (0)
 
-#define outer_()			self->callTop = cs_frame;
+#define outer_()			self->callTop = cs_frame; \
+							ASSERT(self->relTop == self_reltop); 
 
 #define enter_()			CSThread	*self = CSThread::getSelf(); \
 							inner_()
@@ -109,21 +111,6 @@
  * Throwing and catching (the jump stack)
  */
 
-/*
- * Quote from the C99 spec on setjmp:
- * All accessible objects have values, and all other components of the abstract 
- * machine have state, as of the time the longjmp function was called, except that 
- * the values of objects of automatic storage duration that are local to the function 
- * containing the invocation of the corresponding setjmp macro that do not have 
- * volatile-qualified type and have been changed between the setjmp invocation and 
- * longjmp call are indeterminate. 
- *
- * GCC will not put variables into registers for which the address gets taken do use
- * CLOBBER_PROTECT() to protect variables from being clobbered.
- * This came from Jens Thoms Toerring. Thanks!
- */
-#define CLOBBER_PROTECT(a) do { if ((void *)&(a) == (void*) 1) (a) = (a);} while(0)
-
 int prof_setjmp(void);
 
 #define TX_CHK_JMP()		if ((self)->jumpDepth < 0 || (self)->jumpDepth >= CS_JUMP_STACK_SIZE) CSException::throwCoreError(__FUNC__, __FILE__, __LINE__, CS_ERR_JUMP_OVERFLOW)
@@ -168,6 +155,15 @@
 								(self)->relTop++; \
 							} while (0)
 
+#define push_ref_(r)			do { \
+								if ((self)->relTop >= (self)->relStack + CS_RELEASE_STACK_SIZE) {\
+									CSException::throwCoreError(CS_CONTEXT, CS_ERR_RELEASE_OVERFLOW); \
+								} \
+								(self)->relTop->r_type = CS_RELEASE_OBJECT_PTR; \
+								(self)->relTop->x.r_objectPtr = (CSObject **)&(r); \
+								(self)->relTop++; \
+							} while (0)
+
 #define pop_(r)				do { \
 								ASSERT((self)->relTop > (self)->relStack); \
 								if (((self)->relTop - 1)->r_type == CS_RELEASE_OBJECT) {\
@@ -177,7 +173,9 @@
 								} else if (((self)->relTop - 1)->r_type == CS_RELEASE_POOLED) {\
 									ASSERT(((self)->relTop - 1)->x.r_pooled == ((CSPooled *) r)); \
 								} else if (((self)->relTop - 1)->r_type == CS_RELEASE_MEM) {\
-									ASSERT(((self)->relTop - 1)->x.r_mem == ((CSPooled *) r)); \
+									ASSERT(((self)->relTop - 1)->x.r_mem == ((void *) r)); \
+								}  else if (((self)->relTop - 1)->r_type == CS_RELEASE_OBJECT_PTR) {\
+									ASSERT(((self)->relTop - 1)->x.r_objectPtr == ((CSObject **) &(r))); \
 								}  else {\
 									ASSERT(false); \
 								} \
@@ -203,7 +201,13 @@
 									ASSERT(mem == (void *)(r)); \
 									(self)->relTop--; \
 									cs_free(mem); \
-								}  else {\
+								}  else if (((self)->relTop - 1)->r_type == CS_RELEASE_OBJECT_PTR) {\
+									register CSObject **rp; \
+									rp = ((self)->relTop - 1)->x.r_objectPtr; \
+									ASSERT(rp == (CSObject **)&(r)); \
+									(self)->relTop--; \
+									if(*rp) (*rp)->release(); \
+								} else {\
 									ASSERT(false); \
 								} \
 							} while (0)
@@ -218,6 +222,14 @@
 								(self)->relTop++; \
 							} while (0)
 
+#define locked_(r)			do { \
+								if ((self)->relTop >= (self)->relStack + CS_RELEASE_STACK_SIZE) \
+									CSException::throwCoreError(CS_CONTEXT, CS_ERR_RELEASE_OVERFLOW); \
+								(self)->relTop->r_type = CS_RELEASE_MUTEX; \
+								(self)->relTop->x.r_mutex = (r); \
+								(self)->relTop++; \
+							} while (0)
+
 #define unlock_(r)			do {  \
 								register CSMutex *rp; \
 								ASSERT((self)->relTop > (self)->relStack); \

=== modified file 'plugin/pbms/src/cslib/CSHTTPStream.cc'
--- plugin/pbms/src/cslib/CSHTTPStream.cc	2010-10-02 21:15:42 +0000
+++ plugin/pbms/src/cslib/CSHTTPStream.cc	2010-10-18 21:21:20 +0000
@@ -27,12 +27,12 @@
 #include "CSConfig.h"
 
 #include <stdlib.h>
+#include <string.h>
 #include <inttypes.h>
 
-#include "string.h"
-
+#include "CSGlobal.h"
 #include "CSHTTPStream.h"
-#include "CSGlobal.h"
+#include "CSLog.h"
 
 #ifdef DEBUG
 //#define PRINT_HEADER
@@ -85,11 +85,15 @@
 	iValue = value;
 }
 
-void CSHeader::write(CSOutputStream *out)
+void CSHeader::write(CSOutputStream *out, bool trace)
 {
+	if (trace)
+		printf("%s: %s\n", iName->getCString(), iValue->getCString());
+
 	out->print(iName);
 	out->print(": ");
-	out->print(iValue);
+	if (iValue)
+		out->print(iValue);
 	out->print("\r\n");
 }
 
@@ -218,6 +222,14 @@
 	exit_();
 }
 
+void CSHTTPHeaders::addHeader(const char *name, uint64_t value)
+{
+	char buffer[30];
+
+	snprintf(buffer, 30, "%"PRIu64, value);
+	addHeader(name, buffer);
+}
+
 void CSHTTPHeaders::removeHeader(CSString *name)
 {
 	enter_();
@@ -245,37 +257,47 @@
 
 CSString *CSHTTPHeaders::getHeaderValue(const char *name)
 {
-	CSString *n;
 	CSString *v;
 
-	enter_();
-	n = CSString::newString(name);
-	push_(n);
 	v = NULL;
 	if (iHeaders) {
 		CSHeader *h;
 
 		for (uint32_t i=0; i<iHeaders->size(); i++) {
 			h = (CSHeader *) iHeaders->get(i);
-			if (h->getName()->compare(n) == 0) {
+			if (h->getName()->compare(name) == 0) {
 				v = h->getValue();
 				v->retain();
 				break;
 			}
 		}
 	}
-	release_(n);
-	return_(v);
-}
-
-void CSHTTPHeaders::writeHeader(CSOutputStream *out)
-{
-	if (iHeaders) {
-		CSHeader *h;
-
-		for (uint32_t i=0; i<iHeaders->size(); i++) {
-			h = (CSHeader *) iHeaders->get(i);
-			h->write(out);
+	return v;
+}
+
+const char *CSHTTPHeaders::getHeaderCStringValue(const char *name)
+{
+	if (iHeaders) {
+		CSHeader *h;
+
+		for (uint32_t i=0; i<iHeaders->size(); i++) {
+			h = (CSHeader *) iHeaders->get(i);
+			if (h->getName()->compare(name) == 0) {
+				return h->getValue()->getCString();
+			}
+		}
+	}
+	return NULL;
+}
+
+void CSHTTPHeaders::writeHeader(CSOutputStream *out, bool trace)
+{
+	if (iHeaders) {
+		CSHeader *h;
+
+		for (uint32_t i=0; i<iHeaders->size(); i++) {
+			h = (CSHeader *) iHeaders->get(i);
+			h->write(out, trace);
 		}
 	}
 }
@@ -318,7 +340,7 @@
 		iInput->release();
 }
 
-void CSHTTPInputStream::readHead()
+void CSHTTPInputStream::readHead(bool trace)
 {
 	CSStringBuffer	*sb = NULL;
 	bool			first_line = true;
@@ -330,9 +352,11 @@
 		sb = iInput->readLine();
 		if (!sb)
 			break;
-#ifdef PRINT_HEADER
-		printf("HTTP: %s\n", sb->getCString());
-#endif
+		if (trace) {
+			if (first_line)
+				CSL.log(self, CSLog::Protocol, "HTTP Request - Header:\n");
+			printf("%s\n", sb->getCString());
+		}
 		if (sb->length() == 0) {
 			sb->release();
 			break;
@@ -401,6 +425,40 @@
 	exit_();
 }
 
+void CSHTTPInputStream::readBody()
+{
+	uint64_t	body_size;
+	size_t	tfer, len;
+
+	if (getContentLength(&body_size)) {
+		iBody.setLength((size_t) body_size);
+		len = 0;
+		while (len < body_size) {
+			tfer =  read(iBody.getBuffer(len), (size_t)(body_size - len));
+			if (!tfer)
+				CSException::throwException(CS_CONTEXT, CS_ERR_BODY_INCOMPLETE, "POST data incomplete");
+			len += tfer;
+		}
+	}
+	else {
+		CSStringBuffer *sb = NULL;
+
+		/* Read until we have an empty line. */
+		for (;;) {
+			sb = readLine();
+			if (!sb)
+				break;
+			if (sb->length() == 0) {
+				sb->release();
+				break;
+			}
+			iBody.append(sb->getBuffer(0), sb->length());
+			iBody.append((char) '\n');
+			sb->release();
+		}
+	}
+}
+
 bool CSHTTPInputStream::getRange(uint64_t *size, uint64_t *offset)
 {
 	CSString	*val;
@@ -423,7 +481,7 @@
 	return haveRange;
 }
 
-uint64_t CSHTTPInputStream::getContentLength()
+bool CSHTTPInputStream::getContentLength(uint64_t *length)
 {
 	CSString	*val;
 	uint64_t		size = 0;
@@ -434,8 +492,10 @@
 		if (len)  
 			sscanf(len, "%"PRIu64"", &size);		
 		val->release();
+		*length = size;
+		return true;
 	}
-	return size;
+	return false;
 }
 
 const char *CSHTTPInputStream::getMethod()
@@ -531,33 +591,60 @@
 		iOutput->release();
 }
 
-void CSHTTPOutputStream::writeHeaders()
-{
-	writeHeader(this);
+void CSHTTPOutputStream::print(const char *str, bool trace)
+{
+	if (trace)
+		printf("%s", str);
+	iOutput->print(str);
+}
+
+
+void CSHTTPOutputStream::print(int32_t value, bool trace)
+{
+	if (trace)
+		printf("%d", value);
+	iOutput->print(value);
+}
+
+void CSHTTPOutputStream::print(uint64_t value, bool trace)
+{
+	if (trace)
+		printf("%"PRIu64, value);
+	iOutput->print(value);
+}
+
+void CSHTTPOutputStream::writeHeaders(bool trace)
+{
+	writeHeader(this, trace);
 	clearHeaders();
 }
 
-void CSHTTPOutputStream::writeHead()
+void CSHTTPOutputStream::writeHead(bool trace)
 {
-	iOutput->print("HTTP/1.1 ");
-	iOutput->print(iStatus);
-	iOutput->print(" ");
-	iOutput->print(getReasonPhrase(iStatus));
-	iOutput->print("\r\n");
-	writeHeader(iOutput);
-	iOutput->print("Content-Length: ");
-	iOutput->print(iContentLength);
-	iOutput->print("\r\n");
+	enter_();
+	if (trace)
+		CSL.log(self, CSLog::Protocol, "HTTP Reply - Header:\n");
+	print("HTTP/1.1 ", trace);
+	print(iStatus, trace);
+	print(" ", trace);
+	print(getReasonPhrase(iStatus), trace);
+	print("\r\n", trace);
+	writeHeader(iOutput, trace);
+	print("Content-Length: ", trace);
+	print(iContentLength, trace);
+	print("\r\n", trace);
 	if (iRangeSize && (iStatus == 200)) {
-		iOutput->print("Content-Range: bytes ");
-		iOutput->print(iRangeOffset);
-		iOutput->print("-");
-		iOutput->print(iRangeOffset + iRangeSize -1);
-		iOutput->print("/");
-		iOutput->print(iTotalLength);
-		iOutput->print("\r\n");
+		print("Content-Range: bytes ", trace);
+		print(iRangeOffset, trace);
+		print("-", trace);
+		print(iRangeOffset + iRangeSize -1, trace);
+		print("/", trace);
+		print(iTotalLength, trace);
+		print("\r\n", trace);
 	}
-	iOutput->print("\r\n");
+	print("\r\n", trace);
+	exit_();
+
 }
 
 void CSHTTPOutputStream::clearBody()
@@ -580,10 +667,38 @@
 	iContentLength = iBody.length();
 }
 
-void CSHTTPOutputStream::appendBody(int value)
-{
-	iBody.append(value);
-	iContentLength = iBody.length();
+void CSHTTPOutputStream::appendBody(int32_t value)
+{
+	iBody.append(value);
+	iContentLength = iBody.length();
+}
+
+void CSHTTPOutputStream::appendBody(uint32_t value)
+{
+	iBody.append(value);
+	iContentLength = iBody.length();
+}
+
+void CSHTTPOutputStream::appendBody(uint64_t value)
+{
+	iBody.append(value);
+	iContentLength = iBody.length();
+}
+
+const char *CSHTTPOutputStream::getBodyData()
+{
+	return iBody.getCString(); 
+}
+
+size_t CSHTTPOutputStream::getBodyLength()
+{
+	return iBody.length();
+}
+
+void CSHTTPOutputStream::setBody(CSStringBufferImpl *buf)
+{
+	iBody.take(buf);
+	iContentLength = iBody.length(); 
 }
 
 void CSHTTPOutputStream::close()

=== modified file 'plugin/pbms/src/cslib/CSHTTPStream.h'
--- plugin/pbms/src/cslib/CSHTTPStream.h	2010-10-02 21:15:42 +0000
+++ plugin/pbms/src/cslib/CSHTTPStream.h	2010-10-18 21:21:20 +0000
@@ -53,7 +53,7 @@
 	const char *getNameCString() { return iName ? iName->getCString() : ""; }
 	const char *getValueCString() { return iValue ? iValue->getCString() : ""; }
 
-	void write(CSOutputStream *out);
+	void write(CSOutputStream *out, bool trace);
 	
 	friend class CSHTTPHeaders;
 private:
@@ -66,7 +66,7 @@
 
 class CSHTTPHeaders {
 public:
-	CSHTTPHeaders(): iHeaders(NULL), iKeepAlive(false), iExpect100Continue(false), iUnknownEpectHeader(false) { }
+	CSHTTPHeaders(CSVector *list = NULL): iHeaders(list), iKeepAlive(false), iExpect100Continue(false), iUnknownEpectHeader(false) { }
 	virtual ~CSHTTPHeaders() { if (iHeaders) iHeaders->release();}
 
 	void clearHeaders();
@@ -78,10 +78,12 @@
 	void addHeader(const char *name, uint32_t nlen, const char *value, uint32_t vlen);
 	void addHeader(CSString *name, CSString *value);
 	void addHeader(const char *name, CSString *value);
+	void addHeader(const char *name, uint64_t value);
 	void removeHeader(const char *name);
 	void removeHeader(CSString *name);
 	CSString *getHeaderValue(const char *name);
-	void writeHeader(CSOutputStream *out);
+	const char *getHeaderCStringValue(const char *name);
+	void writeHeader(CSOutputStream *out, bool trace);
 	bool keepAlive();
 	bool expect100Continue();
 	bool unknownEpectHeader();
@@ -106,14 +108,24 @@
 	bool iUnknownEpectHeader;
 };
 
+class CSRefHTTPHeaders : public CSHTTPHeaders, public CSRefObject {
+public:
+	CSRefHTTPHeaders(CSVector *list):CSHTTPHeaders(list){}
+	~CSRefHTTPHeaders(){}	
+};
+
 class CSHTTPInputStream : public CSInputStream, public CSHTTPHeaders {
 public:
 	CSHTTPInputStream(CSInputStream *s);
 	virtual ~CSHTTPInputStream();
 
-	void readHead();
-	uint64_t getContentLength();
+	void readHead(bool trace = false);
+	void readBody();
+	bool getContentLength(uint64_t *length);
 	const char *getMethod();
+	char *getBodyData() { return iBody.getCString(); };
+	size_t getBodyLength() { return iBody.length(); };
+	void setBody(CSStringBufferImpl *buf) { iBody.take(buf); }
 	int getStatus() { return iStatus; }
 	CSString *getStatusPhrase() { return iStatusPhrase; }
 	CSString *getRequestURI() { return iRequestURI; }
@@ -127,7 +139,9 @@
 
 	virtual int peek();
 
-	virtual void reset() {iInput->reset();}
+	virtual void reset() { iInput->reset(); }
+
+	virtual const char *identify() { return iInput->identify(); }
 
 	static CSHTTPInputStream *newStream(CSInputStream* i);
 
@@ -135,12 +149,13 @@
 	void freeHead();
 
 	// Request-Line   = Method SP Request-URI SP HTTP-Version CRLF
-	CSInputStream *iInput;
-	int		iStatus;
-	CSString *iMethod;
-	CSString *iRequestURI;
-	CSString *iHTTPVersion;
-	CSString *iStatusPhrase;
+	CSInputStream	*iInput;
+	int				iStatus;
+	CSString		*iMethod;
+	CSString		*iRequestURI;
+	CSString		*iHTTPVersion;
+	CSString		*iStatusPhrase;
+	CSStringBuffer	iBody;
 };
 
 class CSHTTPOutputStream : public CSOutputStream, public CSHTTPHeaders {
@@ -153,14 +168,31 @@
 	void setContentLength(uint64_t size) { iContentLength = size; }
 	void setRange(uint64_t size, uint64_t offset, uint64_t total) { iRangeSize = size; iRangeOffset = offset; iTotalLength = total;}
 
-	void writeHead();	// Writes a standard HTTP header.
-	void writeHeaders(); // Write the current headers.
+	void writeHead(bool trace = false);	// Writes a standard HTTP header.
+	void writeHeaders(bool trace = false); // Write the current headers.
 
 	void clearBody();
 	void writeBody();
 
+	// The virtual and non virtual print() methods
+	// must be kept seperate to avoid possible compiler
+	// warnings about hidden methods.
+	void print(const char *str, bool trace);
+	void print(int32_t value, bool trace);
+	void print(uint64_t value, bool trace);
+
+	virtual void print(const char *str) {print(str, false);}
+	virtual void print(CSString *s) {print(s->getCString(), false);}
+	virtual void print(int32_t value) {print(value, false);}
+	virtual void print(uint64_t value) {print(value, false);}
+
 	void appendBody(const char *str);
-	void appendBody(int value);
+	void appendBody(int32_t value);
+	void appendBody(uint32_t value);
+	void appendBody(uint64_t value);
+	const char *getBodyData();
+	size_t getBodyLength();
+	void setBody(CSStringBufferImpl *buf);
 
 	virtual void close();
 
@@ -172,7 +204,9 @@
 
 	virtual void write(char b);
 
-	virtual void reset() {iOutput->reset();}
+	virtual void reset() { iOutput->reset(); }
+
+	virtual const char *identify() { return iOutput->identify(); }
 
 	static const char *getReasonPhrase(int code);
 

=== modified file 'plugin/pbms/src/cslib/CSLog.cc'
--- plugin/pbms/src/cslib/CSLog.cc	2010-10-02 21:15:42 +0000
+++ plugin/pbms/src/cslib/CSLog.cc	2010-10-18 21:21:20 +0000
@@ -29,6 +29,9 @@
 
 #include <string.h>
 #include <time.h>
+#include <stdarg.h>
+#include <stdio.h>
+#include <stdlib.h>
 
 #include "CSLog.h"
 #include "CSMemory.h"
@@ -36,6 +39,14 @@
 #include "CSStrUtil.h"
 #include "CSThread.h"
 #include "CSGlobal.h"
+
+
+//#ifdef DEBUG
+//#define DEFAULT_LOG_BUFFER_SIZE			10
+//#else
+#define DEFAULT_LOG_BUFFER_SIZE			2000
+//#endif
+
 /*
  * The global logging object.
  */
@@ -160,5 +171,59 @@
 	unlock();
 }
 
-
+void CSLog::log_va(CSThread *self, int level, const char *func, const char *file, int line, const char *fmt, va_list ap)
+{
+	char buffer[DEFAULT_LOG_BUFFER_SIZE];
+	char *log_string = NULL;
+
+	lock();
+
+#if !defined(va_copy) || defined(OS_SOLARIS)
+	int len;
+
+	len = vsnprintf(buffer, DEFAULT_LOG_BUFFER_SIZE-1, fmt, ap);
+	if (len > DEFAULT_LOG_BUFFER_SIZE-1)
+		len = DEFAULT_LOG_BUFFER_SIZE-1;
+	buffer[len] = 0;
+	log_string = buffer;
+#else
+	/* Use the buffer, unless it is too small */
+	va_list ap2;
+
+	va_copy(ap2, ap);
+	if (vsnprintf(buffer, DEFAULT_LOG_BUFFER_SIZE, fmt, ap) >= DEFAULT_LOG_BUFFER_SIZE) {
+		if (vasprintf(&log_string, fmt, ap2) == -1)
+			log_string = NULL;
+	}
+	else
+		log_string = buffer;
+#endif
+
+	if (log_string) {
+		log(self, func, file, line, level, log_string);
+
+		if (log_string != buffer)
+			free(log_string);
+	}
+
+	unlock();
+}
+
+void CSLog::logf(CSThread *self, int level, const char *fmt, ...)
+{
+	va_list	ap;
+
+	va_start(ap, fmt);
+	log_va(self, level, NULL, NULL, 0, fmt, ap);
+	va_end(ap);
+}
+
+void CSLog::logf(CSThread *self, int level, const char *func, const char *file, int line, const char *fmt, ...)
+{
+	va_list	ap;
+
+	va_start(ap, fmt);
+	log_va(self, level, func, file, line, fmt, ap);
+	va_end(ap);
+}
 

=== modified file 'plugin/pbms/src/cslib/CSLog.h'
--- plugin/pbms/src/cslib/CSLog.h	2010-10-02 21:15:42 +0000
+++ plugin/pbms/src/cslib/CSLog.h	2010-10-18 21:21:20 +0000
@@ -29,9 +29,9 @@
 #ifndef __CSLOG_H__
 #define __CSLOG_H__
 
-using namespace std;
+#include <stdio.h>
+#include <stdarg.h>
 
-#include "stdio.h"
 
 #include "CSDefs.h"
 #include "CSString.h"
@@ -47,14 +47,14 @@
 		iStream(s),
 		iHeaderPending(true),
 		iLogLevel(level),
-		iLockThread(0),
 		iLockCount(0)
 		 {
 		pthread_mutex_init(&iMutex, NULL);
+		memset(&iLockThread, 0, sizeof(iLockThread));
 	}
 
 	virtual ~CSLog() {
-		iLockThread = 0;
+		memset(&iLockThread, 0, sizeof(iLockThread));
 		iLockCount = 0;
 		pthread_mutex_destroy(&iMutex);
 	}
@@ -88,6 +88,9 @@
 	void eol(CSThread *self, int level);
 
 	void logLine(CSThread *self, int level, const char *buffer);
+	void log_va(CSThread *self, int level, const char *func, const char *file, int line, const char *fmt, va_list ap);
+	void logf(CSThread *self, int level, const char *fmt, ...);
+	void logf(CSThread *self, int level, const char *func, const char *file, int line, const char *fmt, ...);
 	
 	void flush() {fflush(iStream);}
 private:

=== modified file 'plugin/pbms/src/cslib/CSMd5.cc'
--- plugin/pbms/src/cslib/CSMd5.cc	2010-05-20 22:52:10 +0000
+++ plugin/pbms/src/cslib/CSMd5.cc	2010-10-18 21:21:20 +0000
@@ -55,8 +55,8 @@
 #include <string.h>
 
 #include "CSDefs.h"
-
 #include "CSMd5.h"
+#include "CSStrUtil.h"
 
 #undef BYTE_ORDER	/* 1 = big-endian, -1 = little-endian, 0 = unknown */
 #ifdef ARCH_IS_BIG_ENDIAN
@@ -321,6 +321,9 @@
 void
 CSMd5::md5_init()
 {
+	/* Indicates that there is no digest: */
+	digest_cstr[0] = 0;
+
     md5_state.count[0] = md5_state.count[1] = 0;
     md5_state.abcd[0] = 0x67452301;
     md5_state.abcd[1] = /*0xefcdab89*/ T_MASK ^ 0x10325476;
@@ -339,7 +342,6 @@
 		return;
 	}
 	
-	
    /* Update the message length. */
     md5_state.count[1] += nbytes >> 29;
     md5_state.count[0] += nbits;
@@ -368,8 +370,7 @@
 }
 
 
-void
-CSMd5::md5_digest(Md5Digest *digest)
+void CSMd5::md5_digest()
 {
     static const u_char pad[64] = {
 	0x80, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
@@ -388,5 +389,8 @@
     /* Append the length. */
     md5_append(data, 8);
     for (i = 0; i < 16; ++i)
-		digest->val[i] = (u_char)(md5_state.abcd[i >> 2] >> ((i & 3) << 3));
+		digest[i] = (u_char)(md5_state.abcd[i >> 2] >> ((i & 3) << 3));
+
+	/* Generate the text version: */
+	cs_bin_to_hex(MD5_CHECKSUM_STRING_SIZE, digest_cstr, MD5_CHECKSUM_SIZE, digest);
 }

=== modified file 'plugin/pbms/src/cslib/CSMd5.h'
--- plugin/pbms/src/cslib/CSMd5.h	2010-10-02 21:15:42 +0000
+++ plugin/pbms/src/cslib/CSMd5.h	2010-10-18 21:21:20 +0000
@@ -70,48 +70,70 @@
  *
  */
 #ifndef md5_INCLUDED
-#  define md5_INCLUDED
+#define md5_INCLUDED
+
 #include <string.h>
 
+#include "CSDefs.h"
+
 /* Define the state of the MD5 Algorithm. */
-
-
 typedef unsigned int md5_word_t;
-#define CHECKSUM_VALUE_SIZE	16
 
-typedef struct {
-	u_char val[CHECKSUM_VALUE_SIZE];
-} Md5Digest;
+#define MD5_CHECKSUM_SIZE			CHECKSUM_VALUE_SIZE
+#define MD5_CHECKSUM_STRING_SIZE	(CHECKSUM_VALUE_SIZE * 2 + 1)
 
 class CSMd5 {
 	private:
 	struct md5_state_s {
-		md5_word_t count[2];		/* message length in bits, lsw first */
-		md5_word_t abcd[4];		/* digest buffer */
-		u_char buf[64];		/* accumulate block */
+		md5_word_t	count[2];		/* message length in bits, lsw first */
+		md5_word_t	abcd[4];		/* digest buffer */
+		u_char		buf[64];		/* accumulate block */
 	} md5_state;
 
+	u_char			digest[MD5_CHECKSUM_SIZE];
+	char			digest_cstr[MD5_CHECKSUM_STRING_SIZE];
+
 	void md5_process(const u_char *data /*[64]*/);
+	void md5_digest();
 
 	public:
-	CSMd5()
-	{
+	CSMd5() {
 		md5_init();
 	}
-	
+
 	void md5_init();
 
 	/* Append a string to the message. */
 	void md5_append(const u_char  *data, int nbytes);
-	void md5_append(const char  *data)
-	{
+
+	void md5_append(const char  *data) {
 		md5_append((const u_char  *) data, strlen(data));
 	}
-	
 
 	/* Finish the message and return the digest. */
-	void md5_digest(Md5Digest *digest);
+	
+	// Returns the raw bin digest.
+	const u_char *md5_get_digest() {
+		if (!digest_cstr[0])
+			md5_digest();
+		return digest;
+	}
 
+	// Returns the bin/hex digest.
+	void md5_get_digest(Md5Digest *d) {
+		if (!digest_cstr[0])
+			md5_digest();
+			
+		memcpy(d->val, digest, CHECKSUM_VALUE_SIZE);
+	}
+	
+	// Returns the bin/hex digest.
+	const char *md5_get_digest_cstr() {
+		if (!digest_cstr[0])
+			md5_digest();
+		return digest_cstr;
+	}
+	
 };
 
 #endif /* md5_INCLUDED */

=== modified file 'plugin/pbms/src/cslib/CSMemory.cc'
--- plugin/pbms/src/cslib/CSMemory.cc	2010-10-02 21:15:42 +0000
+++ plugin/pbms/src/cslib/CSMemory.cc	2010-10-18 21:21:20 +0000
@@ -599,7 +599,6 @@
 	if (!track_me && !mm_tracking_id)
 		return;
 
-
 	if (func) {
 		cs_format_context(300, buffer, func, file, line);
 		fprintf(stderr, "TRACKING (%"PRIu32"): %s %2"PRIu32" %s", debug_ptr->md_id, inc ? "INC" : "DEC", ref_cnt, buffer);
@@ -672,3 +671,28 @@
 #endif
 }
 
+#ifdef DEBUG
+uint32_t cs_mm_get_check_point()
+{
+	return mm_alloc_count;
+}
+
+// Reports any memory allocated after the check_point
+// but has not been freed.
+void cs_mm_assert_check_point(uint32_t check_point)
+{
+	uint32_t mm;
+
+	if (!mm_addresses)
+		return;
+
+	pthread_mutex_lock(&mm_mutex);
+	for (mm=0; mm<mm_nr_in_use; mm++) {
+		if (mm_addresses[mm].id >= check_point)
+			mm_throw_assertion(&mm_addresses[mm], mm_addresses[mm].ptr, "Not freed");
+	}
+
+	pthread_mutex_unlock(&mm_mutex);
+
+}
+#endif

=== modified file 'plugin/pbms/src/cslib/CSMemory.h'
--- plugin/pbms/src/cslib/CSMemory.h	2010-10-02 21:15:42 +0000
+++ plugin/pbms/src/cslib/CSMemory.h	2010-10-18 21:21:20 +0000
@@ -66,6 +66,8 @@
 size_t cs_mm_malloc_size(void *ptr);
 void cs_mm_print_track(const char *func, const char *file, uint32_t line, void *p, bool inc, uint32_t ref_cnt, int track_me);
 void cs_mm_track_memory(const char *func, const char *file, uint32_t line, void *p, bool inc, uint32_t ref_cnt, int track_me);
+uint32_t cs_mm_get_check_point();
+void cs_mm_assert_check_point(uint32_t check_point);
 #endif
 
 #endif

=== modified file 'plugin/pbms/src/cslib/CSMutex.cc'
--- plugin/pbms/src/cslib/CSMutex.cc	2010-10-02 21:15:42 +0000
+++ plugin/pbms/src/cslib/CSMutex.cc	2010-10-18 21:21:20 +0000
@@ -28,11 +28,17 @@
 #include "CSConfig.h"
 
 #include <assert.h>
+#ifdef OS_WINDOWS
+extern int gettimeofday(struct timeval *tv, struct timezone *tz);
+#else
 #include <sys/time.h>
+#endif
+
 
 #include "CSException.h"
 #include "CSMutex.h"
 #include "CSGlobal.h"
+#include "CSLog.h"
 
 /*
  * ---------------------------------------------------------------
@@ -40,6 +46,11 @@
  */
 
 CSMutex::CSMutex()
+#ifdef DEBUG
+:
+iLocker(NULL),
+trace(false)
+#endif
 {
 	int err;
 
@@ -58,10 +69,20 @@
 
 	if ((err = pthread_mutex_lock(&iMutex)))
 		CSException::throwOSError(CS_CONTEXT, err);
+#ifdef DEBUG
+	iLocker = CSThread::getSelf();
+	if (trace)
+		CSL.logf(iLocker, CSLog::Protocol, "Mutex locked\n");
+#endif
 }
 
 void CSMutex::unlock()
 {
+#ifdef DEBUG
+	if (trace)
+		CSL.logf(iLocker, CSLog::Protocol, "Mutex unlocked\n");
+	iLocker = NULL;
+#endif
 	pthread_mutex_unlock(&iMutex);
 }
 
@@ -154,24 +175,10 @@
 	struct timespec	abstime;
 	int				lock_count;
 	int				err;
+	uint64_t		micro_sec;
 
 	enter_();
-#ifdef XT_WIN
-	union ft64		now;
-  
-	GetSystemTimeAsFileTime(&now.ft);
-
-	/* System time is measured in 100ns units.
-	 * This calculation will be reversed by the Windows implementation
-	 * of pthread_cond_timedwait(), in order to extract the
-	 * milli-second timeout!
-	 */
-	abstime.tv.i64 = now.i64 + (milli_sec * 10000);
-  
-	abstime.max_timeout_msec = milli_sec;
-#else
 	struct timeval	now;
-	uint64_t			micro_sec;
 
 	/* Get the current time in microseconds: */
 	gettimeofday(&now, NULL);
@@ -183,7 +190,7 @@
 	/* Setup the end time, which is in nano-seconds. */
 	abstime.tv_sec = (long) (micro_sec / 1000000);				/* seconds */
 	abstime.tv_nsec = (long) ((micro_sec % 1000000) * 1000);	/* and nanoseconds */
-#endif
+
 	ASSERT(iLockingThread == self);
 	lock_count = iLockCount;
 	iLockCount = 0;

=== modified file 'plugin/pbms/src/cslib/CSMutex.h'
--- plugin/pbms/src/cslib/CSMutex.h	2010-10-02 21:15:42 +0000
+++ plugin/pbms/src/cslib/CSMutex.h	2010-10-18 21:21:20 +0000
@@ -33,6 +33,8 @@
 
 #include "CSDefs.h"
 
+class CSThread;
+
 class CSMutex {
 public:
 	CSMutex();
@@ -46,6 +48,11 @@
 
 private:
 	pthread_mutex_t	iMutex;
+#ifdef DEBUG
+	CSThread		*iLocker;
+public:
+	bool			trace;
+#endif
 };
 
 class CSLock : public CSMutex {

=== modified file 'plugin/pbms/src/cslib/CSObject.cc'
--- plugin/pbms/src/cslib/CSObject.cc	2010-10-02 21:15:42 +0000
+++ plugin/pbms/src/cslib/CSObject.cc	2010-10-18 21:21:20 +0000
@@ -44,29 +44,26 @@
  * BASIC OBJECTS
  */
 
+#ifdef DEBUG
+void CSObject::retain(const char *func, const char *file, uint32_t line)
+{
+	CSException::throwAssertion(func, file, line, "Non-referenced object cannot be referenced");
+}
+#else
 void CSObject::retain()
 {
 	CSException::throwAssertion(CS_CONTEXT, "Non-referenced object cannot be referenced");
 }
+#endif
 
+#ifdef DEBUG
+void CSObject::release(const char *, const char *, uint32_t )
+#else
 void CSObject::release()
-{
-	delete this;
-}
-
-#ifdef DEBUG
-/*
-void CSObject::retain(const char *func, const char *file, uint32_t line)
-{
-	CSException::throwAssertion(CS_CONTEXT, "Non-referenced object cannot be referenced");
-}
-
-void CSObject::release(const char *func, const char *file, uint32_t line)
-{
-	delete this;
-}
-*/
 #endif
+{
+	delete this;
+}
 
 CSObject *CSObject::getKey() { CSException::throwCoreError(CS_CONTEXT, CS_ERR_IMPL_MISSING, __FUNC__); return NULL; }
 
@@ -85,10 +82,34 @@
 void CSObject::setNextLink(CSObject *) { CSException::throwCoreError(CS_CONTEXT, CS_ERR_IMPL_MISSING, __FUNC__); }
 
 void CSObject::setPrevLink(CSObject *) { CSException::throwCoreError(CS_CONTEXT, CS_ERR_IMPL_MISSING, __FUNC__); }
+
+/*
+ * ---------------------------------------------------------------
+ * STATIC OBJECTS
+ */
+
+#ifdef DEBUG
+void CSStaticObject::retain(const char *, const char *, uint32_t )
+#else
+void CSStaticObject::retain()
+#endif
+{
+}
+
+#ifdef DEBUG
+void CSStaticObject::release(const char *, const char *, uint32_t )
+#else
+void CSStaticObject::release()
+#endif
+{
+	finalize();
+}
+
 /*
  * ---------------------------------------------------------------
  * REFERENCE OBJECTS
  */
+
 CSRefObject::CSRefObject():
 CSObject(),
 iRefCount(1)
@@ -104,47 +125,32 @@
 	ASSERT(iRefCount == 0);
 }
 
+#ifdef DEBUG
+void CSRefObject::retain(const char *func, const char *file, uint32_t line)
+#else
 void CSRefObject::retain()
+#endif
 {
 	if (!iRefCount)
 		CSException::throwAssertion(CS_CONTEXT, "Freed object being retained.");
 		
 	iRefCount++;
 #ifdef DEBUG
-	cs_mm_track_memory(NULL, NULL, 0, this, true, iRefCount, iTrackMe);
+	cs_mm_track_memory(func, file, line, this, true, iRefCount, iTrackMe);
 #endif
 }
 
+#ifdef DEBUG
+void CSRefObject::release(const char *func, const char *file, uint32_t line)
+#else
 void CSRefObject::release()
-{
-	bool terminate;
-
-#ifdef DEBUG
-	cs_mm_track_memory(NULL, NULL, 0, this, false, iRefCount, iTrackMe);
 #endif
-	iRefCount--;
-	if (!iRefCount)
-		terminate = true;
-	else
-		terminate = false;
-
-	if (terminate)
-		delete this;
-}
+{
+	bool terminate;
 
 #ifdef DEBUG
-/*
-void CSRefObject::retain(const char *func, const char *file, uint32_t line)
-{
-	iRefCount++;
-	cs_mm_print_track(func, file, line, this, true, iRefCount);
-}
-
-void CSRefObject::release(const char *func, const char *file, uint32_t line)
-{
-	bool terminate;
-
 	cs_mm_track_memory(func, file, line, this, false, iRefCount, iTrackMe);
+#endif
 	iRefCount--;
 	if (!iRefCount)
 		terminate = true;
@@ -154,7 +160,13 @@
 	if (terminate)
 		delete this;
 }
-*/
+
+#ifdef DEBUG
+void CSRefObject::startTracking()
+{
+	iTrackMe = 1;
+	cs_mm_track_memory(NULL, NULL, 0, this, true, iRefCount, iTrackMe);
+}
 #endif
 
 /*
@@ -178,23 +190,31 @@
 	ASSERT(iRefCount == 0);
 }
 
+#ifdef DEBUG
+void CSSharedRefObject::retain(const char *func, const char *file, uint32_t line)
+#else
 void CSSharedRefObject::retain()
+#endif
 {
 	lock();
 	iRefCount++;
 #ifdef DEBUG
-	cs_mm_track_memory(NULL, NULL, 0, this, true, iRefCount, iTrackMe);
+	cs_mm_track_memory(func, file, line, this, true, iRefCount, iTrackMe);
 #endif
 	unlock();
 }
 
+#ifdef DEBUG
+void CSSharedRefObject::release(const char *func, const char *file, uint32_t line)
+#else
 void CSSharedRefObject::release()
+#endif
 {
 	bool terminate;
 
 	lock();
 #ifdef DEBUG
-	cs_mm_track_memory(NULL, NULL, 0, this, false, iRefCount, iTrackMe);
+	cs_mm_track_memory(func, file, line, this, false, iRefCount, iTrackMe);
 #endif
 	iRefCount--;
 	if (!iRefCount)

=== modified file 'plugin/pbms/src/cslib/CSObject.h'
--- plugin/pbms/src/cslib/CSObject.h	2010-10-02 21:15:42 +0000
+++ plugin/pbms/src/cslib/CSObject.h	2010-10-18 21:21:20 +0000
@@ -42,13 +42,24 @@
 // operator out side of a 'for' loop. :)
 #define RETAIN(x) ((x)->retain(),(x))
 
+// A back reference is a reference stored by an object pointing to it's owner.
+// Since the owner will free all of it's object when it is released the  object
+// itself cannot hold a reference to its owner.
+// Use this macro to make it clear that this is what you intended. 
+#define BACK_REFERENCE(x) (x)
+
 class CSObject {
 public:
 	CSObject() { }
 	virtual ~CSObject() { finalize(); }
 
+#ifdef DEBUG
+	virtual void retain(const char *func, const char *file, uint32_t line);
+	virtual void release(const char *func, const char *file, uint32_t line);
+#else
 	virtual void retain();
 	virtual void release();
+#endif
 
 	/*
 	 * All objects are sortable, hashable and linkable,
@@ -75,8 +86,18 @@
 		cs_mm_free(ptr);
 	}
 
-	//virtual void retain(const char *func, const char *file, uint32_t line);
-	//virtual void release(const char *func, const char *file, uint32_t line);
+#endif
+};
+
+class CSStaticObject : public CSObject {
+#ifdef DEBUG
+	virtual void retain(const char *func, const char *file, uint32_t line);
+	virtual void release(const char *func, const char *file, uint32_t line);
+
+	int		iTrackMe;
+#else
+	virtual void retain();
+	virtual void release();
 #endif
 };
 
@@ -85,17 +106,20 @@
 	CSRefObject();
 	virtual ~CSRefObject();
 
+#ifdef DEBUG
+	virtual void startTracking();
+	virtual void retain(const char *func, const char *file, uint32_t line);
+	virtual void release(const char *func, const char *file, uint32_t line);
+
+	int		iTrackMe;
+#else
 	virtual void retain();
 	virtual void release();
-#ifdef DEBUG
-	//virtual void retain(const char *func, const char *file, uint32_t line);
-	//virtual void release(const char *func, const char *file, uint32_t line);
-	int		iTrackMe;
 #endif
 
-#ifndef DEBUG
+	uint32_t getRefCount() { return iRefCount;}
+
 private:
-#endif
 	uint32_t	iRefCount;
 };
 
@@ -104,32 +128,28 @@
 	CSSharedRefObject();
 	virtual ~CSSharedRefObject();
 
+#ifdef DEBUG
+	virtual void startTracking();
+	virtual void retain(const char *func, const char *file, uint32_t line);
+	virtual void release(const char *func, const char *file, uint32_t line);
+
+	int		iTrackMe;
+#else
 	virtual void retain();
 	virtual void release();
-#ifdef DEBUG
-	virtual void startTracking();
-	//virtual void retain(const char *func, const char *file, uint32_t line);
-	//virtual void release(const char *func, const char *file, uint32_t line);
-	int		iTrackMe;
 #endif
 
-#ifndef DEBUG
+	uint32_t getRefCount() { return iRefCount;}
+
 private:
-#endif
-	uint32_t	iRefCount;
-};
-
-class CSStaticObject : public CSObject {
-	virtual void retain() {}
-	virtual void release(){ finalize();}
-#ifdef DEBUG
-	int		iTrackMe;
-#endif
-	uint32_t	iRefCount;
-};
-
-#ifdef DEBUG
-#define new new(__FUNC__, __FILE__, __LINE__)
+	uint32_t	iRefCount;
+};
+
+#ifdef DEBUG
+#define new			new(__FUNC__, __FILE__, __LINE__)
+
+#define retain()	retain(__FUNC__, __FILE__, __LINE__)
+#define release()	release(__FUNC__, __FILE__, __LINE__)
 #endif
 
 class CSPooled {

=== modified file 'plugin/pbms/src/cslib/CSPath.cc'
--- plugin/pbms/src/cslib/CSPath.cc	2010-10-02 21:15:42 +0000
+++ plugin/pbms/src/cslib/CSPath.cc	2010-10-18 21:21:20 +0000
@@ -28,18 +28,11 @@
 
 #include "CSConfig.h"
 
-#include <unistd.h>
-#include <sys/stat.h>
-#include <sys/param.h>
 #include <string.h>
 #include <stdio.h>
 
-static int unix_file_rename(const char *from, const char *to)
-{
-	return rename(from, to);
-}
-
 #include "CSStrUtil.h"
+#include "CSSys.h"
 #include "CSFile.h"
 #include "CSDirectory.h"
 #include "CSPath.h"
@@ -49,6 +42,7 @@
  * ---------------------------------------------------------------
  * CORE SYSTEM PATH
  */
+CSLock	CSPath::iRename_lock;
 
 CSPath *CSPath::iCWD = NULL;
 
@@ -58,40 +52,36 @@
 		iPath->release();
 }
 
+CSFile *CSPath::try_CreateAndOpen(CSThread *self, int mode, bool retry)
+{
+	try_(a) {
+		return openFile(mode | CSFile::CREATE); // success, do not try again.
+	}
+	catch_(a) {
+		if (retry || !CSFile::isDirNotFound(&self->myException))
+			throw_();
+
+		/* Make sure the parent directory exists: */
+		CSPath	*dir = CSPath::newPath(RETAIN(this), "..");
+		push_(dir);
+		dir->makePath();
+		release_(dir);
+
+	}
+	cont_(a);
+	return NULL; 
+}
+
 CSFile *CSPath::createFile(int mode)
 {
-	CSFile	*file;
-	CSPath	*dir;
+	CSFile	*file = NULL;
 	bool	retry = false;
 
-	CLOBBER_PROTECT(retry);
-
 	enter_();
-	/* Create and open the file: */
-	do {
-		try_(a) {
-			file = openFile(mode | CSFile::CREATE);
-			retry = false;
-		}
-		catch_(a) {
-			if (retry || !CSFile::isDirNotFound(&self->myException))
-				throw_();
-
-			/* Make sure the parent directory exists: */
-			dir = CSPath::newPath(RETAIN(this), "..");
-			try_(b) {
-				dir->makePath();
-			}
-			finally_(b) {
-				dir->release();
-			}
-			finally_end_block(b);
-
-			retry = true;
-		}
-		cont_(a);
+	while (file == NULL) {
+		file = try_CreateAndOpen(self, mode, retry);
+		retry = true;
 	}
-	while (retry);
 	return_(file);
 }
 
@@ -157,14 +147,13 @@
 	}
 
 	path = CSPath::newPath(RETAIN(this), "..");
-	try_(a) {
-		path->makePath();
-		makeDir();
-	}
-	finally_(a) {
-		path->release();
-	}
-	finally_end_block(a);
+	push_(path);
+	
+	path->makePath();
+	makeDir();
+	
+	release_(path);
+
 	exit_();
 }
 
@@ -227,13 +216,12 @@
 
 bool CSPath::isLink()
 {
-	struct stat sb;
+	bool link;
 	enter_();
 
-	if (lstat(iPath->getCString(), &sb) == -1)
-		CSException::throwFileError(CS_CONTEXT, iPath, errno);
-		
-	return_(S_ISLNK(sb.st_mode));
+	link = sys_isLink(iPath->getCString());
+	
+	return_(link);
 }
 
 bool CSPath::isEmpty()
@@ -249,14 +237,12 @@
 		return_(false);
 
 	dir = openDirectory();
-	try_(a) {
-		if (dir->next())
-			result = false;
-	}
-	finally_(a) {
-		dir->release();	
-	}
-	finally_end_block(a);
+	push_(dir);
+	
+	if (dir->next())
+		result = false;
+
+	release_(dir);
 	return_(result);
 }
 
@@ -274,23 +260,49 @@
 		CSException::throwFileError(CS_CONTEXT, iPath, ENOTDIR);
 
 	dir = openDirectory();
-	try_(a) {
-		while (dir->next()) {
-			path = CSPath::newPath(RETAIN(this), dir->name());
-			if (dir->isFile())
-				path->remove();
-			else
-				path->removeDir();
-			path->release();
-			path = NULL;
-		}
-	}
-	finally_(a) {
-		if (path)
-			path->release();
-		dir->release();	
-	}
-	finally_end_block(a);
+	push_(dir);
+
+	while (dir->next()) {
+		path = CSPath::newPath(RETAIN(this), dir->name());
+		push_(path);
+		if (dir->isFile())
+			path->remove();
+		else
+			path->removeDir();
+		release_(path);
+	}
+
+	release_(dir);
+	exit_();
+}
+
+void CSPath::emptyPath()
+{
+	CSDirectory *dir;
+	CSPath		*path = NULL;
+	bool		is_dir;
+
+	enter_();
+	if (!exists(&is_dir))
+		exit_();
+
+	if (!is_dir)
+		CSException::throwFileError(CS_CONTEXT, iPath, ENOTDIR);
+
+	dir = openDirectory();
+	push_(dir);
+
+	while (dir->next()) {
+		path = CSPath::newPath(RETAIN(this), dir->name());
+		push_(path);
+		if (dir->isFile())
+			path->remove();
+		else
+			path->emptyPath();
+		release_(path);
+	}
+
+	release_(dir);
 	exit_();
 }
 
@@ -317,31 +329,31 @@
 
 void CSPath::moveTo(CSPath *in_to_path)
 {
-	CSPath	*to_path = in_to_path;
+	CSPath	*to_path = NULL;
 	bool	is_dir;
 
 	enter_();
+	push_(in_to_path);
+	
 	if (!exists(NULL))
 		CSException::throwFileError(CS_CONTEXT, iPath, ENOENT);
 
-	try_(a) {
-		if (to_path->exists(&is_dir)) {
-			if (is_dir) {
-			 	to_path = CSPath::newPath(RETAIN(in_to_path), getNameCString());
-			 	if (to_path->exists(NULL))
-					CSException::throwFileError(CS_CONTEXT, to_path->getCString(), EEXIST);
-			}
-			else
-				CSException::throwFileError(CS_CONTEXT, to_path->getCString(), ENOTDIR);
+	if (in_to_path->exists(&is_dir)) {
+		if (is_dir) {
+			to_path = CSPath::newPath(RETAIN(in_to_path), getNameCString());
+			push_(to_path);
+			if (to_path->exists(NULL))
+				CSException::throwFileError(CS_CONTEXT, to_path->getCString(), EEXIST);
+			pop_(to_path);
 		}
-		move(to_path);
-	}
-	finally_(a) {
-		if (to_path != in_to_path)
-			to_path->release();
-	}
-	finally_end_block(a);
+		else
+			CSException::throwFileError(CS_CONTEXT, to_path->getCString(), ENOTDIR);
+	} else
+		to_path = RETAIN(in_to_path);
+		
+	move(to_path);
 
+	release_(in_to_path);
 	exit_();
 }
 
@@ -390,6 +402,14 @@
 	return cs_last_name_of_path(str);
 }
 
+off64_t CSPath::getSize(const char *path)
+{
+	off64_t size;
+
+	info(path, NULL, &size, NULL);
+	return size;
+}
+
 off64_t CSPath::getSize()
 {
 	off64_t size;
@@ -406,40 +426,25 @@
 	return is_dir;
 }
 
-/*
- * ---------------------------------------------------------------
- * A UNIX PATH
- */
-
 bool CSPath::exists(bool *is_dir)
 {
-	int err;
-
-	err = access(iPath->getCString(), F_OK);
-	if (err == -1)
+	if (!sys_exists(iPath->getCString()))
 		return false;
+		
 	if (is_dir)
 		*is_dir = isDir();
 	return true;
 }
 
+void CSPath::info(const char *path, bool *is_dir, off64_t *size, CSTime *mod_time)
+{
+	sys_stat(path, is_dir, size, mod_time);
+	
+}
+
 void CSPath::info(bool *is_dir, off64_t *size, CSTime *mod_time)
 {
-	struct stat sb;
-
-	if (stat(iPath->getCString(), &sb) == -1)
-		CSException::throwFileError(CS_CONTEXT, iPath, errno);
-	if (is_dir)
-		*is_dir = sb.st_mode & S_IFDIR;
-	if (size)
-		*size = sb.st_size;
-	if (mod_time)
-#ifdef __USE_MISC
-		/* This is the Linux version: */
-		mod_time->setUTC1970(sb.st_mtim.tv_sec, sb.st_mtim.tv_nsec);
-#else
-		mod_time->setUTC1970(sb.st_mtime, 0);
-#endif
+	info(iPath->getCString(), is_dir, size, mod_time);
 }
 
 CSFile *CSPath::openFile(int mode)
@@ -456,36 +461,7 @@
 
 void CSPath::removeFile()
 {
-	if (unlink(iPath->getCString()) == -1) {
-		int err = errno;
-
-		if (err != ENOENT)
-			CSException::throwFileError(CS_CONTEXT, iPath, err);
-	}
-}
-
-static void cs_set_stats(char *path)
-{
-	char		super_path[PATH_MAX];
-	struct stat	stats;
-	char		*ptr;
-
-	ptr = cs_last_name_of_path(path);
-	if (ptr == path) 
-		strcpy(super_path, ".");
-	else {
-		cs_strcpy(PATH_MAX, super_path, path);
-
-		if ((ptr = cs_last_name_of_path(super_path)))
-			*ptr = 0;
-	}
-	if (stat(super_path, &stats) == -1)
-		CSException::throwFileError(CS_CONTEXT, path, errno);
-
-	if (chmod(path, stats.st_mode) == -1)
-		CSException::throwFileError(CS_CONTEXT, path, errno);
-
-	/*chown(path, stats.st_uid, stats.st_gid);*/
+	sys_removeFile(iPath->getCString());
 }
 
 void CSPath::makeDir()
@@ -495,10 +471,7 @@
 	cs_strcpy(PATH_MAX, path, iPath->getCString());
 	cs_remove_dir_char(path);
 
-	if (mkdir(path, S_IRWXU | S_IRWXG | S_IRWXO) == -1)
-		CSException::throwFileError(CS_CONTEXT, iPath, errno);
-
-	cs_set_stats(path);
+	sys_makeDir(path);
 }
 
 CSDirectory *CSPath::openDirectory()
@@ -516,12 +489,7 @@
 void CSPath::removeDir()
 {			
 	emptyDir();
-	if (rmdir(iPath->getCString()) == -1) {
-		int err = errno;
-
-		if (err != ENOENT)
-			CSException::throwFileError(CS_CONTEXT, iPath, err);
-	}
+	sys_removeDir(iPath->getCString());
 }
 
 void CSPath::rename(const char *name)
@@ -536,13 +504,18 @@
 	cs_add_dir_char(PATH_MAX, new_path);
 	cs_strcat(PATH_MAX, new_path, name);
 	
+	lock_(&iRename_lock); // protect against race condition when testing if the new name exists yet or not.	
+	if (sys_exists(new_path))
+		CSException::throwFileError(CS_CONTEXT, new_path, EEXIST);
+	
 	tmp_path = CSString::newString(new_path);
 	push_(tmp_path);
 	
-	if (unix_file_rename(iPath->getCString(), new_path) == -1)
-		CSException::throwFileError(CS_CONTEXT, iPath, errno);
+	sys_rename(iPath->getCString(), new_path);
 		
 	pop_(tmp_path);
+	unlock_(&iRename_lock);
+	
 	old_path = iPath;
 	iPath = tmp_path;
 	
@@ -552,18 +525,22 @@
 
 void CSPath::move(CSPath *to_path)
 {
+	enter_();
+	lock_(&iRename_lock); // protect against race condition when testing if the new name exists yet or not.	
+	if (to_path->exists())
+		CSException::throwFileError(CS_CONTEXT, to_path->getCString(), EEXIST);
+		
 	/* Cannot move from TD to non-TD: */
-	if (unix_file_rename(iPath->getCString(), to_path->getCString()) == -1)
-		CSException::throwFileError(CS_CONTEXT, iPath, errno);
+	sys_rename(iPath->getCString(), to_path->getCString());
+	unlock_(&iRename_lock);
+	exit_();
 }
 
 CSPath *CSPath::getCWD()
 {
-	char path[MAXPATHLEN];
-
-	if (!getcwd(path, MAXPATHLEN))
-		CSException::throwOSError(CS_CONTEXT, errno);
-
+	char path[PATH_MAX];
+
+	sys_getcwd(path, PATH_MAX);
 	return newPath(path);
 }
 

=== modified file 'plugin/pbms/src/cslib/CSPath.h'
--- plugin/pbms/src/cslib/CSPath.h	2010-10-02 21:15:42 +0000
+++ plugin/pbms/src/cslib/CSPath.h	2010-10-18 21:21:20 +0000
@@ -37,13 +37,12 @@
 #include "CSTime.h"
 #include "CSDefs.h"
 #include "CSString.h"
-
-using namespace std;
+#include "CSSys.h"
 
 class CSFile;
 class CSDirectory;
 
-class CSPath : public CSRefObject {
+class CSPath : public CSRefObject, public CSSys {
 public:
 	virtual CSFile *createFile(int mode);
 
@@ -65,6 +64,9 @@
 	/* Delete the contents of a directory */
 	virtual void emptyDir();
 	
+	/* Recursively delete the contents of a directory */
+	virtual void emptyPath();
+	
 	/* Copy a file or directory to the specified location. */
 	virtual void copyTo(CSPath *to_path, bool overwrite);
 
@@ -76,7 +78,7 @@
 	virtual void remove();
 
 	/* Create an empty file. */
-	virtual void touch(bool create_path);
+	virtual void touch(bool create_path = false);
 
 	virtual CSString *getString();
 
@@ -92,9 +94,13 @@
 
 	virtual bool exists() { return exists(NULL); }
 
+	static void info(const char *path, bool *is_dir, off64_t *size, CSTime *mod_time);
+
 	virtual void info(bool *is_dir, off64_t *size, CSTime *mod_time);
 
-	virtual off64_t getSize();
+	static off64_t getSize(const char *path);
+	
+ 	virtual off64_t getSize();
 
 	virtual bool isDir();
 
@@ -137,6 +143,8 @@
 	static CSPath *newPath(const char *cwd, const char *path);
 
 private:
+	CSFile *try_CreateAndOpen(CSThread *self, int mode, bool retry);
+	static CSLock iRename_lock;
 	CSPath():iPath(NULL) { }
 
 	virtual ~CSPath();

=== modified file 'plugin/pbms/src/cslib/CSS3Protocol.cc'
--- plugin/pbms/src/cslib/CSS3Protocol.cc	2010-10-06 18:22:02 +0000
+++ plugin/pbms/src/cslib/CSS3Protocol.cc	2010-10-18 21:21:20 +0000
@@ -35,12 +35,11 @@
 #ifdef S3_UNIT_TEST
 //#define SHOW_SIGNING
 // Uncomment this line to trace network action during request. Very Usefull!!
-//#define DEBUG_CURL
+#define DEBUG_CURL
 #define DUMP_ERRORS
 #endif
 
-//#define DEBUG_CURL
-#define DUMP_ERRORS
+//#define DUMP_ERRORS
 //#define SHOW_SIGNING
 
 #define HEX_CHECKSUM_VALUE_SIZE (2 *CHECKSUM_VALUE_SIZE)
@@ -128,6 +127,7 @@
 	
 	CSMd5			ms_md5;
 	char			ms_s3Checksum[HEX_CHECKSUM_VALUE_SIZE +1];
+	bool			ms_calculate_md5;
 	
 	bool			ms_notFound; // True if the object could not be found
 	bool			ms_retry; // True if the request failed with a retry error.
@@ -142,12 +142,14 @@
 	bool			ms_throw_error;	// Gets set if an exception occurs in a callback.
 	bool			ms_old_libcurl;
 	char			*ms_safe_url;
+	time_t			ms_last_modified;
 	
 	S3ProtocolCon():
 		ms_curl(NULL),
 		ms_header_list(NULL),
 		ms_inputStream(NULL),
 		ms_outputStream(NULL),
+		ms_calculate_md5(false),
 		ms_notFound(false),
 		ms_retry(false),
 		ms_slowDown(false),
@@ -156,7 +158,8 @@
 		ms_replyStatus(0),
 		ms_throw_error(false),
 		ms_old_libcurl(false),
-		ms_safe_url(NULL)
+		ms_safe_url(NULL),
+		ms_last_modified(0)
 	{
 	
 		ms_curl = curl_easy_init();
@@ -230,6 +233,7 @@
 			//case 307: // Temporary Redirect
 				break;
 			case 404:	// Not Found
+			case 403:	// Forbidden (S3 object not found)
 				ms_notFound = true;
 				break;
 			case 500:	
@@ -367,9 +371,18 @@
 	{
 		enter_();
 		
-		push_(output);	
+		if (output) {
+			push_(output);
+			THROW_CURL_IF(curl_easy_setopt(ms_curl, CURLOPT_HTTPGET, 1L));
+		} else {
+			THROW_CURL_IF(curl_easy_setopt(ms_curl, CURLOPT_NOBODY, 1L));
+		}
+		
+		// 
+		THROW_CURL_IF(curl_easy_setopt(ms_curl, CURLOPT_FILETIME, 1L));
 		THROW_CURL_IF(curl_easy_setopt(ms_curl, CURLOPT_HTTPHEADER, ms_header_list));
-		THROW_CURL_IF(curl_easy_setopt(ms_curl, CURLOPT_HTTPGET, 1L));
+    // Ask curl to parse the Last-Modified header.  This is easier than
+    // parsing it ourselves.
 
 		ms_outputStream = output;	
 		if (curl_easy_perform(ms_curl) && !ms_throw_error) {
@@ -377,16 +390,19 @@
 			CSException::throwException(CS_CONTEXT, CS_ERR_GENERIC_ERROR, ms_curl_error);
 		}
 		ms_outputStream = NULL;	
-		release_(output);	
+		if (output){
+			release_(output);
+		}
 		
 		if (ms_throw_error) 
 			throw_();
 		
 		check_reply_status();
+		curl_easy_getinfo(ms_curl, CURLINFO_FILETIME, &ms_last_modified);
 		exit_();		
 	}
 	
-	inline void ms_execute_put_request(CSInputStream *input, off64_t size, Md5Digest *digest)
+	inline void ms_execute_put_request(CSInputStream *input, off64_t size)
 	{
 		enter_();
 		
@@ -414,20 +430,24 @@
 		
 		check_reply_status();
 		
-		// Check that the checksums agree
-		char checksum[HEX_CHECKSUM_VALUE_SIZE +1];
-		
-		ms_md5.md5_digest(digest);
-		cs_bin_to_hex(HEX_CHECKSUM_VALUE_SIZE, checksum, CHECKSUM_VALUE_SIZE, digest->val);
-		checksum[HEX_CHECKSUM_VALUE_SIZE] = 0;
-		
-		cs_strToUpper(ms_s3Checksum);
-		if (strcmp(checksum, ms_s3Checksum)) {
-			// The request should be restarted in this case.
-			ms_retry = true;
-			CSException::logException(CS_CONTEXT, CS_ERR_CHECKSUM_ERROR, "Calculated checksum did not match S3 checksum");
+		if (ms_calculate_md5) {
+			// If the data was not sent with an md5 checksum then verify
+			// the server's md5 value with the one calculated during the send.
+			char checksum[HEX_CHECKSUM_VALUE_SIZE +1];
+			Md5Digest digest;
+			
+			ms_md5.md5_get_digest(&digest);
+			cs_bin_to_hex(HEX_CHECKSUM_VALUE_SIZE, checksum, CHECKSUM_VALUE_SIZE, digest.val);
+			checksum[HEX_CHECKSUM_VALUE_SIZE] = 0;
+			
+			cs_strToUpper(ms_s3Checksum);
+			if (strcmp(checksum, ms_s3Checksum)) {
+				// The request should be restarted in this case.
+				ms_retry = true;
+				CSException::logException(CS_CONTEXT, CS_ERR_CHECKSUM_ERROR, "Calculated checksum did not match S3 checksum");
+			}
 		}
-		
+
 		exit_();		
 	}
 	
@@ -443,7 +463,8 @@
 	s3_server(NULL),
 	s3_public_key(NULL),
 	s3_private_key(NULL),
-	s3_maxRetrys(5)
+	s3_maxRetries(5),
+	s3_sleepTime(0)
 {
 	new_(s3_server, CSStringBuffer());
 	s3_server->append("s3.amazonaws.com/");
@@ -524,6 +545,29 @@
 // CURL callback functions:
 ////////////////////////////
 //----------------------
+//-----------------
+static bool try_ReadStream(CSThread *self, S3ProtocolCon *con, unsigned char *ptr, size_t buffer_size, size_t *data_sent)
+{
+	volatile bool rtc = true;
+	try_(a) {
+		*data_sent = con->ms_inputStream->read((char*)ptr, buffer_size);
+		if (*data_sent <= con->ms_data_size) {
+			con->ms_data_size -= *data_sent;
+			if (*data_sent)
+				con->ms_md5.md5_append(ptr, *data_sent); // Calculating the checksum for the data sent.
+		} else if (*data_sent > con->ms_data_size) 
+			CSException::RecordException(CS_CONTEXT, CS_ERR_GENERIC_ERROR, "Blob larger than expected.");
+		else if (con->ms_data_size && !*data_sent)
+			CSException::RecordException(CS_CONTEXT, CS_ERR_GENERIC_ERROR, "Blob smaller than expected.");
+		rtc = false;
+	}
+	
+	catch_(a)
+	cont_(a);
+	return rtc;
+}
+
+//----------------------
 static size_t send_callback(void *ptr, size_t objs, size_t obj_size, void *v_con)
 {
 	S3ProtocolCon *con = (S3ProtocolCon*) v_con;
@@ -533,50 +577,63 @@
 		return 0;
 		
 	enter_();
-	try_(a) {
-		data_sent = con->ms_inputStream->read((char*)ptr, buffer_size);
-		if (data_sent <= con->ms_data_size) {
-			con->ms_data_size -= data_sent;
-			if (data_sent)
-				con->ms_md5.md5_append((u_char*)ptr, data_sent); // Calculating the checksum for the data sent.
-		} else if (data_sent > con->ms_data_size) 
-			CSException::RecordException(CS_CONTEXT, CS_ERR_GENERIC_ERROR, "Blob larger than expected.");
-		else if (con->ms_data_size && !data_sent)
-			CSException::RecordException(CS_CONTEXT, CS_ERR_GENERIC_ERROR, "Blob smaller than expected.");
+	if (try_ReadStream(self, con, (unsigned char*)ptr, buffer_size, &data_sent)) {
+		con->ms_throw_error = true;
+		data_sent = (size_t)-1;
 	}
-	catch_(a);
-	con->ms_throw_error = true;
-	data_sent = SIZE_MAX;
-	
-	cont_(a);
-	
-	return data_sent;
+	
+	return_(data_sent);
 }
-//----------------------
-static size_t receive_data(void *vptr, size_t objs, size_t obj_size, void *v_con)
+
+//-----------------
+static bool try_WriteStream(CSThread *self, S3ProtocolCon *con, char *ptr, size_t data_len)
 {
-	S3ProtocolCon *con = (S3ProtocolCon*) v_con;
-	size_t data_len = objs * obj_size;
-
-	enter_();
+	volatile bool rtc = true;
 	try_(a) {
 		if (con->ms_replyStatus >= 400) { // Collect the error reply.
 			if (!con->ms_errorReply)
 				con->ms_errorReply = new CSStringBuffer(50);		
-			con->ms_errorReply->append((char*)vptr, data_len);
+			con->ms_errorReply->append(ptr, data_len);
 		} else if (	con->ms_outputStream)
-			con->ms_outputStream->write((char*)vptr, data_len);
+			con->ms_outputStream->write(ptr, data_len);
+		rtc = false;
 	}
-	catch_(a);
-	con->ms_throw_error = true;
-	data_len = SIZE_MAX; 
 	
+	catch_(a)
 	cont_(a);
+	return rtc;
+}
+
+//----------------------
+static size_t receive_data(void *vptr, size_t objs, size_t obj_size, void *v_con)
+{
+	S3ProtocolCon *con = (S3ProtocolCon*) v_con;
+	size_t data_len = objs * obj_size;
+
+	enter_();
+	if (try_WriteStream(self, con, (char*)vptr, data_len)) {
+		con->ms_throw_error = true;
+		data_len = (size_t)-1;
+	}
+
 	return_(data_len);	
 }
 
 #define IS_REDIRECT(s) ((s >= 300) && (s < 400))
 //----------------------
+static bool try_addHeader(CSThread *self, S3ProtocolCon *con, char *name, uint32_t name_len, char *value, uint32_t value_len)
+{
+	try_(a) {
+		con->ms_reply_headers.addHeader(name, name_len, value, value_len);
+		return false;
+	}
+	
+	catch_(a);
+	cont_(a);
+	return true;
+}
+
+//----------------------
 static size_t receive_header(void *header, size_t objs, size_t obj_size, void *v_con)
 {
 	S3ProtocolCon *con = (S3ProtocolCon*) v_con;
@@ -584,13 +641,6 @@
 	char *end, *ptr = (char*) header, *name, *value = NULL;
 	uint32_t name_len =0, value_len = 0;
 	
-	CLOBBER_PROTECT(con);
-	CLOBBER_PROTECT(size);
-	CLOBBER_PROTECT(ptr);
-	CLOBBER_PROTECT(value);
-	CLOBBER_PROTECT(value_len);
-	CLOBBER_PROTECT(name_len);
-
 //printf(	"receive_header: %s\n", ptr);
 	end = ptr + size;
 	if (*(end -2) == '\r' && *(end -1) == '\n')
@@ -638,7 +688,9 @@
 	while (value[value_len-1] == ' ') value_len--;
 	
 	if (!strncasecmp(name, "ETag", 4)) {
-		value++; value_len -=2; // Strip quotation marks from checksum string.
+		if (*value == '"') {
+			value++; value_len -=2; // Strip quotation marks from checksum string.
+		}
 		if (value_len == HEX_CHECKSUM_VALUE_SIZE) {
 			memcpy(con->ms_s3Checksum, value, value_len);
 			con->ms_s3Checksum[value_len] = 0;
@@ -646,15 +698,10 @@
 	}
 	
 	enter_();
-	try_(a) {
-		con->ms_reply_headers.addHeader(name, name_len, value, value_len);
+	if (try_addHeader(self, con, name, name_len, value, value_len)) {
+		con->ms_throw_error = true;
+		size = (size_t)-1;
 	}
-	
-	catch_(a);
-	con->ms_throw_error = true;
-	size = (size_t)-1;
-		
-	cont_(a);
 	return_(size);
 }
 
@@ -716,11 +763,12 @@
 	con_data->ms_execute_delete_request();
 	
 	if (con_data->ms_retry) {
-		if (retry_count == s3_maxRetrys) {
+		if (retry_count == s3_maxRetries) {
 			CSException::throwException(CS_CONTEXT, CS_ERR_GENERIC_ERROR, "S3 operation aborted after max retries.");
 		}
-	printf("RETRY: s3_delete()\n");
+	//printf("RETRY: s3_delete()\n");
 		retry_count++;
+		self->sleep(s3_sleepTime);
 		goto retry;
 	}
 	
@@ -815,11 +863,12 @@
 	}
 	
 	if (con_data->ms_retry) {
-		if (retry_count == s3_maxRetrys) {
+		if (retry_count == s3_maxRetries) {
 			CSException::throwException(CS_CONTEXT, CS_ERR_GENERIC_ERROR, "S3 operation aborted after max retries.");
 		}
-	printf("RETRY: s3_copy()\n");
+	//printf("RETRY: s3_copy()\n");
 		retry_count++;
+		self->sleep(s3_sleepTime);
 		goto retry;
 	}
 	
@@ -831,7 +880,7 @@
 
 
 //-------------------------------
-CSVector *CSS3Protocol::s3_receive(CSOutputStream *output, const char *bucket, const char *key, bool *found)
+CSVector *CSS3Protocol::s3_receive(CSOutputStream *output, const char *bucket, const char *key, bool *found, S3RangePtr range, time_t *last_modified)
 {
  	CSStringBuffer *s3_buffer;
     char date[64];
@@ -839,10 +888,16 @@
 	uint32_t retry_count = 0;
 	S3ProtocolCon *con_data;
 	CSVector *replyHeaders;
+	CSString *range_header = NULL;
+	const char *http_op;
 
 	enter_();
 
-	push_(output);
+	if (output) {
+		push_(output);
+		http_op = "GET";
+	} else
+		http_op = "HEAD";
 
 	new_(s3_buffer, CSStringBuffer());
 	push_(s3_buffer);
@@ -872,8 +927,16 @@
 	s3_buffer->append(date);
 	con_data->ms_setHeader(s3_buffer->getCString());
 
+	if (range) {
+		char buffer[80];
+		snprintf(buffer, 80,"Range: bytes=%"PRIu64"-%"PRIu64, range->startByte, range->endByte);
+
+		range_header = CSString::newString(buffer);
+	}
 	// Create the authentication signature and add the 'Authorization' header
-	signed_str = s3_getSignature("GET", NULL, NULL, date, bucket, key);
+	if (range_header)
+		con_data->ms_setHeader(range_header->getCString());
+	signed_str = s3_getSignature(http_op, NULL, NULL, date, bucket, key, NULL);
 	push_(signed_str);
 	s3_buffer->setLength(0);
 	s3_buffer->append("Authorization: AWS ");	
@@ -883,23 +946,28 @@
 	release_(signed_str); signed_str = NULL;
 	con_data->ms_setHeader(s3_buffer->getCString());
 	
-	con_data->ms_execute_get_request(RETAIN(output));
+	if (output) output->retain();
+	con_data->ms_execute_get_request(output);
 	
 	if (con_data->ms_retry) {
-		if (retry_count == s3_maxRetrys) {
+		if (retry_count == s3_maxRetries) {
 			CSException::throwException(CS_CONTEXT, CS_ERR_GENERIC_ERROR, "S3 operation aborted after max retries.");
 		}
-	printf("RETRY: s3_receive()\n");
+	//printf("RETRY: s3_receive()\n");
 		retry_count++;
 		output->reset();
+		self->sleep(s3_sleepTime);
 		goto retry;
 	}
 	
+	if (last_modified)
+		*last_modified = con_data->ms_last_modified;
 	*found = !con_data->ms_notFound;
 	replyHeaders = con_data->ms_reply_headers.takeHeaders();
 	release_(con_data);
 	release_(s3_buffer);
-	release_(output);
+	if (output)
+		release_(output);
 	
 	return_(replyHeaders);
 }
@@ -1000,6 +1068,8 @@
 	s3_buffer->append(bucket);
 	s3_buffer->append(".");	
 	s3_buffer->append(s3_server->getCString());
+//s3_buffer->append("/");	
+//s3_buffer->append(bucket);
 	if (key_prefix) {
 		s3_buffer->append("?prefix=");
 		s3_buffer->append(key_prefix);
@@ -1035,12 +1105,13 @@
 	con_data->ms_execute_get_request(RETAIN(output));
 	
 	if (con_data->ms_retry) {
-		if (retry_count == s3_maxRetrys) {
+		if (retry_count == s3_maxRetries) {
 			CSException::throwException(CS_CONTEXT, CS_ERR_GENERIC_ERROR, "S3 operation aborted after max retries.");
 		}
-	printf("RETRY: s3_list()\n");
+	//printf("RETRY: s3_list()\n");
 		retry_count++;
 		output->reset();
+		self->sleep(s3_sleepTime);
 		goto retry;
 	}
 	
@@ -1077,10 +1148,10 @@
  	CSStringBuffer *s3_buffer;
     char date[64];
 	CSString *signed_str;
-	Md5Digest dummy_digest;
 	uint32_t retry_count = 0;
 	S3ProtocolCon *con_data;
 	CSVector *replyHeaders;
+	char checksum[32], *md5 = NULL;
 
 	enter_();
 	push_(input);
@@ -1090,9 +1161,6 @@
 
 	new_(con_data, S3ProtocolCon());
 	push_(con_data);
-
-	if (!digest)
-		digest = &dummy_digest;
 		
 	if (!content_type)
 		content_type = "binary/octet-stream";
@@ -1130,9 +1198,24 @@
 	s3_buffer->append(content_type);
 	con_data->ms_setHeader(s3_buffer->getCString());
 		
+	if (digest) {
+		// Add the Md5 checksum header
+		md5 = checksum;
+		memset(checksum, 0, 32);
+		base64Encode(digest->val, 16, checksum, 32);
+		
+		s3_buffer->setLength(0);
+		s3_buffer->append("Content-MD5: ");	
+		s3_buffer->append(checksum);
+		con_data->ms_setHeader(s3_buffer->getCString());		
+		con_data->ms_calculate_md5 = false;
+	} else 
+		con_data->ms_calculate_md5 = true;
+	
+
 	// Create the authentication signature and add the 'Authorization' header
 	if (!s3Authorization)
-		signed_str = s3_getSignature("PUT", NULL, content_type, date, bucket, key);
+		signed_str = s3_getSignature("PUT", md5, content_type, date, bucket, key);
 	else
 		signed_str = CSString::newString(s3Authorization);
 	push_(signed_str);
@@ -1144,19 +1227,21 @@
 	release_(signed_str); signed_str = NULL;
 	con_data->ms_setHeader(s3_buffer->getCString());
 	
-	con_data->ms_execute_put_request(RETAIN(input), size, digest);
+	con_data->ms_execute_put_request(RETAIN(input), size);
 	
 	if (con_data->ms_retry) {
-		if (retry_count == s3_maxRetrys) {
+		if (retry_count == s3_maxRetries) {
 			CSException::throwException(CS_CONTEXT, CS_ERR_GENERIC_ERROR, "S3 operation aborted after max retries.");
 		}
-	printf("RETRY: s3_send()\n");
+	//printf("RETRY: s3_send()\n");
 		retry_count++;
 		input->reset();
+		self->sleep(s3_sleepTime);
 		goto retry;
 	}
 	
 	replyHeaders = con_data->ms_reply_headers.takeHeaders();
+
 	release_(con_data);
 	release_(s3_buffer);
 	release_(input);
@@ -1211,6 +1296,7 @@
 	printf("Delete object:\n\t%s d <bucket> <object_key>\n", cmd);
 	printf("Delete all object with a given prefix:\n\t%s D <bucket> <object_prefix>\n", cmd);
 	printf("Get object, data will be written to 'prottest.out':\n\t%s g <bucket> <object_key> <timeout>\n", cmd);
+	printf("Get object header only:\n\t%s h <bucket> <object_key> <timeout>\n", cmd);
 	printf("Put (Upload) an object:\n\t%s p <bucket> <object_key> <file>\n", cmd);
 	printf("List objects in the bucket:\n\t%s l <bucket> [<object_prefix> [max_list_size]]\n", cmd);
 	printf("Copy object:\n\t%s c <src_bucket> <src_object_key> <dst_bucket> <dst_object_key> \n", cmd);
@@ -1240,6 +1326,7 @@
 	CSThread *main_thread;
 	const char *pub_key;
 	const char *priv_key;
+	const char *server;
 	CSS3Protocol *prot = NULL;
 	
 	if (argc < 3) {
@@ -1265,7 +1352,10 @@
 		new_(prot, CSS3Protocol());
 		push_(prot);
 		
-		prot->s3_setServer("s3.amazonaws.com/");
+		server = getenv("S3_SERVER");
+		if ((server == NULL) || (*server == 0))
+			server = "s3.amazonaws.com/";
+		prot->s3_setServer(server);
 		prot->s3_setPublicKey(pub_key);
 		prot->s3_setPrivateKey(priv_key);
 		
@@ -1308,15 +1398,22 @@
 				
 				break;
 			case 'g':  // Get the object
-				if (argc == 4) {
+				if ((argc == 4) || (argc == 6)) {
 					CSFile *output;	
 					CSVector *headers;
 					bool found;				
+					S3RangeRec *range_ptr = NULL, range = 	{0,0};		
+					
+					if (argc == 6) {
+						range.startByte = atoi(argv[4]);
+						range.endByte = atoi(argv[5]);
+						range_ptr = &range;
+					}
 					
 					output = CSFile::newFile("prottest.out");
 					push_(output);
 					output->open(CSFile::CREATE | CSFile::TRUNCATE);
-					headers = prot->s3_receive(output->getOutputStream(), argv[2], argv[3], &found);
+					headers = prot->s3_receive(output->getOutputStream(), argv[2], argv[3], &found, range_ptr);
 					if (!found)
 						printf("%s/%s could not be found.\n", argv[2], argv[3]);
 						
@@ -1328,15 +1425,34 @@
 				
 				break;
 				
+			case 'h':  // Get the object header
+				if (argc == 4) {
+					CSVector *headers;
+					bool found;	
+					S3RangeRec range = 	{0,0};		
+					
+					headers = prot->s3_receive(NULL, argv[2], argv[3], &found);
+					if (!found)
+						printf("%s/%s could not be found.\n", argv[2], argv[3]);
+						
+					dump_headers(headers);
+						
+				} else
+					printf("Bad command: h <bucket> <object_key>\n");
+				
+				break;
+				
 			case 'p':  // Put (Upload) the object
 				if (argc == 5) {
 					CSFile *input;
+					Md5Digest digest;
 					CSVector *headers;
 					
 					input = CSFile::newFile(argv[4]);
 					push_(input);
 					input->open(CSFile::READONLY);
-					headers = prot->s3_send(input->getInputStream(), argv[2], argv[3], input->myFilePath->getSize());
+					input->md5Digest(&digest);
+					headers = prot->s3_send(input->getInputStream(), argv[2], argv[3], input->myFilePath->getSize(), NULL, &digest);
 					dump_headers(headers);
 					release_(input);
 				} else

=== modified file 'plugin/pbms/src/cslib/CSS3Protocol.h'
--- plugin/pbms/src/cslib/CSS3Protocol.h	2010-10-02 21:15:42 +0000
+++ plugin/pbms/src/cslib/CSS3Protocol.h	2010-10-18 21:21:20 +0000
@@ -23,6 +23,7 @@
  */
 
 #include "CSHTTPStream.h"
+#include "CSMd5.h"
 class S3ProtocolCon;
 
 #define PUB_KEY_BIT		((uint8_t)1)
@@ -30,6 +31,11 @@
 #define SERVER_BIT		((uint8_t)4)
 
 
+typedef struct S3Range {
+	 off64_t startByte;
+	 off64_t endByte;
+} S3RangeRec, *S3RangePtr;
+
 class CSS3Protocol: public CSRefObject {
 	private:
 	
@@ -39,7 +45,8 @@
 	CSString *s3_private_key;
 	
 	uint8_t	s3_ready; // A flag to indicate if the S3 protocol has been fully initialised.
-	uint32_t s3_maxRetrys; 
+	uint32_t s3_maxRetries; 
+	uint32_t s3_sleepTime; 
 	
 	CSString *s3_getSignature(const char *verb, 
 								const char *md5, 
@@ -95,6 +102,9 @@
 		}
 	}
 
+	void s3_setMaxRetries(uint32_t retries){s3_maxRetries = retries;}
+	void s3_setSleepTime(uint32_t nap_time){s3_sleepTime = nap_time;}
+	
 	const char *s3_getPrivateKey() { return s3_private_key->getCString();}
 	
 	bool s3_isReady() { return ((s3_ready & SERVER_BIT) && (s3_ready & PUB_KEY_BIT) && (s3_ready & PRIV_KEY_BIT));}
@@ -105,7 +115,8 @@
 	CSVector *s3_send(CSInputStream *input, const char *bucket, const char *key, off64_t size, const char *content_type = NULL, Md5Digest *digest = NULL, const char *s3Authorization = NULL, time_t s3AuthorizationTime = 0);
 
 	// s3_receive() returns false if the object was not found.
-	CSVector *s3_receive(CSOutputStream *output, const char *bucket, const char *key, bool *found);
+	// If 'output' is NULL then only the headers will be fetched.
+	CSVector *s3_receive(CSOutputStream *output, const char *bucket, const char *key, bool *found, S3RangePtr range = NULL, time_t *last_modified = NULL);
 
 	// s3_delete() returns false if the object was not found.
 	bool s3_delete(const char *bucket, const char *key);

=== modified file 'plugin/pbms/src/cslib/CSSha1.cc'
--- plugin/pbms/src/cslib/CSSha1.cc	2010-10-02 21:15:42 +0000
+++ plugin/pbms/src/cslib/CSSha1.cc	2010-10-18 21:21:20 +0000
@@ -58,6 +58,7 @@
 
 #include "CSConfig.h"
 #include <string.h>
+#include <stdint.h>
 
 #include "CSDefs.h"
 #include "CSObject.h"

=== modified file 'plugin/pbms/src/cslib/CSSocket.cc'
--- plugin/pbms/src/cslib/CSSocket.cc	2010-10-02 21:15:42 +0000
+++ plugin/pbms/src/cslib/CSSocket.cc	2010-10-18 21:21:20 +0000
@@ -30,12 +30,34 @@
 
 #include <stdio.h>
 #include <sys/types.h>
+
+#ifdef OS_WINDOWS
+#include <winsock.h>
+typedef int socklen_t;
+#define SHUT_RDWR 2
+#define CLOSE_SOCKET(s)	closesocket(s)
+#define IOCTL_SOCKET	ioctlsocket
+#define SOCKET_ERRORNO	WSAGetLastError()
+#define	EWOULDBLOCK		WSAEWOULDBLOCK
+
+#else
+#include <sys/ioctl.h>
 #include <sys/socket.h>
 #include <netdb.h>
 #include <netinet/in.h>
 #include <arpa/inet.h>
 #include <netinet/in.h>
 #include <netinet/tcp.h>
+#include <sys/select.h>
+#include <fcntl.h>
+
+extern void unix_close(int h);
+#define CLOSE_SOCKET(s)	unix_close(s)
+#define IOCTL_SOCKET	ioctl
+#define SOCKET_ERRORNO	errno
+
+#endif
+
 #include <ctype.h>
 #include <string.h>
 #include <stdlib.h>
@@ -46,6 +68,27 @@
 #include "CSStrUtil.h"
 #include "CSFile.h"
 
+void CSSocket::initSockets()
+{
+#ifdef OS_WINDOWS
+	int		err;
+	WSADATA	data;
+	WORD	version = MAKEWORD (1,1);
+	static int inited = 0;
+
+	if (!inited) {
+		err = WSAStartup(version, &data);
+
+		if (err != 0) {
+			CSException::throwException(CS_CONTEXT, err, "WSAStartup error");
+		}
+		
+		inited = 1;
+	}
+	
+#endif
+}
+
 /*
  * ---------------------------------------------------------------
  * CORE SYSTEM SOCKET FACTORY
@@ -53,10 +96,10 @@
 
 CSSocket *CSSocket::newSocket()
 {
-	SCSocket *s;
+	CSSocket *s;
 	
-	new_(s, SCSocket());
-	return (CSSocket *) s;
+	new_(s, CSSocket());
+	return s;
 }
 
 /*
@@ -64,7 +107,7 @@
  * INTERNAL UTILITIES
  */
 
-void SCSocket::formatAddress(size_t size, char *buffer)
+void CSSocket::formatAddress(size_t size, char *buffer)
 {
 	if (iHost) {
 		cs_strcpy(size, buffer, iHost);
@@ -77,7 +120,7 @@
 		cs_strcat(size, buffer, iService);
 }
 
-void SCSocket::throwError(const char *func, const char *file, int line, char *address, int err)
+void CSSocket::throwError(const char *func, const char *file, int line, char *address, int err)
 {
 	if (err)
 		CSException::throwFileError(func, file, line, address, err);
@@ -85,7 +128,7 @@
 		CSException::throwEOFError(func, file, line, address);
 }
 
-void SCSocket::throwError(const char *func, const char *file, int line, int err)
+void CSSocket::throwError(const char *func, const char *file, int line, int err)
 {
 	char address[CS_SOCKET_ADDRESS_SIZE];
 
@@ -93,20 +136,106 @@
 	throwError(func, file, line, address, err);
 }
 
-void SCSocket::setInternalOptions()
+void CSSocket::setNoDelay()
 {
 	int flag = 1;
 
 	if (setsockopt(iHandle, IPPROTO_TCP, TCP_NODELAY, (char *) &flag, sizeof(int)) == -1)
-		CSException::throwOSError(CS_CONTEXT, errno);
-}
-
-void SCSocket::openInternal()
-{
-	iHandle = socket(AF_INET, SOCK_STREAM, 0);
+		CSException::throwOSError(CS_CONTEXT, SOCKET_ERRORNO);
+}
+
+void CSSocket::setNonBlocking()
+{
+	if (iTimeout) {
+		unsigned long block = 1;
+
+		if (IOCTL_SOCKET(iHandle, FIONBIO, &block) != 0)
+			throwError(CS_CONTEXT, SOCKET_ERRORNO);
+	}
+}
+
+void CSSocket::setBlocking()
+{
+	/* No timeout, set blocking: */
+	if (!iTimeout) {
+		unsigned long block = 0;
+
+		if (IOCTL_SOCKET(iHandle, FIONBIO, &block) != 0)
+			throwError(CS_CONTEXT, SOCKET_ERRORNO);
+	}
+}
+
+void CSSocket::openInternal()
+{
+	iHandle = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP);
 	if (iHandle == -1)
-		CSException::throwOSError(CS_CONTEXT, errno);
-	setInternalOptions();
+		CSException::throwOSError(CS_CONTEXT, SOCKET_ERRORNO);
+	setNoDelay();
+	setNonBlocking();
+}
+
+void CSSocket::writeBlock(const void *data, size_t len)
+{
+	ssize_t	out;
+
+	enter_();
+	while (len > 0) {
+		out = send(iHandle, (const char *) data, len, 0);
+		self->interrupted();
+		if (out == -1) {
+			int err = SOCKET_ERRORNO;
+
+			if (err == EWOULDBLOCK || err == EINTR)
+				continue;
+			throwError(CS_CONTEXT, err);
+		}
+		if ((size_t) out > len)
+			break;
+		len -= (size_t) out;
+		data = ((char *) data) + (size_t) out;
+	}
+	exit_();
+}
+
+int CSSocket::timeoutRead(CSThread *self, void *buffer, size_t length)
+{      
+	int			in;
+	uint64_t	start_time;
+	uint64_t	timeout = iTimeout * 1000;
+	
+	start_time = CSTime::getTimeCurrentTicks();
+
+	retry:
+	in = recv(iHandle, (char *) buffer, length, 0);
+	if (in == -1) {
+		if (SOCKET_ERRORNO == EWOULDBLOCK) {
+			fd_set			readfds;
+			uint64_t		time_diff;
+			struct timeval	tv_timeout;
+
+			FD_ZERO(&readfds);
+			self->interrupted();
+
+			time_diff = CSTime::getTimeCurrentTicks() - start_time;
+			if (time_diff >= timeout) {
+				char address[CS_SOCKET_ADDRESS_SIZE];
+
+				formatAddress(CS_SOCKET_ADDRESS_SIZE, address);
+				CSException::throwExceptionf(CS_CONTEXT, CS_ERR_RECEIVE_TIMEOUT, "Receive timeout: %lu ms, on: %s", iTimeout, address);
+			}
+
+			/* Calculate how much time we can wait: */
+			time_diff = timeout - time_diff;
+			tv_timeout.tv_sec = (long)time_diff / 1000000;
+			tv_timeout.tv_usec = (long)time_diff % 1000000;
+
+			FD_SET(iHandle, &readfds);
+			in = select(iHandle+1, &readfds, NULL, NULL, &tv_timeout);
+			if (in != -1)
+				goto retry;
+		}
+	}
+ 	return in;
 }
 
 /*
@@ -114,6 +243,16 @@
  * SOCKET BASED ON THE STANDARD C SOCKET
  */
 
+void CSSocket::setTimeout(uint32_t milli_sec)
+{
+	if (iTimeout != milli_sec) {
+		if ((iTimeout = milli_sec))
+			setNonBlocking();
+		else
+			setBlocking();
+	}
+}
+
 CSOutputStream *CSSocket::getOutputStream()
 {
 	return CSSocketOutputStream::newStream(RETAIN(this));
@@ -124,7 +263,7 @@
 	return CSSocketInputStream::newStream(RETAIN(this));
 }
 
-void SCSocket::publish(char *service, int default_port)
+void CSSocket::publish(char *service, int default_port)
 {
 	enter_();
 	close();
@@ -171,13 +310,13 @@
 		server.sin_port = (uint16_t) servp->s_port;
 
 		if (setsockopt(iHandle, SOL_SOCKET, SO_REUSEADDR, (char *) &flag, sizeof(int)) == -1)
-			CSException::throwOSError(CS_CONTEXT, errno);
+			CSException::throwOSError(CS_CONTEXT, SOCKET_ERRORNO);
 
 		if (bind(iHandle, (struct sockaddr *) &server, sizeof(server)) == -1)
-			CSException::throwOSError(CS_CONTEXT, errno);
+			CSException::throwOSError(CS_CONTEXT, SOCKET_ERRORNO);
 
 		if (listen(iHandle, SOMAXCONN) == -1)
-			CSException::throwOSError(CS_CONTEXT, errno);
+			CSException::throwOSError(CS_CONTEXT, SOCKET_ERRORNO);
 	}
 	catch_(a) {
 		close();
@@ -187,7 +326,7 @@
 	exit_();
 }
 
-void SCSocket::open(CSSocket *listener)
+void CSSocket::open(CSSocket *listener)
 {
 	enter_();
 
@@ -199,7 +338,7 @@
 		socklen_t			addrlen = sizeof(remote);
 
 		/* First get all the information we need from the listener: */
-		listener_handle = ((SCSocket *) listener)->iHandle;
+		listener_handle = ((CSSocket *) listener)->iHandle;
 		listener->formatAddress(CS_SOCKET_ADDRESS_SIZE, address);
 
 		/* I want to make sure no error occurs after the connect!
@@ -211,12 +350,13 @@
 		iHost = (char *) cs_malloc(100);
 		iHandle = accept(listener_handle, (struct sockaddr *) &remote, &addrlen);
 		if (iHandle == -1)
-			throwError(CS_CONTEXT, address, errno);
+			throwError(CS_CONTEXT, address, SOCKET_ERRORNO);
 
 		cs_strcpy(100, iHost, inet_ntoa(remote.sin_addr));
 		iPort = ntohs(remote.sin_port);
 
-		setInternalOptions();
+		setNoDelay();
+		setNonBlocking();
 	}
 	catch_(a) {
 		close();
@@ -226,7 +366,7 @@
 	exit_();
 }
 
-void SCSocket::open(char *address, int default_port)
+void CSSocket::open(char *address, int default_port)
 {
 	enter_();
 	close();
@@ -269,7 +409,7 @@
 		memcpy(&server.sin_addr, hostp->h_addr, (size_t) hostp->h_length);
 		server.sin_port = (uint16_t) servp->s_port;
 		if (connect(iHandle, (struct sockaddr *) &server, sizeof(server)) == -1)
-			throwError(CS_CONTEXT, errno);
+			throwError(CS_CONTEXT, SOCKET_ERRORNO);
 	}
 	catch_(a) {
 		close();
@@ -279,12 +419,13 @@
 	exit_();
 }
 
-void SCSocket::close()
+void CSSocket::close()
 {
+	flush();
 	if (iHandle != -1) {
 		shutdown(iHandle, SHUT_RDWR);
 		/* shutdown does not close the socket!!? */
-		unix_file_close(iHandle);
+		CLOSE_SOCKET(iHandle);
 		iHandle = -1;
 	}
 	if (iHost) {
@@ -295,10 +436,14 @@
 		cs_free(iService);
 		iService = NULL;
 	}
+	if (iIdentity) {
+		cs_free(iIdentity);
+		iIdentity = NULL;
+	}
 	iPort = 0;
 }
 
-size_t SCSocket::read(void *data, size_t len)
+size_t CSSocket::read(void *data, size_t len)
 {
 	ssize_t in;
 
@@ -308,21 +453,27 @@
 	 * So a return of zero means EOF!
 	 */
 	retry:
-	in = recv(iHandle, data, len, 0);
+	if (iTimeout)
+		in = timeoutRead(self, data, len);
+	else
+		in = recv(iHandle, (char *) data, len, 0);
 	self->interrupted();
 	if (in == -1) {
 		/* Note, we actually ignore all errors on the socket.
 		 * If no data was returned by the read so far, then
 		 * the error will be considered EOF.
 		 */
-		if (errno == EAGAIN || errno == EINTR)
+		int err = SOCKET_ERRORNO;
+
+		if (err == EWOULDBLOCK || err == EINTR)
 			goto retry;
+		throwError(CS_CONTEXT, err);
 		in = 0;
 	}
 	return_((size_t) in);
 }
 
-int SCSocket::read()
+int CSSocket::read()
 {
 	int		ch;
 	u_char	buffer[1];
@@ -335,43 +486,78 @@
 	return_(ch);
 }
 
-int SCSocket::peek()
+int CSSocket::peek()
 {
 	return -1;
 }
 
-void SCSocket::write(const void *data, size_t len)
-{
-	ssize_t	out;
-
-	enter_();
-	while (len > 0) {
-		out = send(iHandle, data, len, 0);
-		self->interrupted();
-		if (out == -1) {
-			int err = errno;
-
-			if (err == EAGAIN || errno == EINTR)
-				continue;
-			throwError(CS_CONTEXT, err);
-		}
-		if ((size_t) out > len)
-			break;
-		len -= (size_t) out;
-		data = ((char *) data) + (size_t) out;
-	}
-	exit_();
-}
-
-void SCSocket::write(char ch)
-{
-	enter_();
-	write(&ch, 1);
-	exit_();
-}
-
-void SCSocket::flush()
-{
-}
+void CSSocket::write(const void *data, size_t len)
+{
+#ifdef CS_USE_OUTPUT_BUFFER
+	if (len <= CS_MIN_WRITE_SIZE) {
+		if (iDataLen + len > CS_OUTPUT_BUFFER_SIZE) {
+			/* This is the amount of data that will still fit
+			 * intp the buffer:
+			 */
+			size_t tfer = CS_OUTPUT_BUFFER_SIZE - iDataLen;
+
+			memcpy(iOutputBuffer + iDataLen, data, tfer);
+			flush();
+			len -= tfer;
+			memcpy(iOutputBuffer, ((char *) data) + tfer, len);
+			iDataLen = len;
+		}
+		else {
+			memcpy(iOutputBuffer + iDataLen, data, len);
+			iDataLen += len;
+		}
+	}
+	else {
+		/* If the block give is large enough, the
+		 * writing directly from the block saves copying the
+		 * data to the local output buffer buffer:
+		 */
+		flush();
+		writeBlock(data, len);
+	}
+#else
+	writeBlock(data, len);
+#endif
+}
+
+void CSSocket::write(char ch)
+{
+	enter_();
+	writeBlock(&ch, 1);
+	exit_();
+}
+
+void CSSocket::flush()
+{
+#ifdef CS_USE_OUTPUT_BUFFER
+	uint32_t len;
+
+	if ((len = iDataLen)) {
+		iDataLen = 0;
+		/* Note: we discard the data to be written if an
+		 * exception occurs.
+		 */
+		writeBlock(iOutputBuffer, len);
+	}
+#endif
+}
+
+const char *CSSocket::identify()
+{
+	enter_();
+	if (!iIdentity) {
+		char buffer[200];
+
+		formatAddress(200, buffer);
+		iIdentity = cs_strdup(buffer);
+	}
+	return_(iIdentity);
+}
+
 
 

=== modified file 'plugin/pbms/src/cslib/CSSocket.h'
--- plugin/pbms/src/cslib/CSSocket.h	2010-10-02 21:15:42 +0000
+++ plugin/pbms/src/cslib/CSSocket.h	2010-10-18 21:21:20 +0000
@@ -37,42 +37,65 @@
 #include "CSMemory.h"
 #include "CSMutex.h"
 
-using namespace std;
-
 class CSOutputStream;
 class CSInputStream;
 
+#define CS_SOCKET_ADDRESS_SIZE		300
+
+/* This is only required if you do
+ * not use an output buffer stream!
+ */
+//#define CS_USE_OUTPUT_BUFFER
+
+#ifdef CS_USE_OUTPUT_BUFFER
+#ifdef DEBUG
+#define CS_OUTPUT_BUFFER_SIZE		80
+#define CS_MIN_WRITE_SIZE			40
+#else
+#define CS_OUTPUT_BUFFER_SIZE		(4*1024)
+#define CS_MIN_WRITE_SIZE			(1500)
+#endif
+#endif
+
 class CSSocket : public CSRefObject {
 public:
-	CSSocket() { }
-
-	virtual ~CSSocket() { }
+	CSSocket(): iHandle(-1), iHost(NULL), iService(NULL), iIdentity(NULL), iPort(0), iTimeout(0) {
+#ifdef CS_USE_OUTPUT_BUFFER
+	iDataLen = 0;
+#endif
+	}
+
+	virtual ~CSSocket() {
+		close();
+	}
+
+	void setTimeout(uint32_t milli_sec);
 
 	CSOutputStream *getOutputStream();
 
 	CSInputStream *getInputStream();
 
-	virtual void formatAddress(size_t size, char *address) = 0;
+	virtual void formatAddress(size_t size, char *address);
 
 	/*
 	 * Publish a listener:
 	 */
-	virtual void publish(char *service, int default_port) = 0;
+	virtual void publish(char *service, int default_port);
 
 	/*
 	 * Accept a connection from a listening socket:
 	 */
-	virtual void open(CSSocket *listener) = 0;
+	virtual void open(CSSocket *listener);
 
 	/*
 	 * Connect to a listening socket.
 	 */
-	virtual void open(char *address, int default_port) = 0;
+	virtual void open(char *address, int default_port);
 
 	/*
 	 * Close the socket.
 	 */
-	virtual void close() { }
+	virtual void close();
 
 	/*
 	 * Read at least one byte from the socket.
@@ -84,7 +107,7 @@
 	 * Note: Errors on the socket do not cause
 	 * an exception!
 	 */
-	virtual size_t read(void *data, size_t size) = 0;
+	virtual size_t read(void *data, size_t size);
 
 	/*
 	 * Returns -1 on EOF!
@@ -92,79 +115,58 @@
 	 * Just like read, error on the socket do
 	 * not throw an exception.
 	 */
-	virtual int read() = 0;
+	virtual int read();
 
 	/*
 	 * Look at the next character in the file without
 	 * taking from the input.
 	 */
-	virtual int peek() = 0;
+	virtual int peek();
 
 	/*
 	 * Write the given number of bytes.
 	 * Throws IOException if an error occurs.
 	 */
-	virtual void write(const  void *data, size_t size) = 0;
+	virtual void write(const  void *data, size_t size);
 
 	/*
 	 * Write a character to the file.
 	 */
-	virtual void write(char ch) = 0;
+	virtual void write(char ch);
 
 	/*
 	 * Flush the data written.
 	 */
-	virtual void flush() = 0;
+	virtual void flush();
+
+	virtual const char *identify();
+
+	static void initSockets();
 
 	static CSSocket	*newSocket();
 
-	friend class SCSocket;
-
-private:
-};
-
-#define CS_SOCKET_ADDRESS_SIZE		300
-
-class SCSocket : public CSSocket {
-public:
-	SCSocket(): CSSocket(), iHandle(-1), iHost(NULL), iService(NULL), iPort(0) { }
-
-	virtual ~SCSocket() {
-		close();
-	}
-
-	virtual void formatAddress(size_t size, char *address);
-
-	virtual void publish(char *service, int default_port);
-
-	virtual void open(CSSocket *listener);
-
-	virtual void open(char *address, int default_port);
-
-	virtual void close();
-
-	virtual size_t read(void *data, size_t size);
-
-	virtual int read();
-
-	virtual int peek();
-
-	virtual void write(const void *data, size_t size);
-
-	virtual void write(char ch);
-
-	virtual void flush();
-
 private:
 	void throwError(const char *func, const char *file, int line, char *address, int err);
 	void throwError(const char *func, const char *file, int line, int err);
-	void setInternalOptions();
+	void setNoDelay();
+	void setNonBlocking();
+	void setBlocking();
 	void openInternal();
-
-	int iHandle;
-	char *iHost;
-	char *iService;
-	int iPort;
+	void writeBlock(const void *data, size_t len);
+	int timeoutRead(CSThread *self, void *buffer, size_t length);
+
+	int			iHandle;
+	char		*iHost;
+	char		*iService;
+	char		*iIdentity;
+	int			iPort;
+	uint32_t	iTimeout;
+
+#ifdef CS_USE_OUTPUT_BUFFER
+	char	iOutputBuffer[CS_OUTPUT_BUFFER_SIZE];
+	size_t	iDataLen;
+#endif
+
 };
 
 #endif

=== modified file 'plugin/pbms/src/cslib/CSStorage.cc'
--- plugin/pbms/src/cslib/CSStorage.cc	2010-10-02 21:15:42 +0000
+++ plugin/pbms/src/cslib/CSStorage.cc	2010-10-20 20:21:41 +0000
@@ -55,7 +55,16 @@
 void CSHashTable::setSize(uint32_t size)
 {
 	enter_();
-	cs_realloc((void **) &iTable, sizeof(CSObject *) * size);
+	if (size == 0) {
+		if (iTable) {
+			cs_free(iTable);
+			iTable = NULL;
+		}
+	}
+	else {
+		cs_realloc((void **) &iTable, sizeof(CSObject *) * size);
+		memset(iTable, 0, sizeof(CSObject *) * size);
+	}
 	iSize = size;
 	exit_();
 }
@@ -83,7 +92,7 @@
 	return NULL;
 }
 
-void CSHashTable::remove(CSObject *key)
+bool CSHashTable::remove(CSObject *key)
 {
 	uint32_t h = key->hashKey();
 	CSObject *item, *prev_item;
@@ -98,11 +107,12 @@
 			else
 				iTable[h % iSize] = item->getHashLink();
 			item->release();
-			break;
+			return true;
 		}
 		prev_item = item;
 		item = item->getHashLink();
 	}
+	return false;
 }
 
 void CSHashTable::clear()
@@ -182,8 +192,8 @@
 		
 	item = 	iList[idx];
 	
+	iInUse--;
 	memmove(&iList[idx], &iList[idx+1], (iInUse - idx) * sizeof(CSObject *));
-	iInUse--;
 	return item;
 }
 
@@ -193,8 +203,8 @@
 	uint32_t		idx;
 
 	if ((item = search(key, idx))) {
+		iInUse--;
 		memmove(&iList[idx], &iList[idx+1], (iInUse - idx) * sizeof(CSObject *));
-		iInUse--;
 		item->release();
 	}
 }
@@ -263,6 +273,30 @@
 		item->release();
 }
 
+void CSLinkedList::addBack(CSObject *item)
+{
+	if (iListBack != item) {
+		remove(item);
+		item->setNextLink(iListBack);
+		item->setPrevLink(NULL);
+		
+		if (iListBack)
+			iListBack->setPrevLink(item);
+		else
+			iListFront = item;
+		iListBack = item;
+		iSize++;
+	}
+	else
+		/* Must do this or I will have one reference too
+		 * many!
+		 * The input object was given to me referenced,
+		 * but I already have the object on my list, and
+		 * referenced!
+		 */
+		item->release();
+}
+
 bool CSLinkedList::remove(CSObject *item)
 {
 	bool on_list = false;
@@ -505,6 +539,7 @@
 {
 	uint32_t pos;
 
+	// If search fails then pos will be > iUsage
 	search(idx, pos);
 	return pos;
 }
@@ -655,8 +690,8 @@
 		CSOrderedListItemRec ir;
 
 		memcpy(&ir, item, sizeof(CSOrderedListItemRec));
+		iInUse--;
 		memmove(&iList[idx], &iList[idx+1], (iInUse - idx) * sizeof(CSOrderedListItemRec));
-		iInUse--;
 		if (ir.li_key)
 			ir.li_key->release();
 		if (ir.li_item)

=== modified file 'plugin/pbms/src/cslib/CSStorage.h'
--- plugin/pbms/src/cslib/CSStorage.h	2010-10-02 21:15:42 +0000
+++ plugin/pbms/src/cslib/CSStorage.h	2010-10-20 20:21:41 +0000
@@ -51,7 +51,7 @@
 	/* Value is returned NOT referenced. */
 	CSObject *find(CSObject *key);
 	
-	void remove(CSObject *key);
+	bool remove(CSObject *key);
 
 private:
 	uint32_t iSize;
@@ -100,7 +100,7 @@
 
 class CSLinkedItem : public CSRefObject {
 public:
-	CSLinkedItem(): CSRefObject() { }
+	CSLinkedItem(): CSRefObject(), iNextLink(NULL), iPrevLink(NULL) { }
 	virtual ~CSLinkedItem() { }
 
 	virtual CSObject *getNextLink() { return iNextLink; }
@@ -129,6 +129,7 @@
 
 	/* Value must be given referenced. */
 	void addFront(CSObject *item);
+	void addBack(CSObject *item);
 
 	bool remove(CSObject *item);
 
@@ -149,6 +150,11 @@
 	CSObject *iListBack;
 };
 
+class CSSyncLinkedList : public CSLinkedList, public CSSync {
+public:
+	CSSyncLinkedList(): CSLinkedList(), CSSync() { }
+};
+
 class CSVector : public CSObject {
 public:
 	CSVector(uint32_t growSize): iGrowSize(growSize), iMaxSize(0), iUsage(0), iArray(NULL) { }
@@ -216,19 +222,19 @@
 
 	void clear();
 
-	CSObject *take(uint32_t idx);
+	CSObject *take(uint32_t sparse_idx);
 
-	void remove(uint32_t idx);
+	void remove(uint32_t sparse_idx);
 
 	void removeFirst();
 
 	CSObject *itemAt(uint32_t idx);
 
-	CSObject *get(uint32_t idx);
-	
-	uint32_t getIndex(uint32_t idx);
-	
-	void set(uint32_t idx, CSObject *);
+	CSObject *get(uint32_t sparse_idx);
+	
+	uint32_t getIndex(uint32_t sparse_idx);
+	
+	void set(uint32_t sparse_idx, CSObject *);
 
 	uint32_t size() { return iUsage; }
 

=== modified file 'plugin/pbms/src/cslib/CSStrUtil.cc'
--- plugin/pbms/src/cslib/CSStrUtil.cc	2010-10-02 21:15:42 +0000
+++ plugin/pbms/src/cslib/CSStrUtil.cc	2010-10-20 20:21:41 +0000
@@ -29,15 +29,31 @@
 #include "CSConfig.h"
 #include <inttypes.h>
 
+#include <stdlib.h>
 #include <stdio.h>
 #include <string.h>
 #include <ctype.h>
 
+#ifndef OS_WINDOWS
+#include <fnmatch.h>
+#endif
+
 #include "CSDefs.h"
 #include "CSStrUtil.h"
 #include "CSMemory.h"
 #include "CSGlobal.h"
 
+const char *cs_version()
+{
+	static char version[124];
+	
+	if (!version[0]) {
+		snprintf(version, 124, "%s(Built %s %s)", VERSION, __DATE__, __TIME__);
+	}
+	
+	return version;
+}
+
 void cs_strcpy(size_t size, char *to, const char *from, size_t len)
 {
 	if (size > 0) {
@@ -59,6 +75,25 @@
 	}
 }
 
+/* This function adds '...' to the end of the string.
+ * if it does not fit!
+ */
+void cs_strcpy_dottt(size_t size, char *d, const char *s, size_t len)
+{
+	if (len+1 <= size) {
+		cs_strcpy(size, d, s, len);
+		return;
+	}
+	if (size < 5) {
+		/* Silly, but anyway... */
+		cs_strcpy(size, d, "...");
+		return;
+	}
+	memcpy(d, s, size-4);
+	memcpy(d+size-4, "...", 3);
+	d[size-1] = 0;
+}
+
 void cs_strcpy_left(size_t size, char *to, const char *from, char ch)
 {
 	if (size > 0) {
@@ -174,6 +209,96 @@
 		cs_strcat(size, buffer, ")");
 }
 
+int cs_path_depth(const char *path)
+{
+	int count = 0;
+	while (*path) {
+		if (IS_DIR_CHAR(*path))
+			count++;
+
+		path++;
+	}
+	return count;
+}
+
+static const char *find_wildcard(const char *pattern)
+{
+	bool escaped = false;
+	while (*pattern) {
+		if ((*pattern == '*' || *pattern == '?' ) && !escaped)
+			return pattern;
+			
+		if (*pattern == '\\')
+			escaped = !escaped;
+		else
+			escaped = false;
+			
+		pattern++;
+	}
+	
+	return NULL;
+}
+
+// Check if the path contains any variable components.
+bool cs_fixed_pattern(const char *str)
+{
+	return (find_wildcard(str) == NULL);
+}
+
+#ifdef OS_WINDOWS
+/* 
+bool cs_match_patern(const char *pattern, const char *str, bool ignore_case)
+{
+	bool escaped = false;
+	
+	while (*pattern && *str) {
+		if ((*pattern == '*' || *pattern == '?' ) && !escaped) {
+			if (*pattern == '?') {
+				pattern++;
+				str++;	
+				continue;			
+			}
+			
+			while (*pattern == '*' || *pattern == '?' ) pattern++; // eat the pattern matching characters.
+			
+			if (!*pattern) // A * at the end of the pattern matches everything.
+				return true;
+				
+			// This is where it gets complicted.
+			
+			coming soon!
+			
+		}
+					
+		if (*pattern == '\\')
+			escaped = !escaped;
+		else
+			escaped = false;
+			
+		if (escaped)
+			pattern++;
+		else {
+			if (ignore_case) {
+				if (toupper(*pattern) != toupper(*str))
+					return false;
+			} else if (*pattern != *str)
+				return false;
+			pattern++;
+			str++;				
+		}
+		
+	}
+	
+	return ((!*pattern) && (!*str));
+}
+*/
+#else
+bool cs_match_patern(const char *pattern, const char *str, bool ignore_case)
+{
+	return (fnmatch(pattern, str, (ignore_case)?FNM_CASEFOLD:0) == 0);
+}
+#endif
+
 /* This function returns "" if the path ends with a dir char */
 char *cs_last_name_of_path(const char *path, int count)
 {
@@ -202,20 +327,20 @@
 }
 
 /* This function returns the last name component, even if the path ends with a dir char */
-char *cs_last_directory_of_path(char *path)
+char *cs_last_directory_of_path(const char *path)
 {
 	size_t	length;
-	char	*ptr;
+	const char	*ptr;
 
 	length = strlen(path);
 	if (!length)
-		return(path);
+		return((char *)path);
 	ptr = path + length - 1;
 	if (IS_DIR_CHAR(*ptr))
 		ptr--;
 	while (ptr != path && !IS_DIR_CHAR(*ptr)) ptr--;
 	if (IS_DIR_CHAR(*ptr)) ptr++;
-	return(ptr);
+	return((char *)ptr);
 }
 
 const char *cs_find_extension(const char *file_name)
@@ -402,6 +527,12 @@
 		return NULL;
 
 	str = (char *) cs_malloc(len + 1);
+
+	// Allow for allocation of an oversized buffer.
+	size_t str_len = strlen(in_str);
+	if (len > str_len)
+		len = str_len;
+		
 	memcpy(str, in_str, len);
 	str[len] = 0;
 	return str;
@@ -418,6 +549,47 @@
 	return *cstr || !*w_cstr;
 }
 
+bool cs_ends_with(const char *cstr, const char *w_cstr)
+{
+	size_t		len = strlen(cstr);
+	size_t		w_len = strlen(w_cstr);
+	const char	*ptr = cstr + len - 1;
+	const char	*w_ptr = w_cstr + w_len - 1;
+	
+	if (w_len > len)
+		return false;
+
+	if (w_len == 0)
+		return false;
+
+	while (w_ptr >= w_cstr) {
+		if (*w_ptr != *ptr)
+			return false;
+		w_ptr--;
+		ptr--;
+	}
+
+	return true;
+}
+
+void cs_replace_string(size_t size, char *into, const char *find_str, const char *str)
+{
+	char *ptr;
+
+	if ((ptr = strstr(into, find_str))) {
+		size_t len = strlen(into);
+		size_t len2 = strlen(str);
+		size_t len3 = strlen(find_str);
+		
+		if (len + len2 + len3 >= size)
+			len2 = size - len;
+		
+		memmove(ptr+len2, ptr+len3, len - (ptr + len3 - into));
+		memcpy(ptr, str, len2);
+		into[len + len2 - len3] = 0;
+	}
+}
+
 void cs_replace_string(size_t size, char *into, const char ch, const char *str)
 {
 	char *ptr;
@@ -469,126 +641,124 @@
 	return value;
 }
 
-int64_t cs_byte_size_to_int8(const char *ptr)
+int64_t cs_byte_size_to_int8(const char *ptr, bool *invalid)
 {
-	char	number[101], *num_ptr;
+	char	*end_ptr;
 	int64_t	size;
 
+	if (invalid)
+		*invalid = false;
+
 	while (*ptr && isspace(*ptr))
 		ptr++;
 
-	num_ptr = number;
-	while (*ptr && isdigit(*ptr)) {
-		if (num_ptr < number+100) {
-			*num_ptr = *ptr;
-			num_ptr++;
-		}
-		ptr++;
-	}
-	*num_ptr = 0;
-	size = cs_str_to_int8(number, NULL);
-
+	if (!isdigit(*ptr) && *ptr != '.')
+		goto failed;
+
+	size = (int64_t) strtod(ptr, &end_ptr);
+
+	ptr = end_ptr;
 	while (*ptr && isspace(*ptr))
 		ptr++;
 	
 	switch (toupper(*ptr)) {
+		case 'P':
+			size *= (int64_t) 1024;
+		case 'T':
+			size *= (int64_t) 1024;
 		case 'G':
-			size *= 1024LL * 1024LL * 1024LL;
-			break;
+			size *= (int64_t) 1024;
 		case 'M':
-			size *= 1024LL * 1024LL;
-			break;
+			size *= (int64_t) 1024;
 		case 'K':
-			size *= 1024LL;
-			break;
-	}
-	
-	return size;
-}
-
-
-/*--------------------------------------------------------------------------------------------------*/
-size_t cs_hex_to_bin(size_t size, void *bin, size_t len, const char *hex)
-{	
-	unsigned char *bin_ptr, *hex_ptr, c, val;
-	size_t result = 0;
-
-	if (len %2)  /* The hex string must be an even number of bytes. */
-		len--;
-		
-	if (len > (2 *size)) {
-		len = 2 * size;
-	} 
-			
-	bin_ptr = (unsigned char *) bin;	
-	hex_ptr = (unsigned char *) hex;	
-
-	
-	for (; len > 0; len--, hex_ptr++) {
-		c = *hex_ptr;
-		if ((c >= '0') && (c <= '9')) {
-			val = c - '0';
-		}
-		else {
-			c = toupper(c);
-			if ((c >= 'A') && (c <= 'F')) {
-				val = c - 'A' + 10;
-			}
-			else
-				return(result);
-		}
-		
-		if ( len & 0X01) {
-			*bin_ptr += val;
-			bin_ptr++;
-			result++;
-		}
-		else {
-			*bin_ptr = val << 4;
-		}
-	}
-	
-	return(result);
-}
-	
-/*--------------------------------------------------------------------------------------------------*/
-size_t cs_bin_to_hex(size_t size, char *hex, size_t len, const void *bin)
-{
-	static uint16_t hex_table[256], initialized = 0;
-	uint16_t *hex_ptr = (uint16_t *)hex;
-	unsigned char *bin_ptr = (unsigned char *)bin;
-	size_t	result = 0;
-
-	/* init the hex table if required */
-	if (!initialized) {
-		char buf[20];
-		int i;
-		for ( i=0; i < 256; i++) {
-			snprintf(buf, 20,"%X", i + 256);
-			memcpy(&(hex_table[i]), buf +1, 2);
-		}
-		
-		initialized = 1;
-	}
-	/*----------------------------------*/	
-
-	if (size < len *2) {
-		len = size/2;
-	}
-		
-	result = len *2;
-		
-	hex_ptr += len -1;
-	bin_ptr += len -1;
-	for (; len != 0; len--, hex_ptr--, bin_ptr--) {
-		memcpy(hex_ptr, hex_table + *bin_ptr, 2);
-	}
-	
-	// If there is room null terminate the hex string.
-	if (size > result)
-		hex[result] = 0;
-		
-	return(result);
+			size *= (int64_t) 1024;
+			ptr++;
+			break;
+		case '\0':
+			break;
+		default:
+			goto failed;
+	}
+	
+	if (toupper(*ptr) == 'B')
+		ptr++;
+
+	while (*ptr && isspace(*ptr))
+		ptr++;
+
+	if (*ptr)
+		goto failed;
+
+	return (int64_t) size;
+
+	failed:
+	if (invalid)
+		*invalid = true;
+	return 0;
+}
+
+
+static uint32_t cs_hex_value(char ch)
+{
+	u_char uch = (u_char) ch;
+
+	if (uch >= '0' && uch <= '9')
+		return uch - '0';
+	if (uch >= 'A' && uch <= 'F')
+		return uch - 'A' + 10; 
+	if (uch >= 'a' && uch <= 'f')
+		return uch - 'a' + 10;
+	return 0;
+}
+
+size_t cs_hex_to_bin(size_t size, void *v_bin, size_t len, const char *hex)
+{
+	size_t	tot_size = size;
+	uint32_t	val = 0;
+	size_t	shift = 0;
+	u_char *bin = (u_char *) v_bin;
+
+	if (len & 1)
+		shift = 1;
+	for (size_t i=shift; i<len+shift && size > 0; i++) {
+		if (i & 1) {
+			val = val | cs_hex_value(*hex);
+			*bin = val;
+			bin++;
+			size--;
+		}
+		else
+			val = cs_hex_value(*hex) << 4;
+		hex++;
+	}
+	return tot_size - size;
+}
+
+size_t cs_hex_to_bin(size_t size, void *bin, const char *hex)
+{
+	return cs_hex_to_bin(size, bin, strlen(hex), hex);
+}
+
+#define HEX_DIGIT(x)	((x) <= 9 ? '0' + (x) : 'A' + ((x) - 10))
+
+void cs_bin_to_hex(size_t size, char *hex, size_t len, const void *v_bin)
+{
+	const u_char *bin = (u_char *) v_bin;
+	if (size == 0)
+		return;
+	size--;
+	for (size_t i=0; i<len && size > 0; i++) {
+		*hex = HEX_DIGIT(*bin >> 4);
+		hex++;
+		size--;
+		if (size == 0)
+			break;
+		*hex = HEX_DIGIT(*bin & 0x0F);
+		hex++;
+		size--;
+		bin++;
+	}
+	*hex = 0;
 }	
 
 void cs_strToUpper(char *ptr)
@@ -607,3 +777,15 @@
 	}
 }
 
+/*
+ * Return failed if this is not a valid number.
+ */
+bool cs_str_to_value(const char *ptr, uint32_t *value, uint8_t base)
+{
+	char *endptr;
+
+	*value = strtoul(ptr, &endptr, base);
+	return *endptr ? false : true;
+}
+
+

=== modified file 'plugin/pbms/src/cslib/CSStrUtil.h'
--- plugin/pbms/src/cslib/CSStrUtil.h	2010-10-02 21:15:42 +0000
+++ plugin/pbms/src/cslib/CSStrUtil.h	2010-10-20 20:21:41 +0000
@@ -29,12 +29,13 @@
 #ifndef __CSUTIL_H__
 #define __CSUTIL_H__
 
-using namespace std;
-
 #include "CSDefs.h"
 
+const char	*cs_version();
+
 void		cs_strcpy(size_t size, char *d, const char *s, size_t len);
 void		cs_strcpy(size_t size, char *d, const char *s);
+void		cs_strcpy_dottt(size_t size, char *d, const char *s, size_t len);
 void		cs_strcpy_left(size_t size, char *d, const char *s, char ch);
 void		cs_strcpy_right(size_t size, char *d, const char *s, char ch);
 void		cs_strcat(size_t size, char *d, const char *s);
@@ -47,9 +48,12 @@
 void		cs_strcat(size_t size, char *to, uint64_t i);
 void		cs_strcat_hex(size_t size, char *to, uint64_t i);
 void		cs_format_context(size_t size, char *buffer, const char *func, const char *file, int line);
+int			cs_path_depth(const char *path);
+bool		cs_fixed_pattern(const char *path);
+bool		cs_match_patern(const char *pattern, const char *str, bool ignore_case);
 char		*cs_last_name_of_path(const char *path, int count);
 char		*cs_last_name_of_path(const char *path);
-char		*cs_last_directory_of_path(char *path);
+char		*cs_last_directory_of_path(const char *path);
 const char	*cs_find_extension(const char *file_name);
 void		cs_remove_extension(char *file_name);
 bool		cs_is_extension(const char *file_name, const char *ext);
@@ -64,14 +68,19 @@
 char		*cs_strdup(int value);
 char		*cs_strdup(const char *in_str, size_t len);
 bool		cs_starts_with(const char *cstr, const char *w_cstr);
+bool		cs_ends_with(const char *cstr, const char *w_cstr);
 void		cs_make_absolute_path(size_t max, char *path, const char *rel_path, const char *cwd);
+void		cs_replace_string(size_t size, char *into, const char *find_str, const char *str);
 void		cs_replace_string(size_t size, char *into, const char ch, const char *str);
 int64_t		cs_str_to_int8(const char *ptr, bool *overflow);
-int64_t		cs_byte_size_to_int8(const char *ptr);
-uint64_t	cs_str_to_word8(const char *ptr, bool *overflow);
+int64_t		cs_byte_size_to_int8(const char *ptr, bool *invalid = NULL);
+uint64_t	cs_str_to_word8(const char *ptr, bool *overflow = NULL);
+size_t		cs_hex_to_bin(size_t size, void *bin, const char *hex);
 size_t		cs_hex_to_bin(size_t size, void *bin, size_t len, const char *hex);
-size_t		cs_bin_to_hex(size_t size, char *hex, size_t len, const void *bin);
+void		cs_bin_to_hex(size_t size, char *hex, size_t len, const void *bin);
 void		cs_strToUpper(char *ptr);
 void		cs_strToLower(char *ptr);
 
+const char *cs_version();
+bool		cs_str_to_value(const char *ptr, uint32_t *value, uint8_t base);
 #endif

=== modified file 'plugin/pbms/src/cslib/CSStream.cc'
--- plugin/pbms/src/cslib/CSStream.cc	2010-10-02 21:15:42 +0000
+++ plugin/pbms/src/cslib/CSStream.cc	2010-10-18 21:21:20 +0000
@@ -39,8 +39,6 @@
 #include "CSStream.h"
 #include "CSGlobal.h"
 
-using namespace std;
-
 /*
  * ---------------------------------------------------------------
  * STREAM UTILITIES
@@ -183,7 +181,15 @@
 	return_(size == 0 ? -1 : (int) ch);
 }
 
-void CSFileInputStream::reset() {iReadOffset = 0;}
+void CSFileInputStream::reset()
+{
+	iReadOffset = 0;
+}
+
+const char *CSFileInputStream::identify()
+{
+	return iFile->myFilePath->getCString();
+}
 
 int CSFileInputStream::peek()
 {
@@ -267,7 +273,15 @@
 	exit_();
 }
 
-void CSFileOutputStream::reset() {iWriteOffset = 0;}
+void CSFileOutputStream::reset()
+{
+	iWriteOffset = 0;
+}
+
+const char *CSFileOutputStream::identify()
+{
+	return iFile->myFilePath->getCString();
+}
 
 void CSFileOutputStream::close()
 {
@@ -344,6 +358,10 @@
 	exit_();
 }
 
+const char *CSSocketInputStream::identify()
+{
+	return iSocket->identify();
+}
 
 CSSocketInputStream *CSSocketInputStream::newStream(CSSocket *s)
 {
@@ -403,6 +421,11 @@
 	exit_();
 }
 
+const char *CSSocketOutputStream::identify()
+{
+	return iSocket->identify();
+}
+
 CSSocketOutputStream *CSSocketOutputStream::newStream(CSSocket *s)
 {
 	CSSocketOutputStream *str;
@@ -490,6 +513,11 @@
 	iStream->reset();
 }
 
+const char *CSBufferedInputStream::identify()
+{
+	return iStream->identify();
+}
+
 CSBufferedInputStream *CSBufferedInputStream::newStream(CSInputStream* i)
 {
 	CSBufferedInputStream *s;
@@ -524,8 +552,11 @@
 {
 	size_t tfer;
 
+	// If the length of the data being written is greater than half
+	// the buffer size then the data is written directly through
+	// with out buffering.
 	enter_();
-	if (iBuffTotal < CS_STREAM_BUFFER_SIZE) {
+	if (iBuffTotal < CS_STREAM_BUFFER_SIZE/2) {
 		tfer = CS_STREAM_BUFFER_SIZE - iBuffTotal;
 		
 		if (tfer > len)
@@ -537,7 +568,7 @@
 	}
 	if (len > 0) {
 		flush();
-		if (len > CS_STREAM_BUFFER_SIZE)
+		if (len > CS_STREAM_BUFFER_SIZE/2)
 			iStream->write(b, len);
 		else {
 			memcpy(iBuffer, b, len);
@@ -549,10 +580,16 @@
 
 void CSBufferedOutputStream::flush()
 {
+	size_t len;
+
 	enter_();
-	if (iBuffTotal > 0) {
-		iStream->write((char *) iBuffer, iBuffTotal);
+	if ((len = iBuffTotal)) {
+		/* Discard the contents of the buffer
+		 * if flush fails, because we do
+		 * not know how much was written anyway!
+		 */
 		iBuffTotal = 0;
+		iStream->write((char *) iBuffer, len);
 	}
 	exit_();
 }
@@ -573,6 +610,11 @@
 	iStream->reset();
 }
 
+const char *CSBufferedOutputStream::identify()
+{
+	return iStream->identify();
+}
+
 CSBufferedOutputStream *CSBufferedOutputStream::newStream(CSOutputStream* i)
 {
 	CSBufferedOutputStream *s;
@@ -658,6 +700,11 @@
 	iMemSpace = iMemTotal;
 }
 
+const char *CSMemoryOutputStream::identify()
+{
+	return "memory stream";
+}
+
 /*
  * ---------------------------------------------------------------
  * STATIC (user) MEMORY OUTPUT STREAM

=== modified file 'plugin/pbms/src/cslib/CSStream.h'
--- plugin/pbms/src/cslib/CSStream.h	2010-10-02 21:15:42 +0000
+++ plugin/pbms/src/cslib/CSStream.h	2010-10-18 21:21:20 +0000
@@ -74,6 +74,9 @@
 	 */
 	virtual void reset() = 0; 
 
+	/* Return the name of the file, or whatever: */
+	virtual const char *identify() = 0;
+
 	/*
 	 * Read a line from the input stream. This function
 	 * handles all types of line endings. The function
@@ -109,15 +112,17 @@
 	virtual void flush() = 0;
 
 	/*
-	 *  Writes the specified byte to this output stream.
+	 * Writes the specified byte to this output stream.
 	 */
 	virtual void write(char b) = 0; 
 
 	/*
-	 *  Reset this output stream to the start inorder to restart the write.
+	 * Reset this output stream to the start inorder to restart the write.
 	 */
 	virtual void reset() = 0; 
 
+	virtual const char *identify() = 0;
+
 	/*
 	 * Write a line. Terminator is specific to the
 	 * type of output stream and may depend on the
@@ -159,6 +164,8 @@
 	 */
 	virtual void reset(); 
 
+	virtual const char *identify();
+
 	static CSFileInputStream *newStream(CSFile* f);
 	static CSFileInputStream *newStream(CSFile* f, off64_t offset);
 
@@ -184,6 +191,8 @@
 
 	virtual void reset(); 
 
+	virtual const char *identify();
+
 	static CSFileOutputStream *newStream(CSFile* f);
 	static CSFileOutputStream *newStream(CSFile* f, off64_t offset);
 
@@ -209,6 +218,8 @@
 
 	virtual void reset(); 
 
+	virtual const char *identify();
+
 	static CSSocketInputStream *newStream(CSSocket *s);
 
 private:
@@ -232,6 +243,8 @@
 
 	virtual void reset(); 
 
+	virtual const char *identify();
+
 	static CSSocketOutputStream *newStream(CSSocket *s);
 
 private:
@@ -239,11 +252,10 @@
 };
 
 /* Buffered Stream: */
-#ifdef DEBUG
-#define CS_STREAM_BUFFER_SIZE			11
-//#define CS_STREAM_BUFFER_SIZE			(64 * 1024)
+#ifdef DEBUG_disabled
+#define CS_STREAM_BUFFER_SIZE			80
 #else
-#define CS_STREAM_BUFFER_SIZE			(64 * 1024)
+#define CS_STREAM_BUFFER_SIZE			(32 * 1024)
 #endif
 
 class CSBufferedInputStream : public CSInputStream {
@@ -261,6 +273,8 @@
 
 	virtual void reset(); 
 
+	virtual const char *identify();
+
 	static CSBufferedInputStream *newStream(CSInputStream* i);
 
 private:
@@ -287,6 +301,8 @@
 
 	virtual void reset(); 
 
+	virtual const char *identify();
+
 	static CSBufferedOutputStream *newStream(CSOutputStream* i);
 
 private:
@@ -331,6 +347,11 @@
 
 	virtual void reset() {iMemPos = 0;}
 	
+	virtual const char *identify() 
+	{
+		return "memory stream";
+	}
+
 	static CSMemoryInputStream *newStream(const u_char* buffer, uint32_t length);
 
 private:
@@ -362,6 +383,8 @@
 	
 	virtual void reset();
 	
+	virtual const char *identify();
+	
 	static CSMemoryOutputStream *newStream(size_t init_length, size_t min_alloc);
 
 private:
@@ -392,6 +415,11 @@
 		iMemSpace = iMemSize;
 	}
 	
+	virtual const char *identify() 
+	{
+		return "memory stream";
+	}
+	
 	off64_t getSize() { return iMemPos - iMemory; }
 
 private:
@@ -466,6 +494,11 @@
 		doReset = false;
 	}
 
+	virtual const char *identify() 
+	{
+		return "callback stream";
+	}
+
 	static CSCallbackInputStream *newStream(CSStreamReadCallbackFunc callback, void *user_data);
 
 private:
@@ -475,6 +508,7 @@
 	bool havePeek;
 	bool doReset;
 };
+
 #endif
 
 

=== modified file 'plugin/pbms/src/cslib/CSString.cc'
--- plugin/pbms/src/cslib/CSString.cc	2010-10-02 21:15:42 +0000
+++ plugin/pbms/src/cslib/CSString.cc	2010-10-18 21:21:20 +0000
@@ -47,7 +47,7 @@
  * Core System String Buffers
  */
 
-CSStringBuffer_::CSStringBuffer_():
+CSStringBufferImpl::CSStringBufferImpl():
 iBuffer(NULL),
 iGrow(0),
 iSize(0),
@@ -56,7 +56,7 @@
 	iGrow = 20;
 }
 
-CSStringBuffer_::CSStringBuffer_(uint32_t grow):
+CSStringBufferImpl::CSStringBufferImpl(uint32_t grow):
 iBuffer(NULL),
 iGrow(0),
 iSize(0),
@@ -65,12 +65,12 @@
 	iGrow = grow;
 }
 
-CSStringBuffer_::~CSStringBuffer_()
+CSStringBufferImpl::~CSStringBufferImpl()
 {
 	clear();
 }
 
-void CSStringBuffer_::clear()
+void CSStringBufferImpl::clear()
 {
 	if (iBuffer)
 		cs_free(iBuffer);
@@ -79,7 +79,7 @@
 	myStrLen = 0; 
 }
 
-void CSStringBuffer_::append(char ch)
+void CSStringBufferImpl::append(char ch)
 {
 	if (iSize == myStrLen) {
 		cs_realloc((void **) &iBuffer, iSize + iGrow);
@@ -89,7 +89,7 @@
 	myStrLen++;
 }
 
-void CSStringBuffer_::append(const char *str, size_t len)
+void CSStringBufferImpl::append(const char *str, size_t len)
 {
 	if (myStrLen + len > iSize) {
 		size_t add = len;
@@ -103,23 +103,31 @@
 	myStrLen += len;
 }
 
-void CSStringBuffer_::append(int value)
-{
-	char buffer[100];
-
-	snprintf(buffer, 100, "%"PRId32"", value);
-	append(buffer);
-}
-
-void CSStringBuffer_::append(uint32_t value)
-{
-	char buffer[100];
-
-	snprintf(buffer, 100, "%"PRIu32"", value);
-	append(buffer);
-}
-
-char *CSStringBuffer_::getCString()
+void CSStringBufferImpl::append(int value)
+{
+	char buffer[100];
+
+	snprintf(buffer, 100, "%d", value);
+	append(buffer);
+}
+
+void CSStringBufferImpl::append(uint32_t value)
+{
+	char buffer[100];
+
+	snprintf(buffer, 100, "%"PRIu32, value);
+	append(buffer);
+}
+
+void CSStringBufferImpl::append(uint64_t value)
+{
+	char buffer[100];
+
+	snprintf(buffer, 100, "%"PRIu64, value);
+	append(buffer);
+}
+
+char *CSStringBufferImpl::getCString()
 {
 	if (iSize == myStrLen) {
 		cs_realloc((void **) &iBuffer, iSize + 1);
@@ -129,7 +137,7 @@
 	return iBuffer;
 }
 
-char *CSStringBuffer_::take()
+char *CSStringBufferImpl::take()
 {
 	char *buf;
 
@@ -144,7 +152,7 @@
 	return buf;
 }
 
-void CSStringBuffer_::setLength(uint32_t len)
+void CSStringBufferImpl::setLength(uint32_t len)
 {
 	if (len > iSize) {
 		cs_realloc((void **) &iBuffer, len + 1);
@@ -153,28 +161,28 @@
 	myStrLen = len;
 }
 
-uint32_t CSStringBuffer_::ignore(uint32_t pos, char ch)
+uint32_t CSStringBufferImpl::ignore(uint32_t pos, char ch)
 {
 	while (pos < myStrLen && iBuffer[pos] == ch)
 		pos++;
 	return pos;
 }
 
-uint32_t CSStringBuffer_::find(uint32_t pos, char ch)
+uint32_t CSStringBufferImpl::find(uint32_t pos, char ch)
 {
 	while (pos < myStrLen && iBuffer[pos] != ch)
 		pos++;
 	return pos;
 }
 
-uint32_t CSStringBuffer_::trim(uint32_t pos, char ch)
+uint32_t CSStringBufferImpl::trim(uint32_t pos, char ch)
 {
 	while (pos > 0 && iBuffer[pos-1] == ch)
 		pos--;
 	return pos;
 }
 
-CSString *CSStringBuffer_::substr(uint32_t pos, uint32_t len)
+CSString *CSStringBufferImpl::substr(uint32_t pos, uint32_t len)
 {
 	CSString *s = CSString::newString(iBuffer + pos, len);
 
@@ -186,21 +194,6 @@
  * Generic Strings
  */
 
-CSString *CSString::newString(const char *cstr)
-{
-	return CSCString::newString(cstr);
-}
-
-CSString *CSString::newString(const char *bytes, uint32_t len)
-{
-	return CSCString::newString(bytes, len);
-}
-
-CSString *CSString::newString(CSStringBuffer *sb)
-{
-	return CSCString::newString(sb);
-}
-
 CSString *CSString::concat(CSString *cat_str)
 {
 	CSString *new_str = NULL;
@@ -267,10 +260,10 @@
 	return (h);
 }
 
-uint32_t CSString::locate(const char *w_cstr, s_int count)
+uint32_t CSString::locate(const char *w_cstr, int32_t count)
 {
-	s_int len = length();
-	s_int i;
+	int32_t len = length();
+	int32_t i;
 
 	if (count >= 0) {
 		i = 0;
@@ -285,7 +278,7 @@
 	}
 	else {
 		count = -count;
-		i = len - (s_int) strlen(w_cstr);
+		i = len - (int32_t) strlen(w_cstr);
 		while (i >= 0) {
 			if (startsWith((uint32_t) i, w_cstr)) {
 				count--;
@@ -357,12 +350,12 @@
 	
 }
 
-CSString *CSString::left(const char *w_cstr, s_int count)
+CSString *CSString::left(const char *w_cstr, int32_t count)
 {
 	uint32_t idx = locate(w_cstr, count);
 
 	if (idx == (uint32_t)-1)
-		return CSCString::newString("");
+		return CSString::newString("");
 	return substr(0, idx);
 }
 
@@ -371,13 +364,12 @@
 	return left(w_cstr, 1);
 }
 
-CSString *CSString::right(const char *w_cstr, s_int count)
+CSString *CSString::right(const char *w_cstr, int32_t count)
 {
 	uint32_t idx = locate(w_cstr, count);
 
 	if (idx == (uint32_t)-1) {
-		this->retain();
-		return this;
+		return RETAIN(this);
 	}
 	
 	if (idx == length())
@@ -436,45 +428,50 @@
  * Standard C String
  */
 
-CSCString::CSCString():
-CSString(),
+CSString::CSString():
 myCString(NULL),
 myStrLen(0)
 {
 }
 
-CSCString::~CSCString()
+CSString::CSString(const char *cstr):
+myCString(cs_strdup(cstr)),
+myStrLen(strlen(cstr))
+{
+}
+
+CSString::~CSString()
 {
 	if (myCString)
 		cs_free(myCString);
 }
 
-const char *CSCString::getCString()
+const char *CSString::getCString()
 {
 	return myCString;
 }
 
-CS_CHAR CSCString::charAt(uint32_t pos)
+CS_CHAR CSString::charAt(uint32_t pos)
 {
 	if (pos < myStrLen)
 		return (CS_CHAR) (unsigned char) myCString[pos];
 	return (CS_CHAR) 0;
 }
 
-CS_CHAR CSCString::upperCharAt(uint32_t pos)
+CS_CHAR CSString::upperCharAt(uint32_t pos)
 {
 	if (pos < myStrLen)
 		return (CS_CHAR) (unsigned char) toupper(myCString[pos]);
 	return (CS_CHAR) 0;
 }
 
-void CSCString::setCharAt(uint32_t pos, CS_CHAR ch)
+void CSString::setCharAt(uint32_t pos, CS_CHAR ch)
 {
 	if (pos < myStrLen)
 		myCString[pos] = (unsigned char) ch;
 }
 
-int CSCString::compare(const char *val, uint32_t len)
+int CSString::compare(const char *val, uint32_t len)
 {
 	const char *pa = myCString, *pb = val;
 	int r = 0;
@@ -497,12 +494,12 @@
 	return_(r);
 }
 
-int CSCString::compare(CSString *val)
+int CSString::compare(CSString *val)
 {
 	return compare(val->getCString(), (uint32_t)-1);
 }
 
-bool CSCString::startsWith(uint32_t index, const char *w_str)
+bool CSString::startsWith(uint32_t index, const char *w_str)
 {
 	uint32_t len = strlen(w_str);
 	char *str;
@@ -519,19 +516,19 @@
 	return (*w_str == 0);
 }
 
-void CSCString::setLength(uint32_t len)
+void CSString::setLength(uint32_t len)
 {
 	cs_realloc((void **) &myCString, len+1);
 	myCString[len] = 0;
 	myStrLen = len;
 }
 
-CSString *CSCString::clone(uint32_t pos, uint32_t len)
+CSString *CSString::clone(uint32_t pos, uint32_t len)
 {
-	CSCString *str = NULL;
+	CSString *str = NULL;
 	
 	enter_();
-	new_(str, CSCString());
+	new_(str, CSString());
 	push_(str);
 	
 	str->myCString = (char *) cs_malloc(len + 1);
@@ -554,22 +551,22 @@
 	return_(str);
 }
 
-CSObject *CSCString::getKey()
+CSObject *CSString::getKey()
 {
 	return (CSObject *) this;
 }
 
-int CSCString::compareKey(CSObject *key)
+int CSString::compareKey(CSObject *key)
 {
 	return compare((CSString *) key);
 }
 
-CSCString *CSCString::newString(const char *cstr)
+CSString *CSString::newString(const char *cstr)
 {
-	CSCString *str;
+	CSString *str;
 	
 	enter_();
-	new_(str, CSCString());
+	new_(str, CSString());
 	push_(str);
 	str->myCString = cs_strdup(cstr);
 	str->myStrLen = strlen(cstr);
@@ -577,12 +574,12 @@
 	return_(str);
 }
 
-CSCString *CSCString::newString(const char *bytes, uint32_t len)
+CSString *CSString::newString(const char *bytes, uint32_t len)
 {
-	CSCString *str;
+	CSString *str;
 	
 	enter_();
-	new_(str, CSCString());
+	new_(str, CSString());
 	push_(str);
 	str->myStrLen = len;
 	str->myCString = (char *) cs_malloc(len+1);
@@ -592,18 +589,18 @@
 	return_(str);
 }
 
-CSCString *CSCString::newString(CSStringBuffer *sb)
+CSString *CSString::newString(CSStringBuffer *sb)
 {
-	CSCString *str;
+	CSString *str;
 	
 	enter_();
 	push_(sb);
-	new_(str, CSCString());
+	new_(str, CSString());
 	push_(str);
+	str->myStrLen = sb->length();
 	str->myCString = sb->take();
-	str->myStrLen = sb->length();
 	pop_(str);
-	release_(sb);
+	pop_(sb); // pop this do not release it because CSStringBuffer is NOT a CSRefObject. 
 	return_(str);
 }
 

=== modified file 'plugin/pbms/src/cslib/CSString.h'
--- plugin/pbms/src/cslib/CSString.h	2010-10-02 21:15:42 +0000
+++ plugin/pbms/src/cslib/CSString.h	2010-10-18 21:21:20 +0000
@@ -45,13 +45,14 @@
  */
 #define unichar			uint16_t
 
-// CSStringBuffer_ is just a base class to define
-// referenced and non referenced versions on.
-class CSStringBuffer_ {
+/* CSStringBufferImpl is the string buffer implementation.
+ * Of this implementation we have various types.
+ */
+class CSStringBufferImpl {
 public:
-	CSStringBuffer_();
-	CSStringBuffer_(uint32_t grow);
-	~CSStringBuffer_();
+	CSStringBufferImpl();
+	CSStringBufferImpl(uint32_t grow);
+	~CSStringBufferImpl();
 
 	void clear();
 
@@ -65,6 +66,8 @@
 
 	void append(uint32_t value);
 
+	void append(uint64_t value);
+
 	char *getCString();
 
 	char *getBuffer(uint32_t pos) { return iBuffer ? iBuffer + pos : NULL; }
@@ -85,6 +88,14 @@
 
 	CSString *substr(uint32_t pos, uint32_t len);
 
+	void take(CSStringBufferImpl *buf) {
+		clear();
+		iGrow = buf->iGrow;
+		iSize = buf->iSize;
+		myStrLen = buf->myStrLen;
+		iBuffer = buf->take();
+	}
+
 private:
 	char *iBuffer;
 	uint32_t iGrow;
@@ -92,16 +103,24 @@
 	uint32_t myStrLen;
 };
 
-class CSStringBuffer : public CSStringBuffer_, public CSObject {
-public:
-	CSStringBuffer(): CSStringBuffer_(){ }
-	CSStringBuffer(uint32_t grow): CSStringBuffer_(grow){ }
-};
-
-class CSRefStringBuffer : public CSStringBuffer_, public CSRefObject {
-public:
-	CSRefStringBuffer(): CSStringBuffer_(){ }
-	CSRefStringBuffer(uint32_t grow): CSStringBuffer_(grow){ }
+class CSStringBuffer : public CSStringBufferImpl, public CSObject {
+public:
+	CSStringBuffer(): CSStringBufferImpl(), CSObject() { }
+	CSStringBuffer(uint32_t grow): CSStringBufferImpl(grow), CSObject() { }
+};
+
+class CSStaticStringBuffer : public CSStringBufferImpl, public CSStaticObject {
+	virtual void finalize() { clear(); }
+
+public:
+	CSStaticStringBuffer(): CSStringBufferImpl(), CSStaticObject() { }
+	CSStaticStringBuffer(uint32_t grow): CSStringBufferImpl(grow), CSStaticObject() { }
+};
+
+class CSRefStringBuffer : public CSStringBufferImpl, public CSRefObject {
+public:
+	CSRefStringBuffer(): CSStringBufferImpl(), CSRefObject() { }
+	CSRefStringBuffer(uint32_t grow): CSStringBufferImpl(grow), CSRefObject() { }
 };
 
 class CSSyncStringBuffer : public CSStringBuffer, public CSSync {
@@ -114,8 +133,9 @@
 
 class CSString : public CSRefObject {
 public:
-	CSString() { }
-	virtual ~CSString() { }
+	CSString();
+	CSString(const char *cstr);
+	virtual ~CSString();
 
 	/*
 	 * Construct a string from a C-style UTF-8
@@ -134,14 +154,14 @@
 	 * The returned string must be
 	 * not be freed by the caller.
 	 */
-	virtual const char *getCString() = 0;
+	virtual const char *getCString();
 
 	/*
 	 * Return the character at a certain point:
 	 */
-	virtual CS_CHAR charAt(uint32_t pos) = 0;
-	virtual CS_CHAR upperCharAt(uint32_t pos) = 0;
-	virtual void setCharAt(uint32_t pos, CS_CHAR ch) = 0;
+	virtual CS_CHAR charAt(uint32_t pos);
+	virtual CS_CHAR upperCharAt(uint32_t pos);
+	virtual void setCharAt(uint32_t pos, CS_CHAR ch);
 
 	/*
 	 * Returns < 0 if this string is
@@ -149,24 +169,24 @@
 	 * > 0 if sortede after.
 	 * The comparison is case-insensitive.
 	 */
-	virtual int compare(CSString *val) = 0;
-	virtual int compare(const char *val, uint32_t len = ((uint32_t) 0xFFFFFFFF)) = 0;
+	virtual int compare(CSString *val);
+	virtual int compare(const char *val, uint32_t len = ((uint32_t) 0xFFFFFFFF));
 
 	/*
 	 * Case sensitive match.
 	 */
-	virtual bool startsWith(uint32_t index, const char *) = 0;
+	virtual bool startsWith(uint32_t index, const char *);
 
 	/* Returns the string length in characters. */
-	virtual uint32_t length() = 0;
-	virtual void setLength(uint32_t len) = 0;
+	virtual uint32_t length() { return myStrLen; }
+	virtual void setLength(uint32_t len);
 
 	virtual bool equals(const char *str);
 
 	/*
 	 * Return a copy of this string.
 	 */
-	virtual CSString *clone(uint32_t pos, uint32_t len) = 0;
+	virtual CSString *clone(uint32_t pos, uint32_t len);
 
 	/*
 	 * Concatinate 2 strings.
@@ -196,7 +216,7 @@
 	 * the length of the string, and search
 	 * from right to left will return 0.
 	 */
-	virtual uint32_t locate(const char *, s_int count);
+	virtual uint32_t locate(const char *, int32_t count);
 	virtual uint32_t locate(uint32_t pos, const char *);
 	virtual uint32_t locate(uint32_t pos, CS_CHAR ch);
 
@@ -205,10 +225,10 @@
 	virtual CSString *substr(uint32_t index, uint32_t size);
 	virtual CSString *substr(uint32_t index);
 
-	virtual CSString *left(const char *, s_int count);
+	virtual CSString *left(const char *, int32_t count);
 	virtual CSString *left(const char *);
 
-	virtual CSString *right(const char *, s_int count);
+	virtual CSString *right(const char *, int32_t count);
 	virtual CSString *right(const char *);
 
 	virtual bool startsWith(const char *);
@@ -221,48 +241,14 @@
 
 	virtual CSString *clone(uint32_t len);
 	virtual CSString *clone();
-};
-
-class CSCString : public CSString {
-public:
-	char *myCString;
-	uint32_t myStrLen;
-
-	CSCString();
-	~CSCString();
-
-
-	virtual const char *getCString();
-
-	virtual CS_CHAR charAt(uint32_t pos);
-
-	virtual CS_CHAR upperCharAt(uint32_t pos);
-
-	virtual void setCharAt(uint32_t pos, CS_CHAR ch);
-
-	virtual int compare(CSString *val);
-	virtual int compare(const char *val, uint32_t len = ((uint32_t) 0xFFFFFFFF));
-
-	virtual bool startsWith(uint32_t index, const char *);
-	virtual bool startsWith(const char *str) { return CSString::startsWith(str);}
-
-	virtual uint32_t length() { return myStrLen; }
-
-	virtual void setLength(uint32_t len);
-
-	virtual CSString *clone(uint32_t pos, uint32_t len);
-	virtual CSString *clone(uint32_t len){ return CSString::clone(len);}
-	virtual CSString *clone(){ return CSString::clone();}
 
 	virtual CSObject *getKey();
 
 	virtual int compareKey(CSObject *);
 
-	static CSCString *newString(const char *cstr);
-
-	static CSCString *newString(const char *bytes, uint32_t len);
-
-	static CSCString *newString(CSStringBuffer *sb);
+private:
+	char *myCString;
+	uint32_t myStrLen;
 };
 
 #endif

=== added file 'plugin/pbms/src/cslib/CSSys.h'
--- plugin/pbms/src/cslib/CSSys.h	1970-01-01 00:00:00 +0000
+++ plugin/pbms/src/cslib/CSSys.h	2010-10-20 20:21:41 +0000
@@ -0,0 +1,128 @@
+/* Copyright (c) 2010 PrimeBase Technologies GmbH, Germany
+ *
+ * PrimeBase Media Stream for MySQL
+ *
+ * This program is free software; you can redistribute it and/or modify
+ * it under the terms of the GNU General Public License as published by
+ * the Free Software Foundation; either version 2 of the License, or
+ * (at your option) any later version.
+ *
+ * This program is distributed in the hope that it will be useful,
+ * but WITHOUT ANY WARRANTY; without even the implied warranty of
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+ * GNU General Public License for more details.
+ *
+ * You should have received a copy of the GNU General Public License
+ * along with this program; if not, write to the Free Software
+ * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
+ *
+ * Author: Barry Leslie
+ *
+ * 2010-02-05
+ *
+ * CORE SYSTEM:
+ * Basic system specific file I/O classes.
+ * The classes in this header are defined in CSSys_unix.cc and  CSSys_win.cc
+ *
+ */
+
+#ifndef __CSSYSFILE_H__
+#define __CSSYSFILE_H__
+
+#ifdef OS_WINDOWS
+#define INVALID_DIR_HANDLE (INVALID_HANDLE_VALUE)
+#define INVALID_FILE_HANDLE (INVALID_HANDLE_VALUE)
+#else
+#include <dirent.h>
+#define INVALID_DIR_HANDLE (NULL)
+#define INVALID_FILE_HANDLE (-1)
+#endif
+
+#include "CSString.h"
+#include "CSException.h"
+#include "CSTime.h"
+
+class CSSysDir {
+public:
+	CSSysDir(): 
+		sd_path(NULL),
+		sd_filter(NULL),
+		sd_dir(INVALID_DIR_HANDLE){}
+	
+	~CSSysDir();
+	
+	void open();
+	void close();
+	bool next();
+
+	bool entryIsFile();
+	const char *entryName();
+	void getEntryPath(char *path, size_t size);
+	
+	CSString *sd_path;
+	
+private:
+	CSStringBuffer *sd_filter;
+#ifdef OS_WINDOWS
+	WIN32_FIND_DATAA sd_entry;
+	HANDLE sd_dir;
+#else
+	struct dirent sd_entry;
+	DIR *sd_dir;
+#endif
+};
+
+
+class CSSysFile {
+public:
+
+	static bool isDirNotFound(CSException *e);
+	static bool isFileNotFound(CSException *e);
+	static bool isDirExists(CSException *e);
+
+CSSysFile(): sf_path(NULL), sf_fh(INVALID_FILE_HANDLE){}
+
+	~CSSysFile(){ sf_close();}
+	
+	bool fs_isOpen() { return ( sf_fh != INVALID_FILE_HANDLE);}
+	
+	void sf_open(const char *path, bool readonly, bool create);
+	void sf_close();
+	
+	size_t sf_pread(void *data, size_t size, off64_t offset);
+	void sf_pwrite(const void *data, size_t size, off64_t offset);
+	
+	off64_t sf_getEOF();
+	void sf_setEOF(off64_t offset);
+	
+	void sf_sync();
+	
+	void sf_lock(bool shared);
+	void sf_unlock();
+	
+private:
+	CSString *sf_path;
+#ifdef OS_WINDOWS
+	HANDLE	sf_fh;
+#else
+	int		sf_fh;
+#endif
+	
+};
+
+class CSSys {
+	public:
+	static bool sys_exists(const char *path);
+	static void sys_makeDir(const char *path);
+	static void sys_removeDir(const char *path);
+	static void sys_removeFile(const char *path);
+	static void sys_rename(const char *old_path, const char *new_path);
+	static void sys_stat(const char *path, bool *is_dir, off64_t *size, CSTime *mod_time);
+	static bool sys_isLink(const char *path);
+	static void sys_getcwd(char *path, size_t size);
+	static void sys_setcwd(const char *path);
+	static uint32_t sys_getpid();
+	static bool sys_isAlive(uint32_t pid);
+};
+
+#endif

=== added file 'plugin/pbms/src/cslib/CSSys_unix.cc'
--- plugin/pbms/src/cslib/CSSys_unix.cc	1970-01-01 00:00:00 +0000
+++ plugin/pbms/src/cslib/CSSys_unix.cc	2010-10-20 20:21:41 +0000
@@ -0,0 +1,361 @@
+/* Copyright (c) 2010 PrimeBase Technologies GmbH, Germany
+ *
+ * PrimeBase Media Stream for MySQL
+ *
+ * This program is free software; you can redistribute it and/or modify
+ * it under the terms of the GNU General Public License as published by
+ * the Free Software Foundation; either version 2 of the License, or
+ * (at your option) any later version.
+ *
+ * This program is distributed in the hope that it will be useful,
+ * but WITHOUT ANY WARRANTY; without even the implied warranty of
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+ * GNU General Public License for more details.
+ *
+ * You should have received a copy of the GNU General Public License
+ * along with this program; if not, write to the Free Software
+ * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
+ *
+ * Author: Barry Leslie
+ *
+ * 2010-02-05
+ *
+ * CORE SYSTEM:
+ * Basic UNIX specific file I/O classes.
+ *
+ */
+
+#include "CSConfig.h"
+#include <sys/stat.h>
+#include <fcntl.h>
+#include <errno.h>
+#include <string.h>
+#include <sys/file.h>
+#include <unistd.h>
+#include <signal.h>
+
+#include "CSGlobal.h"
+#include "CSDefs.h"
+#include "CSStrUtil.h"
+#include "CSSys.h"
+
+#define CS_MASK		((S_IRUSR | S_IWUSR) | (S_IRGRP | S_IWGRP) | (S_IROTH))
+//=====================
+// CSSysFile
+//=====================
+bool CSSysFile::isDirNotFound(CSException *e) { return e->getErrorCode() == ENOENT; }
+bool CSSysFile::isFileNotFound(CSException *e) { return e->getErrorCode() == ENOENT; }
+bool CSSysFile::isDirExists(CSException *e) { return e->getErrorCode() == EEXIST; }
+
+//--------------
+void CSSysFile::sf_open(const char *path, bool readonly, bool create)
+{
+	int flags;
+	
+	flags = (readonly)?O_RDONLY:O_RDWR;
+	
+	if (create)
+		flags |= O_CREAT;
+		
+	if (sf_fh != -1)
+		sf_close();
+	
+	sf_path = CSString::newString(path);
+	
+	sf_fh = open(path, flags, CS_MASK);
+	if (sf_fh == -1) {
+		sf_path->release();
+		sf_path = NULL;
+		CSException::throwFileError(CS_CONTEXT, path, errno);
+	}
+}
+
+//--------------
+void CSSysFile::sf_close()
+{
+	if (sf_fh != -1) {
+		close(sf_fh);
+		sf_fh = -1;
+		sf_path->release();
+		sf_path = NULL;
+	}	
+}
+
+//--------------
+size_t CSSysFile::sf_pread(void *data, size_t size, off64_t offset)
+{
+	ssize_t read_size;
+	
+	read_size = pread(sf_fh, data, size, offset);
+	if (read_size ==  -1)
+		CSException::throwFileError(CS_CONTEXT, sf_path->getCString(), errno);
+
+	return read_size;
+}
+
+//--------------
+void CSSysFile::sf_pwrite(const void *data, size_t size, off64_t offset)
+{
+	size_t write_size;
+	
+	write_size = pwrite(sf_fh, data, size, offset);
+	if (write_size != size)
+		CSException::throwFileError(CS_CONTEXT, sf_path->getCString(), errno);
+
+}
+
+//--------------
+void CSSysFile::sf_setEOF(off64_t offset)
+{
+	if (ftruncate(sf_fh, offset) == -1)
+		CSException::throwFileError(CS_CONTEXT, sf_path->getCString(), errno);
+}
+
+//--------------
+off64_t CSSysFile::sf_getEOF()
+{
+	off64_t eof;
+
+	if ((eof = lseek(sf_fh, 0, SEEK_END)) == ((off64_t)-1))
+		CSException::throwFileError(CS_CONTEXT, sf_path->getCString(), errno);
+
+     return eof;
+}
+
+//--------------
+void CSSysFile::sf_sync()
+{
+	fsync(sf_fh);
+}
+
+//--------------
+void CSSysFile::sf_lock(bool shared)
+{
+	if (flock(sf_fh, (shared)?LOCK_SH:LOCK_EX) == -1)
+		CSException::throwOSError(CS_CONTEXT, errno);
+}
+
+//--------------
+void CSSysFile::sf_unlock()
+{
+	if (flock(sf_fh, LOCK_UN) == -1)
+		CSException::throwOSError(CS_CONTEXT, errno);
+}
+
+//=====================
+// CSSys
+//=====================
+//--------------
+bool CSSys::sys_exists(const char *path)
+{
+	 return (access(path, F_OK) != -1);
+}
+
+//--------------
+void CSSys::sys_makeDir(const char *path)
+{
+	char		super_path[PATH_MAX];
+	struct stat	stats;
+	char		*ptr;
+
+	if (mkdir(path, S_IRWXU | S_IRWXG | S_IRWXO) == -1)
+		CSException::throwFileError(CS_CONTEXT, path, errno);
+
+	// Set the access privileges.
+	ptr = cs_last_name_of_path(path);
+	if (ptr == path) 
+		strcpy(super_path, ".");
+	else {
+		cs_strcpy(PATH_MAX, super_path, path);
+
+		if ((ptr = cs_last_name_of_path(super_path)))
+			*ptr = 0;
+	}
+	
+	if (stat(super_path, &stats) == -1)
+		CSException::throwFileError(CS_CONTEXT, path, errno);
+
+	if (chmod(path, stats.st_mode) == -1)
+		CSException::throwFileError(CS_CONTEXT, path, errno);
+
+}
+
+//--------------
+void CSSys::sys_removeDir(const char *path)
+{
+	if (rmdir(path) == -1) {
+		int err = errno;
+
+		if (err != ENOENT)
+			CSException::throwFileError(CS_CONTEXT, path, err);
+	}
+}
+
+//--------------
+void CSSys::sys_removeFile(const char *path)
+{
+	if (unlink(path) == -1) {
+		int err = errno;
+
+		if (err != ENOENT)
+			CSException::throwFileError(CS_CONTEXT, path, err);
+	}
+}
+
+
+//--------------
+
+void CSSys::sys_stat(const char *path, bool *is_dir, off64_t *size, CSTime *mod_time)
+{
+	struct stat sb;
+
+	if (stat(path, &sb) == -1)
+		CSException::throwFileError(CS_CONTEXT, path, errno);
+	if (is_dir)
+		*is_dir = sb.st_mode & S_IFDIR;
+	if (size)
+		*size = sb.st_size;
+	if (mod_time)
+#ifdef __USE_MISC
+		/* This is the Linux version: */
+		mod_time->setUTC1970(sb.st_mtim.tv_sec, sb.st_mtim.tv_nsec);
+#else
+		/* This is the Mac OS X version: */
+		mod_time->setUTC1970(sb.st_mtimespec.tv_sec, sb.st_mtimespec.tv_nsec);
+#endif
+}
+
+//--------------
+bool CSSys::sys_isLink(const char *path)
+{
+	struct stat sb;
+
+	if (lstat(path, &sb) == -1)
+		CSException::throwFileError(CS_CONTEXT, path, errno);
+		
+	return S_ISLNK(sb.st_mode);
+}
+
+//--------------
+void CSSys::sys_rename(const char *old_path, const char *new_path)
+{
+	 if (rename(old_path, new_path) == -1)
+		CSException::throwFileError(CS_CONTEXT, old_path, errno);
+}
+
+//--------------
+void CSSys::sys_getcwd(char *path, size_t size)
+{
+	if (getcwd(path, size) == NULL)
+		CSException::throwOSError(CS_CONTEXT, errno);
+}
+
+//--------------
+void CSSys::sys_setcwd(const char *path)
+{
+	if (chdir(path) == -1)
+		CSException::throwFileError(CS_CONTEXT, path, errno);
+}
+
+//--------------
+uint32_t CSSys::sys_getpid()
+{
+	return getpid();
+}
+
+//--------------
+bool CSSys::sys_isAlive(uint32_t pid)
+{
+	return (kill(pid, 0) == 0);
+}
+
+//=====================
+// CSSysDir
+//=====================
+CSSysDir::~CSSysDir()
+{
+	close();
+	if (sd_path)
+		sd_path->release();
+		
+	if (sd_filter)
+		sd_filter->release();
+}
+
+//--------------
+void CSSysDir::open()
+{
+	enter_();
+	if (!(sd_dir = opendir(sd_path->getCString())))
+		CSException::throwFileError(CS_CONTEXT, sd_path->getCString(), errno);
+	exit_();
+}
+
+//--------------
+void CSSysDir::close()
+{
+	enter_();
+	if (sd_dir) {
+		closedir(sd_dir);
+		sd_dir = NULL;
+	}
+	exit_();
+}
+
+//--------------
+bool CSSysDir::next()
+{
+	int				err;
+	struct dirent	*result;
+
+	enter_();
+	for (;;) {
+		err = readdir_r(sd_dir, &sd_entry, &result);
+		self->interrupted();
+		if (err)
+			CSException::throwFileError(CS_CONTEXT, sd_path->getCString(), err);
+		if (!result)
+			break;
+		/* Filter out '.' and '..': */
+		if (sd_entry.d_name[0] == '.') {
+			if (sd_entry.d_name[1] == '.') {
+				if (sd_entry.d_name[2] == '\0')
+					continue;
+			}
+			else {
+				if (sd_entry.d_name[1] == '\0')
+					continue;
+			}
+		}
+		break;
+	}
+	return_(result ? true : false);
+}
+
+
+//--------------
+void CSSysDir::getEntryPath(char *path, size_t size)
+{
+	cs_strcpy(size, path, sd_path->getCString());
+	cs_add_dir_char(size, path);
+	cs_strcat(size, path, sd_entry.d_name);
+}
+
+//--------------
+const char *CSSysDir::entryName()
+{
+	return sd_entry.d_name;
+}
+
+//--------------
+bool CSSysDir::entryIsFile()
+{
+	if (sd_entry.d_type & DT_DIR)
+		return false;
+	return true;
+}
+
+//--------------
+extern void unix_close(int h);
+void unix_close(int h) {close(h);}
+

=== added file 'plugin/pbms/src/cslib/CSSys_win.cc'
--- plugin/pbms/src/cslib/CSSys_win.cc	1970-01-01 00:00:00 +0000
+++ plugin/pbms/src/cslib/CSSys_win.cc	2010-10-20 20:21:41 +0000
@@ -0,0 +1,501 @@
+/* Copyright (c) 2010 PrimeBase Technologies GmbH, Germany
+ *
+ * PrimeBase Media Stream for MySQL
+ *
+ * This program is free software; you can redistribute it and/or modify
+ * it under the terms of the GNU General Public License as published by
+ * the Free Software Foundation; either version 2 of the License, or
+ * (at your option) any later version.
+ *
+ * This program is distributed in the hope that it will be useful,
+ * but WITHOUT ANY WARRANTY; without even the implied warranty of
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+ * GNU General Public License for more details.
+ *
+ * You should have received a copy of the GNU General Public License
+ * along with this program; if not, write to the Free Software
+ * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
+ *
+ * Author: Barry Leslie
+ *
+ * 2010-02-05
+ *
+ * CORE SYSTEM:
+ * Basic UNIX specific file I/O classes.
+ *
+ */
+
+#include "CSConfig.h"
+
+#include <io.h>
+
+#include "CSGlobal.h"
+#include "CSDefs.h"
+#include "CSStrUtil.h"
+#include "CSSys.h"
+
+#define CS_MASK		((S_IRUSR | S_IWUSR) | (S_IRGRP | S_IWGRP) | (S_IROTH))
+//=====================
+// CSSysFile
+//=====================
+static int get_win_error()
+{
+	return (int) GetLastError();
+}
+
+bool CSSysFile::isDirNotFound(CSException *e) { return e->getErrorCode() == ERROR_PATH_NOT_FOUND; }
+bool CSSysFile::isFileNotFound(CSException *e) { return e->getErrorCode() == ERROR_FILE_NOT_FOUND; }
+bool CSSysFile::isDirExists(CSException *e) { return e->getErrorCode() == ERROR_ALREADY_EXISTS; }
+
+
+//--------------
+void CSSysFile::sf_open(const char *path, bool readonly, bool create)
+{
+	SECURITY_ATTRIBUTES	sa = { sizeof(SECURITY_ATTRIBUTES), 0, 0 };
+	DWORD				flags;
+
+	flags = (create)?OPEN_ALWAYS:OPEN_EXISTING;
+	
+	if (sf_fh != INVALID_HANDLE_VALUE)
+		sf_close();
+	
+	sf_path = CSString::newString(path);
+	
+	sf_fh = CreateFileA(
+		path,
+		readonly ? GENERIC_READ : (GENERIC_READ | GENERIC_WRITE),
+		FILE_SHARE_READ | FILE_SHARE_WRITE,
+		&sa,
+		flags,
+		FILE_FLAG_RANDOM_ACCESS,
+		NULL);
+		
+	if (sf_fh == INVALID_HANDLE_VALUE) {
+		sf_path->release();
+		sf_path = NULL;
+		CSException::throwFileError(CS_CONTEXT, path, get_win_error());
+	}
+}
+
+//--------------
+void CSSysFile::sf_close()
+{
+	if (sf_fh != INVALID_HANDLE_VALUE) {
+		CloseHandle(sf_fh);
+		sf_fh = INVALID_HANDLE_VALUE;
+		sf_path->release();
+		sf_path = NULL;
+	}	
+}
+
+//--------------
+size_t CSSysFile::sf_pread(void *data, size_t size, off64_t offset)
+{
+	LARGE_INTEGER	liDistanceToMove;
+	DWORD			result;
+
+	liDistanceToMove.QuadPart = offset;
+	if (!SetFilePointerEx(sf_fh, liDistanceToMove, NULL, FILE_BEGIN))
+		CSException::throwFileError(CS_CONTEXT, sf_path->getCString(), get_win_error());
+
+	if (!ReadFile(sf_fh, data, size, &result, NULL))
+		CSException::throwFileError(CS_CONTEXT, sf_path->getCString(), get_win_error());
+
+	return (size_t) result;
+}
+
+//--------------
+void CSSysFile::sf_pwrite(const void *data, size_t size, off64_t offset)
+{
+	LARGE_INTEGER	liDistanceToMove;
+	DWORD			result;
+	
+	liDistanceToMove.QuadPart = offset;
+	if (!SetFilePointerEx(sf_fh, liDistanceToMove, NULL, FILE_BEGIN))
+		CSException::throwFileError(CS_CONTEXT, sf_path->getCString(), get_win_error());
+
+	if (!WriteFile(sf_fh, data, size, &result, NULL))
+		CSException::throwFileError(CS_CONTEXT, sf_path->getCString(), get_win_error());
+
+	if (result != size)
+		CSException::throwFileError(CS_CONTEXT, sf_path->getCString(), ERROR_HANDLE_EOF);
+	
+}
+
+//--------------
+void CSSysFile::sf_setEOF(off64_t offset)
+{
+	LARGE_INTEGER liDistanceToMove;
+	
+	liDistanceToMove.QuadPart = offset;
+	if (!SetFilePointerEx(sf_fh, liDistanceToMove, NULL, FILE_BEGIN)) 
+		CSException::throwFileError(CS_CONTEXT, sf_path->getCString(), get_win_error());
+
+	if (!SetEndOfFile(sf_fh)) 
+		CSException::throwFileError(CS_CONTEXT, sf_path->getCString(), get_win_error());
+}
+
+//--------------
+off64_t CSSysFile::sf_getEOF()
+{
+	DWORD			result;
+	LARGE_INTEGER	lpFileSize;
+
+	result = SetFilePointer(sf_fh, 0, NULL, FILE_END);
+	if (result == 0xFFFFFFFF)  
+		CSException::throwFileError(CS_CONTEXT, sf_path->getCString(), get_win_error());
+
+	if (!GetFileSizeEx(sf_fh, &lpFileSize))  
+		CSException::throwFileError(CS_CONTEXT, sf_path->getCString(), get_win_error());
+
+	return lpFileSize.QuadPart;
+}
+
+//--------------
+void CSSysFile::sf_sync()
+{
+	if (!FlushFileBuffers(sf_fh))  
+		CSException::throwFileError(CS_CONTEXT, sf_path->getCString(), get_win_error());
+}
+
+//--------------
+void CSSysFile::sf_lock(bool shared)
+{
+	OVERLAPPED overlap = {0};
+	
+	if (!LockFileEx(sf_fh, (shared)? 0: LOCKFILE_EXCLUSIVE_LOCK, 0, 512, 0, &overlap))  
+		CSException::throwFileError(CS_CONTEXT, sf_path->getCString(), get_win_error());
+}
+
+//--------------
+void CSSysFile::sf_unlock()
+{
+	OVERLAPPED overlap = {0};
+
+	if (!UnlockFileEx(sf_fh, 0, 512, 0, &overlap))  
+		CSException::throwFileError(CS_CONTEXT, sf_path->getCString(), get_win_error());
+}
+
+//=====================
+// CSSys
+//=====================
+//--------------
+bool CSSys::sys_exists(const char *path)
+{
+	 return (access(path, 0) != -1);
+}
+
+//--------------
+void CSSys::sys_getcwd(char *path, size_t size)
+{
+	DWORD	len;
+
+	len = GetCurrentDirectoryA(size, path);
+	if (len == 0)
+		CSException::throwFileError(CS_CONTEXT, "GetCurrentDirectory()" , get_win_error());
+	else if (len > (size -1))
+		CSException::throwFileError(CS_CONTEXT, "GetCurrentDirectory()overflow " , len);
+
+}
+
+//--------------
+void CSSys::sys_setcwd(const char *path)
+{
+	if (!SetCurrentDirectoryA(path))
+		CSException::throwFileError(CS_CONTEXT, "SetCurrentDirectory()" , get_win_error());
+}
+
+//--------------
+void CSSys::sys_makeDir(const char *path)
+{
+	SECURITY_ATTRIBUTES	sa = { sizeof(SECURITY_ATTRIBUTES), 0, 0 };
+
+	if (!CreateDirectoryA(path, &sa))
+		CSException::throwFileError(CS_CONTEXT, path, get_win_error());
+
+}
+
+#define FILE_NOT_FOUND(x)		((x) == ERROR_FILE_NOT_FOUND || (x) == ERROR_PATH_NOT_FOUND)
+//--------------
+void CSSys::sys_removeDir(const char *path)
+{
+	if (!RemoveDirectoryA(path)) {
+		int err = get_win_error();
+
+		if (!FILE_NOT_FOUND(err)) 
+			CSException::throwFileError(CS_CONTEXT, path, err);
+	}
+}
+
+//--------------
+void CSSys::sys_removeFile(const char *path)
+{
+	if (!DeleteFileA(path)) {
+		int err = get_win_error();
+
+		if (!FILE_NOT_FOUND(err)) 
+			CSException::throwFileError(CS_CONTEXT, path, err);
+	}
+}
+
+
+//--------------
+
+void CSSys::sys_stat(const char *path, bool *is_dir, off64_t *size, CSTime *mod_time)
+{
+	HANDLE						fh;
+	BY_HANDLE_FILE_INFORMATION	info;
+	SECURITY_ATTRIBUTES			sa = { sizeof(SECURITY_ATTRIBUTES), 0, 0 };
+
+	fh = CreateFileA(
+		path,
+		0,
+		FILE_SHARE_READ,
+		&sa,
+		OPEN_EXISTING,
+		FILE_ATTRIBUTE_NORMAL | FILE_FLAG_BACKUP_SEMANTICS, // FILE_FLAG_BACKUP_SEMANTICS allows you to open directories.
+		NULL);
+		
+	if (fh == INVALID_HANDLE_VALUE) {
+		CSException::throwFileError(CS_CONTEXT, path, get_win_error());
+	}
+
+	if (!GetFileInformationByHandle(fh, &info)) {
+		CloseHandle(fh);
+		CSException::throwFileError(CS_CONTEXT, path, get_win_error());
+	}
+
+	CloseHandle(fh);
+	if (is_dir)
+		*is_dir = ((info.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) == FILE_ATTRIBUTE_DIRECTORY);
+
+	if (size)
+		*size = (off64_t) info.nFileSizeLow | (((off64_t) info.nFileSizeHigh) << 32);
+		
+	if (mod_time) {
+		SYSTEMTIME st;
+		FileTimeToSystemTime(&info.ftLastWriteTime, &st);
+		mod_time->setUTC(st.wYear, st.wMonth, st.wDay, st.wHour, st.wMinute, st.wSecond, st.wMilliseconds * 1000000);
+	}
+
+}
+
+//--------------
+bool CSSys::sys_isLink(const char *path)
+{
+	CSException::throwFileError(CS_CONTEXT, "CSSys::sys_isLink() not implimented on windows.", -1);
+	return false;	
+}
+
+//--------------
+void CSSys::sys_rename(const char *old_path, const char *new_path)
+{
+	 if (rename(old_path, new_path) == -1)
+		CSException::throwFileError(CS_CONTEXT, old_path, errno);
+}
+
+//--------------
+uint32_t CSSys::sys_getpid()
+{
+	return GetCurrentProcessId();
+}
+
+//--------------
+bool CSSys::sys_isAlive(uint32_t pid)
+{
+	HANDLE h;
+	bool isAlive = false;
+	DWORD code;
+
+	h = OpenProcess(PROCESS_QUERY_INFORMATION, FALSE, pid);
+	if (h) {
+		if (GetExitCodeProcess(h, &code) && (code == STILL_ACTIVE))
+			isAlive = true;
+			
+		CloseHandle(h);
+	} else {
+		int err;
+
+		err = HRESULT_CODE(GetLastError());
+		if (err != ERROR_INVALID_PARAMETER)
+			isAlive = true;
+		else
+			fprintf(stderr, "ERROR CSSys::sys_isAlive(%d):OpenProcess %d\n", pid, err);
+	}
+
+	return(isAlive);	
+}
+
+
+
+//=====================
+// CSSysDir
+//=====================
+CSSysDir::~CSSysDir()
+{
+	close();
+	if (sd_path)
+		sd_path->release();
+		
+	if (sd_filter)
+		sd_filter->release();
+}
+
+//--------------
+void CSSysDir::open()
+{
+	enter_();
+
+	if (! CSSys::sys_exists(sd_path->getCString()))
+		CSException::throwFileError(CS_CONTEXT, sd_path->getCString(), ERROR_PATH_NOT_FOUND);
+
+	sd_filter = new CSStringBuffer();
+	sd_filter->append(sd_path->getCString());
+
+	if (IS_DIR_CHAR(*(sd_filter->getBuffer(sd_filter->length()-1))))
+		sd_filter->append("*");
+	else
+		sd_filter->append(CS_DIR_DELIM"*");
+
+	exit_();
+}
+
+//--------------
+void CSSysDir::close()
+{
+	if (sd_dir != INVALID_HANDLE_VALUE) {
+		FindClose(sd_dir);
+		sd_dir = INVALID_HANDLE_VALUE;
+	}
+}
+
+//--------------
+bool CSSysDir::next()
+{
+	int err = 0;
+
+	while (true) {
+		if (sd_dir == INVALID_HANDLE_VALUE) {
+			sd_dir = FindFirstFileA(sd_filter->getCString(), &sd_entry);
+			if (sd_dir == INVALID_HANDLE_VALUE)
+				err = get_win_error();
+		}
+		else {
+			if (!FindNextFileA(sd_dir, &sd_entry))
+				err = get_win_error();
+		}
+
+		if (err) {
+			if ((err != ERROR_NO_MORE_FILES) && (err != ERROR_FILE_NOT_FOUND)){
+				CSException::throwFileError(CS_CONTEXT, sd_path->getCString(), err);
+			}
+			return false;
+		}
+
+		/* Filter out '.' and '..': */
+		if (sd_entry.cFileName[0] == '.') {
+			if (sd_entry.cFileName[1] == '.') {
+				if (sd_entry.cFileName[2] == '\0')
+					continue;
+			}
+			else {
+				if (sd_entry.cFileName[1] == '\0')
+					continue;
+			}
+		}
+		break;
+	}
+
+	return true;
+}
+
+
+//--------------
+void CSSysDir::getEntryPath(char *path, size_t size)
+{
+	cs_strcpy(size, path, sd_path->getCString());
+	cs_add_dir_char(size, path);
+	cs_strcat(size, path, entryName());
+}
+
+//--------------
+const char *CSSysDir::entryName()
+{
+	return (const char*) sd_entry.cFileName;
+}
+
+//--------------
+bool CSSysDir::entryIsFile()
+{
+	if (sd_entry.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)
+		return false;
+	return true;
+}
+
+///////////////////////////////////////////
+// A windows version of gettimeofday() as taken from:
+// http://www.suacommunity.com/dictionary/gettimeofday-entry.php
+
+#if defined(_MSC_VER) || defined(_MSC_EXTENSIONS)
+  #define DELTA_EPOCH_IN_MICROSECS  11644473600000000Ui64
+#else
+  #define DELTA_EPOCH_IN_MICROSECS  11644473600000000ULL
+#endif
+ 
+struct timezone
+{
+  int  tz_minuteswest; /* minutes W of Greenwich */
+  int  tz_dsttime;     /* type of dst correction */
+};
+ 
+// Definition of a gettimeofday function
+ 
+int gettimeofday(struct timeval *tv, struct timezone *tz)
+{
+// Define a structure to receive the current Windows filetime
+  FILETIME ft;
+ 
+// Initialize the present time to 0 and the timezone to UTC
+  unsigned __int64 tmpres = 0;
+  static int tzflag = 0;
+ 
+  if (NULL != tv)
+  {
+    GetSystemTimeAsFileTime(&ft);
+ 
+// The GetSystemTimeAsFileTime returns the number of 100 nanosecond 
+// intervals since Jan 1, 1601 in a structure. Copy the high bits to 
+// the 64 bit tmpres, shift it left by 32 then or in the low 32 bits.
+    tmpres |= ft.dwHighDateTime;
+    tmpres <<= 32;
+    tmpres |= ft.dwLowDateTime;
+ 
+// Convert to microseconds by dividing by 10
+    tmpres /= 10;
+ 
+// The Unix epoch starts on Jan 1 1970.  Need to subtract the difference 
+// in seconds from Jan 1 1601.
+    tmpres -= DELTA_EPOCH_IN_MICROSECS;
+ 
+// Finally change microseconds to seconds and place in the seconds value. 
+// The modulus picks up the microseconds.
+    tv->tv_sec = (long)(tmpres / 1000000UL);
+    tv->tv_usec = (long)(tmpres % 1000000UL);
+  }
+ 
+  if (NULL != tz)
+  {
+    if (!tzflag)
+    {
+      _tzset();
+      tzflag++;
+    }
+  
+// Adjust for the timezone west of Greenwich
+      tz->tz_minuteswest = _timezone / 60;
+    tz->tz_dsttime = _daylight;
+  }
+ 
+  return 0;
+}
+

=== modified file 'plugin/pbms/src/cslib/CSThread.cc'
--- plugin/pbms/src/cslib/CSThread.cc	2010-10-02 21:15:42 +0000
+++ plugin/pbms/src/cslib/CSThread.cc	2010-10-20 20:21:41 +0000
@@ -30,12 +30,15 @@
 
 #ifdef OS_WINDOWS
 #include <signal.h>
-#include "uniwin.h"
+//#include "uniwin.h"
+#define SIGUSR1 30
+#define SIGUSR2 31
+
 #else
 #include <signal.h>
 #include <sys/signal.h>
-#endif
 #include <unistd.h>
+#endif
 #include <errno.h>
 
 #include "CSGlobal.h"
@@ -208,6 +211,7 @@
 void CSThread::removeFromList()
 {
 	if (myThreadList && isRunning) {
+		CSThread *myself = this; // pop_() wants to take a reference to its parameter.
 		enter_();
 		/* Retain the thread in order to ensure
 		 * that after it is removed from the list,
@@ -215,11 +219,11 @@
 		 * unlock_() call invalid, because it requires
 		 * on the thread.
 		 */
-		push_(this);
+		push_(myself);
 		lock_(myThreadList);
-		myThreadList->remove(RETAIN(this));
+		myThreadList->remove(RETAIN(myself));
 		unlock_(myThreadList);
-		pop_(this);
+		pop_(myself);
 		outer_();
 	}
 	this->release();
@@ -288,7 +292,7 @@
 	return NULL;
 }
 
-void CSThread::start()
+void CSThread::start(bool detached)
 {
 	int err;
 
@@ -303,6 +307,10 @@
 			break;
 		usleep(10);
 	}
+	
+	isDetached = detached;
+	if (detached)
+		pthread_detach(iThread);
 }
 
 void CSThread::stop()
@@ -313,12 +321,17 @@
 
 void *CSThread::join()
 {
-	void	*return_data;
+	void	*return_data = NULL;
 	int		err;
 
 	enter_();
-	if ((err = pthread_join(iThread, &return_data)))
+	if (isDetached) {
+		while (isRunning && !pthread_kill(iThread, 0)) 
+			usleep(100);
+	} else if ((err = pthread_join(iThread, &return_data))) {
 		CSException::throwOSError(CS_CONTEXT, err);
+	}
+
 	return_(return_data);
 }
 
@@ -334,6 +347,7 @@
 
 void CSThread::signal(unsigned int sig)
 {
+#ifndef OS_WINDOWS // Currently you cannot signal threads on windows.
 	int err;
 
 	setSignalPending(sig);
@@ -343,6 +357,7 @@
 		if (err != ESRCH) /* No such process */
 			CSException::throwOSError(CS_CONTEXT, err);
 	}
+#endif
 }
 
 void CSThread::throwSignal()
@@ -389,6 +404,14 @@
 				if (relTop->x.r_pooled)
 					relTop->x.r_pooled->returnToPool();
 				break;
+			case CS_RELEASE_MEM:
+				if (relTop->x.r_mem)
+					cs_free(relTop->x.r_mem);
+				break;
+			case CS_RELEASE_OBJECT_PTR:
+				if ((relTop->x.r_objectPtr) && (obj = *(relTop->x.r_objectPtr)))
+					obj->release();
+				break;
 		}
 	}
 }
@@ -463,7 +486,6 @@
 	isUp = false;
 	if ((err = pthread_key_create(&sThreadKey, NULL))) {
 		CSException::logOSError(CS_CONTEXT, errno);
-		return false;
 	} else
 		isUp = true;
 		
@@ -512,10 +534,14 @@
 		return (CSThread*) NULL;
 		
 #ifdef DEBUG
-	if (self->iRefCount == 0) {
+	/* PMC - Problem is, if this is called when releasing a
+	 * thread, then we have the reference count equal to
+	 * zero.
+	if (self && !self->iRefCount) {
 		pthread_setspecific(sThreadKey, NULL);
-		CSException::throwAssertion(CS_CONTEXT, "Bad self pointer.");
-	}	
+		CSException::throwAssertion(CS_CONTEXT, "Freed self pointer referenced.");
+	}
+	*/
 #endif
 
 	return self;
@@ -614,19 +640,10 @@
 {
 }
 
-void *CSDaemon::run()
+void CSDaemon::try_Run(CSThread *self, const bool c_must_sleep)
 {
-	bool must_sleep = false;
-
-	CLOBBER_PROTECT(must_sleep);
-
-	enter_();
-	CLOBBER_PROTECT(self);
-
-	myMustQuit = !initializeWork();
-
-	restart:
 	try_(a) {
+		 bool must_sleep = c_must_sleep; // This done to avoid longjmp() clobber.
 		while (!myMustQuit) {
 			if (must_sleep) {
 				lock_(this);
@@ -646,9 +663,19 @@
 			myMustQuit = true;
 	}
 	cont_(a);
-	if (!myMustQuit) {
+}
+
+void *CSDaemon::run()
+{
+	bool must_sleep = false;
+
+	enter_();
+
+	myMustQuit = !initializeWork();
+
+	while  (!myMustQuit) {
+		try_Run(self, must_sleep);
 		must_sleep = true;
-		goto restart;
 	}
 
 	/* Prevent signals from going off in completeWork! */

=== modified file 'plugin/pbms/src/cslib/CSThread.h'
--- plugin/pbms/src/cslib/CSThread.h	2010-10-02 21:15:42 +0000
+++ plugin/pbms/src/cslib/CSThread.h	2010-10-19 14:03:11 +0000
@@ -35,10 +35,9 @@
 #include "CSDefs.h"
 #include "CSMutex.h"
 #include "CSException.h"
+#include "CSObject.h"
 #include "CSStorage.h"
 
-using namespace std;
-
 #define CS_THREAD_TYPE				int
 
 /* Types of threads: */
@@ -59,6 +58,7 @@
 #define CS_RELEASE_MUTEX		2
 #define CS_RELEASE_POOLED		3
 #define CS_RELEASE_MEM			4
+#define CS_RELEASE_OBJECT_PTR	5
 
 typedef struct CSRelease {
 	int						r_type;
@@ -67,6 +67,7 @@
 		CSMutex				*r_mutex;					/* The mutex to be unlocked! */
 		CSPooled			*r_pooled;
 		void				*r_mem;
+		CSObject			**r_objectPtr;
 	} x;
 } CSReleaseRec, *CSReleasePtr;
 
@@ -124,24 +125,26 @@
 	/* Set to true when the thread must quit (never reset!): */
 	bool			myMustQuit;	
 	
+	CSException		myException;
+#if defined(MYSQL_SERVER) ||  defined(DRIZZLED)
+
 	/* Set to true when this tread was initialized through the internal PBMS api. */
 	/* When this is the case than it must only be freed via the API as well. */
 	bool			pbms_api_owner;
 
-	CSException		myException;
-
 	/* Transaction references. */
 #ifdef DRIZZLED
 	CSSortedList	mySavePoints;
 #endif
-	uint32_t			myTID;			// Current transaction ID
-	uint32_t			myTransRef;		// Reference to the current transaction cache index
+	uint32_t		myTID;			// Current transaction ID
+	uint32_t		myTransRef;		// Reference to the current transaction cache index
 	bool			myIsAutoCommit;	// Is the current transaction in auto commit mode.
-	uint32_t			myCacheVersion;	// The last transaction cache version checked. Used during overflow.
+	uint32_t		myCacheVersion;	// The last transaction cache version checked. Used during overflow.
 	bool			myStartTxn;		// A flag to indicate the start of a new transaction.
-	uint32_t			myStmtCount;	// Counts the number of statements in the current transaction.
-	uint32_t			myStartStmt;	// The myStmtCount at the start of the last logical statement. (An update is 2 statements but only 1 logical statement.)
-	void			*myInfo;		// A place to hang some info. (Be carefull with this!)
+	uint32_t		myStmtCount;	// Counts the number of statements in the current transaction.
+	uint32_t		myStartStmt;	// The myStmtCount at the start of the last logical statement. (An update is 2 statements but only 1 logical statement.)
+	void			*myInfo;
+#endif
 	
 	/* The call stack */
 	int				callTop;
@@ -163,6 +166,7 @@
 		ignoreSignals(false),
 		isRunning(false),
 		myMustQuit(false),
+#if defined(MYSQL_SERVER) ||  defined(DRIZZLED)
 		pbms_api_owner(false),
 		myTID(0),
 		myTransRef(0),
@@ -172,10 +176,12 @@
 		myStmtCount(0),
 		myStartStmt(0),
 		myInfo(NULL),
+#endif
 		callTop(0),
 		jumpDepth(0),
 		relTop(relStack),
 		iIsMain(false),
+		isDetached(false),
 		iRunFunc(NULL),
 		iNextLink(NULL),
 		iPrevLink(NULL)
@@ -199,7 +205,7 @@
 	 *
      * @exception CSSystemException thrown if thread is invalid.
 	 */
-	void start();
+	void start(bool detached = false);
 
 	/*
 	 * Stop execution of the thread.
@@ -250,7 +256,7 @@
 	void releaseObjects(CSReleasePtr top);
 	void throwException();
 	void caught();
-	bool isMe(CSThread *me) { return (me->iThread == iThread);}
+	bool isMe(CSThread *me) { return (pthread_equal(me->iThread,iThread) != 0);}
 	/* Make this object linkable: */
 	virtual CSObject *getNextLink() { return iNextLink; }
 	virtual CSObject *getPrevLink() { return iPrevLink; }
@@ -262,6 +268,7 @@
 private:
 	pthread_t		iThread;
 	bool			iIsMain;
+	bool			isDetached;
 	ThreadRunFunc	iRunFunc;
 	CSObject		*iNextLink;
 	CSObject		*iPrevLink;
@@ -318,12 +325,18 @@
 
 	virtual void *run();
 
+	/* Return false if startup failed, and the thread must quit. */
 	virtual bool initializeWork() { return true; };
 
+	/* Return true of the thread should sleep before doing more work. */
 	virtual bool doWork();
 
 	virtual void *completeWork() { return NULL; };
 
+	/* Return false if the excpetion is not handled and the thread must quit.
+	 * Set must_sleep to true of the thread should pause before doing work
+	 * again.
+	 */
 	virtual bool handleException();
 
 	virtual void stop();
@@ -348,8 +361,9 @@
 	void suspendedWait(time_t milli_sec);
 
 private:
-	bool			iSuspended;
-	uint32_t			iSuspendCount;
+	void		try_Run(CSThread *self, const bool must_sleep);
+	bool		iSuspended;
+	uint32_t	iSuspendCount;
 };
 
 #endif

=== modified file 'plugin/pbms/src/cslib/CSTime.cc'
--- plugin/pbms/src/cslib/CSTime.cc	2010-10-02 21:15:42 +0000
+++ plugin/pbms/src/cslib/CSTime.cc	2010-10-18 21:21:20 +0000
@@ -27,10 +27,12 @@
 #include "CSConfig.h"
 
 #include <string.h>
-#include <time.h>
 
 #ifdef OS_WINDOWS
 #include <Windows.h>
+extern int gettimeofday(struct timeval *tv, struct timezone *tz);
+#else
+#include <sys/time.h>
 #endif
 
 #include "CSTime.h"
@@ -186,6 +188,18 @@
 	setUTC(ltime.tm_year + 1900, ltime.tm_mon + 1, ltime.tm_mday, ltime.tm_hour, ltime.tm_min, ltime.tm_sec, nsec);
 }
 
+bool CSTime::olderThen(time_t max_age)
+{
+	time_t secs, now;
+	s_int nsec;
+	
+	getUTC1970(secs, nsec);
+	
+	now = time(NULL);
+	
+	return ((now - secs) > max_age);
+}
+
 void CSTime::getUTC1970(time_t& sec, s_int& nsec)
 {
 #ifdef OS_WINDOWS
@@ -267,4 +281,14 @@
 
 #endif
 
+uint64_t CSTime::getTimeCurrentTicks()
+{
+	struct timeval	now;
+
+	/* Get the current time in microseconds: */
+	gettimeofday(&now, NULL);
+	return (uint64_t) now.tv_sec * (uint64_t) 1000000 + (uint64_t) now.tv_usec;
+}
+
+
 

=== modified file 'plugin/pbms/src/cslib/CSTime.h'
--- plugin/pbms/src/cslib/CSTime.h	2010-10-02 21:15:42 +0000
+++ plugin/pbms/src/cslib/CSTime.h	2010-10-18 21:21:20 +0000
@@ -36,8 +36,6 @@
 #include "CSDefs.h"
 #include "CSObject.h"
 
-using namespace std;
-
 class CSTime : public CSObject  {
 public:
 	CSTime(): iIsNull(true) { }
@@ -91,6 +89,12 @@
 	void setUTC1601(uint64_t nsec100);
 	uint64_t getUTC1601();
 
+	/* 
+	 * Tests if the time is more than 'max_age' seconds in the past.
+	 */
+	bool olderThen(time_t max_age);
+
+	static	uint64_t getTimeCurrentTicks();
 private:
 	bool	iIsNull;
 	char	iCString[100];

=== modified file 'plugin/pbms/src/cslib/CSXML.cc'
--- plugin/pbms/src/cslib/CSXML.cc	2010-10-02 21:15:42 +0000
+++ plugin/pbms/src/cslib/CSXML.cc	2010-10-18 21:21:20 +0000
@@ -1293,13 +1293,13 @@
 
 	switch (err) {
 		case CS_XML_ERR_OUT_OF_MEMORY:
-			sprintf(err_message, "AES parse error- insufficient memory");			
+			snprintf(err_message, CS_XML_ERR_MSG_SIZE, "AES parse error- insufficient memory");			
 			break;
 		case CS_XML_ERR_CHAR_TOO_LARGE:
-			sprintf(err_message, "AES parse error- UNICODE character too large to be encoded as UTF-8");			
+			snprintf(err_message, CS_XML_ERR_MSG_SIZE, "AES parse error- UNICODE character too large to be encoded as UTF-8");			
 			break;
 		default:
-			sprintf(err_message, "AES parse error- %s", strerror(err));
+			snprintf(err_message, CS_XML_ERR_MSG_SIZE, "AES parse error- %s", strerror(err));
 			break;
 	}
 }

=== modified file 'plugin/pbms/src/database_ms.cc'
--- plugin/pbms/src/database_ms.cc	2010-10-05 18:58:54 +0000
+++ plugin/pbms/src/database_ms.cc	2010-10-18 21:21:20 +0000
@@ -545,11 +545,11 @@
 	push_(file);
 	if ((repo = file->myRepo)) {
 		if (repo->isRemovingFP) {
-			repo->removeRepoFile(file);
+			repo->removeRepoFile(RETAIN(file));
 			myRepostoryList->wakeup();
 		}
 		else
-			repo->returnRepoFile(file);
+			repo->returnRepoFile(RETAIN(file));
 		repo->release(); /* [++] here is the release.  */
 	}
 	release_(file);
@@ -1076,7 +1076,7 @@
 						int len = ptr - dir_name;
 						ptr++;
 						if ((strtoul(ptr, NULL, 10) == db_id) && len) {
-							db = getDatabase(CSCString::newString(dir_name, len), true);
+							db = getDatabase(CSString::newString(dir_name, len), true);
 							ASSERT(db->myDatabaseID == db_id);
 						}
 					}
@@ -1259,7 +1259,7 @@
 	 * The database name is case sensitive here if the file system names are
 	 * case sensitive. This is desirable.
 	 */
-	path = CSPath::newPath(ms_my_get_mysql_home_path().c_str(), RETAIN(db_name));
+	path = CSPath::newPath(ms_my_get_mysql_home_path(), RETAIN(db_name));
 	push_(path);
 	if (create && !path->exists()) {
 		CSException::throwException(CS_CONTEXT, MS_ERR_UNKNOWN_DB, db_name->getCString());
@@ -1597,7 +1597,7 @@
 			gDatabaseList->remove(doomedDatabase->getKey());
 		if (!self->myMustQuit) 
 			unlock_(gDatabaseList); 
-		ASSERT(doomedDatabase->iRefCount == 1);
+		ASSERT(doomedDatabase->getRefCount() == 1);
 		release_(doomedDatabase);
 		
 	} else {
@@ -1605,7 +1605,7 @@
 		bool create = false;
 		uint32_t db_id;
 		
-		path = createDatabasePath(ms_my_get_mysql_home_path().c_str(), CSString::newString(db_name), &db_id, &create);
+		path = createDatabasePath(ms_my_get_mysql_home_path(), CSString::newString(db_name), &db_id, &create);
 		
 		if (path) {
 			MSTransactionManager::dropDatabase(db_id);
@@ -1636,7 +1636,7 @@
 // 3: <database>/<table>
 bool MSDatabase::convertTablePathToIDs(const char *table_path, uint32_t *db_id, uint32_t *tab_id, bool create) 
 {
-	const char	*base = ms_my_get_mysql_home_path().c_str();
+	const char	*base = ms_my_get_mysql_home_path();
 	CSString	*table_url;
 	CSString	*db_path = NULL;
 	CSString	*db_name = NULL;
@@ -1736,7 +1736,7 @@
 	MSDatabase *db;
 	enter_();
 	
-	db = newDatabase(ms_my_get_mysql_home_path().c_str(), db_name, 0, create);
+	db = newDatabase(ms_my_get_mysql_home_path(), db_name, 0, create);
 	
 	if (db) {
 		push_(db);

=== modified file 'plugin/pbms/src/database_ms.h'
--- plugin/pbms/src/database_ms.h	2010-10-07 01:20:26 +0000
+++ plugin/pbms/src/database_ms.h	2010-10-18 21:21:20 +0000
@@ -30,6 +30,10 @@
 #ifndef __DATABASE_MS_H__
 #define __DATABASE_MS_H__
 
+#include "cslib/CSDefs.h"
+#include "cslib/CSStorage.h"
+#include "cslib/CSStrUtil.h"
+
 #include "table_ms.h"
 #include "repository_ms.h"
 #ifdef HAVE_ALIAS_SUPPORT
@@ -39,10 +43,6 @@
 #include "compactor_ms.h"
 #include "cloud_ms.h"
 
-#include "cslib/CSDefs.h"
-#include "cslib/CSStorage.h"
-#include "cslib/CSStrUtil.h"
-
 class MSOpenTable;
 class MSBackup;
 

=== modified file 'plugin/pbms/src/defs_ms.h'
--- plugin/pbms/src/defs_ms.h	2010-10-07 01:20:26 +0000
+++ plugin/pbms/src/defs_ms.h	2010-10-18 21:21:20 +0000
@@ -28,8 +28,6 @@
 #ifndef __DEFS_MS_H__
 #define __DEFS_MS_H__
 
-#include <boost/dynamic_bitset.hpp>
-
 #include "cslib/CSDefs.h"
 
 #define MS_IDENTIFIER_CHAR_COUNT	64
@@ -117,6 +115,7 @@
 #define MS_DEFAULT_KEEP_ALIVE		(10) 
 
 #ifdef DRIZZLED
+#include <boost/dynamic_bitset.hpp>
 /* Drizzle is stuck at this level: */
 #define MYSQL_VERSION_ID					60005
 

=== modified file 'plugin/pbms/src/discover_ms.cc'
--- plugin/pbms/src/discover_ms.cc	2010-10-02 21:15:42 +0000
+++ plugin/pbms/src/discover_ms.cc	2010-10-18 21:21:20 +0000
@@ -1148,7 +1148,7 @@
     goto err;
   }
 
-  LOCK_open.lock;
+  pthread_mutex_lock(&LOCK_open);
   if (!internal_tmp_table && !(create_info->options & HA_LEX_CREATE_TMP_TABLE))
   {
     if (!access(path,F_OK))
@@ -1256,7 +1256,7 @@
 #endif
   error= FALSE;
 unlock_and_end:
-  LOCK_open.unlock;
+  pthread_mutex_unlock(&LOCK_open);
 
 err:
   thd_proc_info(thd, "After create");

=== modified file 'plugin/pbms/src/discover_ms.h'
--- plugin/pbms/src/discover_ms.h	2010-10-02 21:15:42 +0000
+++ plugin/pbms/src/discover_ms.h	2010-10-18 21:21:20 +0000
@@ -30,6 +30,7 @@
  * ---------------------------------------------------------------
  * TABLE DISCOVERY HANDLER
  */
+#define NOVAL 0
 
 typedef struct dt_field_info
 	{

=== modified file 'plugin/pbms/src/engine_ms.cc'
--- plugin/pbms/src/engine_ms.cc	2010-10-02 21:15:42 +0000
+++ plugin/pbms/src/engine_ms.cc	2010-10-18 21:21:20 +0000
@@ -202,26 +202,19 @@
 		}
 	}
 	
-	return NULL;
+	return (const PBMSEnginePtr)NULL;
 }
 #endif	
 
 //---------------
-int32_t	MSEngine::createBlob(const char *db_name, const char *tab_name, char *blob, size_t blob_len, PBMSBlobURLPtr blob_url, PBMSResultPtr result)
+bool MSEngine::try_createBlob(CSThread *self, const char *db_name, const char *tab_name, char *blob, size_t blob_len, PBMSBlobURLPtr blob_url)
 {
-
-	CSThread		*self;
-	int32_t			err = MS_OK;
-	MSOpenTable		*otab;
-	CSInputStream	*i_stream = NULL;
+	volatile bool rtc = true;
 	
-	CLOBBER_PROTECT(err);
-
-	if ((err = enterConnectionNoThd(&self, result)))
-		return err;
-
-	inner_();
 	try_(a) {
+		MSOpenTable		*otab;
+		CSInputStream	*i_stream = NULL;
+		
 		otab = openTable(db_name, tab_name, true);
 		frompool_(otab);
 		
@@ -232,68 +225,85 @@
 			CSException::throwException(CS_CONTEXT, MS_ERR_RECOVERY_IN_PROGRESS, "Cannot create BLOBs during repository recovery.");
 
 		backtopool_(otab);
+		rtc = false;			
 	}
-	catch_(a) {
+	catch_(a);
+	cont_(a);
+	return rtc;
+}
+
+//---------------
+int32_t	MSEngine::createBlob(const char *db_name, const char *tab_name, char *blob, size_t blob_len, PBMSBlobURLPtr blob_url, PBMSResultPtr result)
+{
+
+	CSThread		*self;
+	int32_t			err = MS_OK;
+	
+	if ((err = enterConnectionNoThd(&self, result)))
+		return err;
+
+	inner_();
+	if (try_createBlob(self, db_name, tab_name, blob, blob_len, blob_url))
 		err = exceptionToResult(&self->myException, result);
-	}
-	cont_(a);
+
 	return_(err);
 }
 
 //---------------
+bool MSEngine::try_referenceBlob(CSThread *self, const char *db_name, const char *tab_name, PBMSBlobURLPtr ret_blob_url, char *blob_url, uint16_t col_index)
+{
+	volatile bool rtc = true;
+	try_(a) {
+		MSBlobURLRec	blob;
+		MSOpenTable		*otab;
+		
+		if (! PBMSBlobURLTools::couldBeURL(blob_url, &blob)){
+			char buffer[CS_EXC_MESSAGE_SIZE];
+			
+			cs_strcpy(CS_EXC_MESSAGE_SIZE, buffer, "Incorrect URL: ");
+			cs_strcat(CS_EXC_MESSAGE_SIZE, buffer, blob_url);
+			CSException::throwException(CS_CONTEXT, MS_ERR_INCORRECT_URL, buffer);
+		}
+		
+		otab = openTable(db_name, tab_name, true);
+		frompool_(otab);
+
+		otab->useBlob(blob.bu_type, blob.bu_db_id, blob.bu_tab_id, blob.bu_blob_id, blob.bu_auth_code, col_index, blob.bu_blob_size, blob.bu_blob_ref_id, ret_blob_url);
+
+		backtopool_(otab);
+		rtc = false;			
+	}
+	catch_(a);
+	cont_(a);
+	return rtc;
+}
+
+//---------------
 int32_t	MSEngine::referenceBlob(const char *db_name, const char *tab_name, PBMSBlobURLPtr ret_blob_url, char *blob_url, uint16_t col_index, PBMSResultPtr result)
 {
 
 	CSThread		*self;
 	int32_t			err = MS_OK;
-	MSBlobURLRec	blob;
-	MSOpenTable		*otab;
 	
-	CLOBBER_PROTECT(err);
-
 	if ((err = enterConnectionNoThd(&self, result)))
 		return err;
 
 	inner_();
-	try_(a) {
-		
-		if (! PBMSBlobURLTools::couldBeURL(blob_url, &blob)){
-			char buffer[CS_EXC_MESSAGE_SIZE];
-			
-			cs_strcpy(CS_EXC_MESSAGE_SIZE, buffer, "Incorrect URL: ");
-			cs_strcat(CS_EXC_MESSAGE_SIZE, buffer, blob_url);
-			CSException::throwException(CS_CONTEXT, MS_ERR_INCORRECT_URL, buffer);
-		}
-		
-		otab = openTable(db_name, tab_name, true);
-		frompool_(otab);
-
-		otab->useBlob(blob.bu_type, blob.bu_db_id, blob.bu_tab_id, blob.bu_blob_id, blob.bu_auth_code, col_index, blob.bu_blob_size, blob.bu_blob_ref_id, ret_blob_url);
-
-		backtopool_(otab);
-	}
-	catch_(a) {
+	if (try_referenceBlob(self, db_name, tab_name, ret_blob_url, blob_url, col_index))
 		err = exceptionToResult(&self->myException, result);
-	}
-	cont_(a);
+
 	return_(err);
+
 }
 
 //---------------
-int32_t	MSEngine::dereferenceBlob(const char *db_name, const char *tab_name, char *blob_url, PBMSResultPtr result)
+bool MSEngine::try_dereferenceBlob(CSThread *self, const char *db_name, const char *tab_name, char *blob_url)
 {
-	CSThread		*self;
-	int32_t			err = MS_OK;
-	MSBlobURLRec	blob;
-	MSOpenTable		*otab;
-
-	CLOBBER_PROTECT(err);
-
-	if ((err = enterConnectionNoThd(&self, result)))
-		return err;
-
-	inner_();
+	volatile bool rtc = true;
 	try_(a) {
+		MSBlobURLRec	blob;
+		MSOpenTable		*otab;
+
 		if (! PBMSBlobURLTools::couldBeURL(blob_url, &blob)){
 			char buffer[CS_EXC_MESSAGE_SIZE];
 
@@ -323,35 +333,58 @@
 			CSException::throwException(CS_CONTEXT, MS_ERR_INCORRECT_URL, buffer);
 		}
 		
-		backtopool_(otab);				
+		backtopool_(otab);	
+		rtc = false;			
 	}
-	catch_(a) {
+	catch_(a);
+	cont_(a);
+	return rtc;
+}
+
+int32_t	MSEngine::dereferenceBlob(const char *db_name, const char *tab_name, char *blob_url, PBMSResultPtr result)
+{
+	CSThread		*self;
+	int32_t			err = MS_OK;
+
+	if ((err = enterConnectionNoThd(&self, result)))
+		return err;
+
+	inner_();
+	if (try_dereferenceBlob(self, db_name, tab_name, blob_url))
 		err = exceptionToResult(&self->myException, result);
-	}
-	cont_(a);
+
 	return_(err);
 }
 
+bool MSEngine::try_dropDatabase(CSThread *self, const char *db_name)
+{
+	volatile bool rtc = true;
+	try_(a) {
+		MSDatabase::dropDatabase(db_name);
+		rtc = false;
+	}
+	catch_(a);
+	cont_(a);
+	
+	return rtc;
+}
+
 int32_t MSEngine::dropDatabase(const char *db_name, PBMSResultPtr result)
 {
 	CSThread *self;
-	int		err = MS_OK;;
+	int		err = MS_OK;
 	
-	CLOBBER_PROTECT(err);
-
 	if ((err = enterConnectionNoThd(&self, result)))
 		return err;
+
 	inner_();
 	
-	try_(a) {
-		MSDatabase::dropDatabase(db_name);
-	}
-	catch_(a) {
+	if (try_dropDatabase(self, db_name))
 		err = exceptionToResult(&self->myException, result);
-	}
-	cont_(a);
+
 	return_(err);
 }
+
 //---------------
 typedef struct UnDoInfo {
 	bool udo_WasRename;
@@ -361,17 +394,10 @@
 	CSString *udo_NewName;
 } UnDoInfoRec, *UnDoInfoPtr;
 
-int32_t	MSEngine::dropTable(const char *db_name, const char *tab_name, PBMSResultPtr result)
+//---------------
+bool MSEngine::try_dropTable(CSThread *self, const char *db_name, const char *tab_name)
 {
-	CSThread	*self;
-	int			err = MS_OK;
-
-	CLOBBER_PROTECT(err);
-
-	if ((err = enterConnectionNoThd(&self, result)))
-		return err;
-
-	inner_();
+	volatile bool rtc = true;
 	try_(a) {
 
 		CSPath			*new_path;
@@ -382,8 +408,9 @@
 		UnDoInfoPtr		undo_info = NULL;
 
 		otab = openTable(db_name, tab_name, false);
-		if (!otab)
-			goto exit;
+		if (!otab) {
+			goto end_try;
+		}
 		
 		// If we are recovering do not delete the table.
 		// It is normal for MySQL recovery scripts to delete any table they aare about to
@@ -391,7 +418,7 @@
 		// then this would delete all the recovered BLOBs in the table.
 		if (otab->getDB()->isRecovering()) {
 			otab->returnToPool();
-			goto exit;
+			goto end_try;
 		}
 
 		frompool_(otab);
@@ -421,7 +448,7 @@
 		frompool_(tab_pool);
 
 		if (old_path->exists())
-			old_path->move(new_path);
+			old_path->move(RETAIN(new_path));
 		tab->myDatabase->dropTable(RETAIN(tab));
 		
 		/* Add the table to the temp delete list if we are not recovering... */
@@ -437,15 +464,28 @@
 		
 		undo_info->udo_WasRename = false;
 		self->myInfo = undo_info;
-		
-		
-exit: ;
+				
+end_try:
+		rtc = false;	
 	}
-	
-	catch_(a) {
+	catch_(a);
+	cont_(a);
+	return rtc;
+}
+
+//---------------
+int32_t	MSEngine::dropTable(const char *db_name, const char *tab_name, PBMSResultPtr result)
+{
+	CSThread	*self;
+	int			err = MS_OK;
+
+	if ((err = enterConnectionNoThd(&self, result)))
+		return err;
+
+	inner_();
+	if (try_dropTable(self, db_name, tab_name))
 		err = exceptionToResult(&self->myException, result);
-	}
-	cont_(a);
+
 	outer_();
 	exitConnection();
 	return err;
@@ -506,7 +546,7 @@
 	tab_pool = MSTableList::lockTablePoolForDeletion(otab);
 	frompool_(tab_pool);
 
-	from_path->move(to_path);
+	from_path->move(RETAIN(to_path));
 	tab->myDatabase->renameTable(tab, to_table);
 
 	backtopool_(tab_pool);	// The will unlock and close the table pool freeing all tables in it.
@@ -518,17 +558,9 @@
 }
 
 //---------------
-int32_t	MSEngine::renameTable(const char *from_db_name, const char *from_table, const char *to_db_name, const char *to_table, PBMSResultPtr result)
+bool MSEngine::try_renameTable(CSThread *self, const char *from_db_name, const char *from_table, const char *to_db_name, const char *to_table)
 {
-	CSThread	*self;
-	int err = MS_OK;
-
-	CLOBBER_PROTECT(err);
-
-	if ((err = enterConnectionNoThd(&self, result)))
-		return err;
-
-	inner_();
+	volatile bool rtc = true;
 	try_(a) {
 		UnDoInfoPtr undo_info = (UnDoInfoPtr) cs_malloc(sizeof(UnDoInfoRec));
 		push_ptr_(undo_info);
@@ -554,11 +586,26 @@
 		}
 		self->myInfo = undo_info;
 		pop_(undo_info);
+		rtc = false;			
 	}
-	catch_(a) {
+	catch_(a);
+	cont_(a);
+	return rtc;
+}
+
+//---------------
+int32_t	MSEngine::renameTable(const char *from_db_name, const char *from_table, const char *to_db_name, const char *to_table, PBMSResultPtr result)
+{
+	CSThread	*self;
+	int err = MS_OK;
+
+	if ((err = enterConnectionNoThd(&self, result)))
+		return err;
+
+	inner_();
+	if (try_renameTable(self, from_db_name, from_table, to_db_name, to_table))
 		err = exceptionToResult(&self->myException, result);
-	}
-	cont_(a);
+
 	outer_();
 	exitConnection();
 	return err;
@@ -591,6 +638,26 @@
 	}
 	exit_();
 }
+
+//---------------
+static bool try_CompleteTransaction(CSThread *self, bool ok)
+{
+	volatile bool rtc = true;
+	try_(a) {
+		if (ok)
+			MSTransactionManager::commit();
+		else if (self->myIsAutoCommit)
+			MSTransactionManager::rollback();
+		else
+			MSTransactionManager::rollbackToPosition(self->myStartStmt); // Rollback the last logical statement.
+		rtc = false;
+	}
+	catch_(a)
+	cont_(a);
+	
+	return rtc;
+}
+
 //---------------
 void MSEngine::callCompleted(bool ok)
 {
@@ -611,18 +678,9 @@
 		self->myInfo = NULL;
 	} else if (self->myTID && (self->myIsAutoCommit || !ok)) {
 		inner_();
-		try_(a) {
-			if (ok)
-				MSTransactionManager::commit();
-			else if (self->myIsAutoCommit)
-				MSTransactionManager::rollback();
-			else
-				MSTransactionManager::rollbackToPosition(self->myStartStmt); // Rollback the last logical statement.
-		}
-		catch_(a) {
+		if (try_CompleteTransaction(self, ok)) {
 			self->logException();
 		}
-		cont_(a);
 		outer_();
 	}
 	

=== modified file 'plugin/pbms/src/engine_ms.h'
--- plugin/pbms/src/engine_ms.h	2010-10-07 01:20:26 +0000
+++ plugin/pbms/src/engine_ms.h	2010-10-18 21:21:20 +0000
@@ -32,14 +32,6 @@
 
 #include "defs_ms.h"
 
-#include "cslib/CSConfig.h"
-#include <inttypes.h>
-
-#include "cslib/CSGlobal.h"
-#include "cslib/CSSocket.h"
-#include "cslib/CSStrUtil.h"
-#include "cslib/CSHTTPStream.h"
-
 #include "pbms.h"
 class MSOpenTable;
 
@@ -85,6 +77,13 @@
 	static bool couldBeURL(const char *url, size_t length);
 	
 	private:
+	static bool try_createBlob(CSThread *self, const char *db_name, const char *tab_name, char *blob, size_t blob_len, PBMSBlobURLPtr blob_url);
+	static bool try_referenceBlob(CSThread *self, const char *db_name, const char *tab_name, PBMSBlobURLPtr ret_blob_url, char *blob_url, uint16_t col_index);
+	static bool try_dereferenceBlob(CSThread *self, const char *db_name, const char *tab_name, char *blob_url);
+	static bool try_dropDatabase(CSThread *self, const char *db_name);
+	static bool try_dropTable(CSThread *self, const char *db_name, const char *tab_name);
+	static bool try_renameTable(CSThread *self, const char *from_db_name, const char *from_table, const char *to_db_name, const char *to_table);
+	
 	static MSOpenTable *openTable(const char *db_name, const char *tab_name, bool create);
 	static bool renameTable(const char *db_name, const char *from_table, const char *to_db_name, const char *to_table);
 	static void completeRenameTable(struct UnDoInfo *info, bool ok);

=== modified file 'plugin/pbms/src/ha_pbms.cc'
--- plugin/pbms/src/ha_pbms.cc	2010-10-02 21:15:42 +0000
+++ plugin/pbms/src/ha_pbms.cc	2010-10-18 21:21:20 +0000
@@ -697,7 +697,11 @@
 	inner_();
 	try_(a) {
 		ha_open_tab = MSSystemTableShare::openSystemTable(table_path, table);
+#ifdef DRIZZLED
 		ha_lock.init(&ha_open_tab->myShare->myThrLock);
+#else
+		thr_lock_data_init(&ha_open_tab->myShare->myThrLock, &ha_lock, NULL);
+#endif
 		ref_length = ha_open_tab->getRefLen();
 	}
 	catch_(a) {

=== modified file 'plugin/pbms/src/mysql_ms.cc'
--- plugin/pbms/src/mysql_ms.cc	2010-10-05 18:58:54 +0000
+++ plugin/pbms/src/mysql_ms.cc	2010-10-18 21:21:20 +0000
@@ -56,9 +56,9 @@
 }
 
 #ifdef DRIZZLED
-const std::string ms_my_get_mysql_home_path()
+const char *ms_my_get_mysql_home_path()
 {
-	return drizzled::getDataHomeCatalog().file_string();
+	return drizzled::getDataHomeCatalog().file_string().c_str();
 }
 
 bool ms_is_autocommit()
@@ -67,7 +67,7 @@
 }
 
 #else
-const std::string ms_my_get_mysql_home_path()
+const char *ms_my_get_mysql_home_path()
 {
 	return mysql_real_data_home;
 }

=== modified file 'plugin/pbms/src/mysql_ms.h'
--- plugin/pbms/src/mysql_ms.h	2010-10-05 18:58:54 +0000
+++ plugin/pbms/src/mysql_ms.h	2010-10-18 21:21:20 +0000
@@ -30,11 +30,9 @@
 #ifndef __MS_MYSQL_H__
 #define __MS_MYSQL_H__
 
-#include <string>
-
 void		*ms_my_get_thread();
 uint64_t	ms_my_1970_to_mysql_time(time_t t);
-const std::string ms_my_get_mysql_home_path();
+const char	*ms_my_get_mysql_home_path();
 void		ms_my_set_column_name(const char *table, uint16_t col_index, char *col_name);
 bool		ms_is_autocommit();
 

=== modified file 'plugin/pbms/src/network_ms.cc'
--- plugin/pbms/src/network_ms.cc	2010-10-07 01:20:26 +0000
+++ plugin/pbms/src/network_ms.cc	2010-10-20 20:21:41 +0000
@@ -26,16 +26,15 @@
  * Network interface.
  *
  */
-
-#include "config.h"
-
-#include "network_ms.h"
-#include "connection_handler_ms.h"
-
 #include "cslib/CSConfig.h"
+
+#include "defs_ms.h"
+
 #include "cslib/CSGlobal.h"
 #include "cslib/CSLog.h"
 
+#include "network_ms.h"
+#include "connection_handler_ms.h"
 
 MSSystemThread		*MSNetwork::gSystemThread;
 time_t				MSNetwork::gCurrentTime;
@@ -171,16 +170,46 @@
  * Return NULL of a connection cannot be openned, and the
  * thread must quit.
  */
+class OpenConnectioCleanUp : public CSRefObject {
+	bool do_cleanup;
+
+	public:
+	
+	OpenConnectioCleanUp(): CSRefObject(),
+		do_cleanup(false){}
+		
+	~OpenConnectioCleanUp() 
+	{
+		if (do_cleanup) {
+			MSNetwork::unlockListenerSocket();
+		}
+	}
+	
+	void setCleanUp()
+	{
+		do_cleanup = true;
+	}
+	
+	void cancelCleanUp()
+	{
+		do_cleanup = false;
+	}
+	
+};
+
+/*
+ * Return NULL if a connection cannot be openned, and the
+ * thread must quit.
+ */
 CSSocket *MSNetwork::openConnection(MSConnectionHandler *handler)
 {
 	CSSocket *sock = NULL;
-
-	CLOBBER_PROTECT(sock);
+	OpenConnectioCleanUp *cleanup;
 
 	enter_();
 	
 	if(!MSNetwork::gListenerSocket) {
-		return NULL;
+		return_(NULL);
 	}
 	
 	sock = CSSocket::newSocket();
@@ -192,19 +221,18 @@
 		return_(NULL);
 	}
 
-	try_(a) {
-		sock->open(MSNetwork::gListenerSocket);
-	}
-	catch_(a) {
-		unlockListenerSocket();
-		throw_();
-	}
-	cont_(a);
+	new_(cleanup, OpenConnectioCleanUp());
+	push_(cleanup);
+	
+	cleanup->setCleanUp();
+	sock->open(MSNetwork::gListenerSocket);
+	cleanup->cancelCleanUp();
 
 	handler->lastUse = gCurrentTime;
 
 	unlockListenerSocket();
 
+	release_(cleanup);
 	pop_(sock);
 	return_(sock);
 }

=== modified file 'plugin/pbms/src/network_ms.h'
--- plugin/pbms/src/network_ms.h	2010-10-07 01:20:26 +0000
+++ plugin/pbms/src/network_ms.h	2010-10-18 21:21:20 +0000
@@ -30,10 +30,10 @@
 #ifndef __NETWORK_MS_H__
 #define __NETWORK_MS_H__
 
-#include "connection_handler_ms.h"
 #include "cslib/CSDefs.h"
 #include "cslib/CSSocket.h"
 #include "cslib/CSThread.h"
+#include "connection_handler_ms.h"
 
 class MSSystemThread : public CSDaemon {
 public:

=== modified file 'plugin/pbms/src/open_table_ms.cc'
--- plugin/pbms/src/open_table_ms.cc	2010-10-07 01:20:26 +0000
+++ plugin/pbms/src/open_table_ms.cc	2010-10-18 21:21:20 +0000
@@ -26,8 +26,14 @@
  * Media Stream Tables.
  *
  */
-
-#include "config.h"
+#include "cslib/CSConfig.h"
+
+#include "defs_ms.h"
+
+#include "cslib/CSGlobal.h"
+#include "cslib/CSLog.h"
+#include "cslib/CSStrUtil.h"
+#include "cslib/CSPath.h"
 
 #include "open_table_ms.h"
 #include "table_ms.h"
@@ -36,14 +42,6 @@
 #include "transaction_ms.h"
 #include "parameters_ms.h"
 
-#include "cslib/CSConfig.h"
-
-#include "cslib/CSGlobal.h"
-#include "cslib/CSLog.h"
-#include "cslib/CSStrUtil.h"
-#include "cslib/CSPath.h"
-
-
 /*
  * ---------------------------------------------------------------
  * OPEN TABLES
@@ -101,6 +99,42 @@
 	MSTableList::releaseTable(this);
 }
 
+// This cleanup class is used to reset the 
+// repository size if something goes wrong.
+class CreateBlobCleanUp : public CSRefObject {
+	bool do_cleanup;
+	uint64_t old_size;
+	MSOpenTable *ot;
+	MSRepository *repo;
+
+	public:
+	
+	CreateBlobCleanUp(): CSRefObject(),
+		do_cleanup(false){}
+		
+	~CreateBlobCleanUp() 
+	{
+		if (do_cleanup) {
+			repo->setRepoFileSize(ot, old_size);
+
+		}
+	}
+	
+	void setCleanUp(MSOpenTable *ot_arg, MSRepository *repo_arg, uint64_t size)
+	{
+		old_size = size;
+		repo = repo_arg;
+		ot = ot_arg;
+		do_cleanup = true;
+	}
+	
+	void cancelCleanUp()
+	{
+		do_cleanup = false;
+	}
+	
+};
+
 void MSOpenTable::createBlob(PBMSBlobURLPtr bh, uint64_t blob_size, char *metadata, uint16_t metadata_size, CSInputStream *stream, CloudKeyPtr cloud_key, Md5Digest *checksum)
 {
 	uint64_t repo_offset;
@@ -114,11 +148,12 @@
 	uint64_t repo_id;
 	Md5Digest my_checksum;
 	CloudKeyRec cloud_key_rec;
+	CreateBlobCleanUp *cleanup;
 	enter_();
 	
-	CLOBBER_PROTECT(cloud_key);
-	CLOBBER_PROTECT(checksum);
-
+	new_(cleanup, CreateBlobCleanUp());
+	push_(cleanup);
+	
 	if (!checksum)
 		checksum = &my_checksum;
 		
@@ -129,51 +164,49 @@
 	repo_size = myWriteRepo->getRepoFileSize();
 	temp_time =	myWriteRepo->myLastTempTime;
 
-	try_(a) {
-		head_size = myWriteRepo->getDefaultHeaderSize(metadata_size);
-		if (getDB()->myBlobType == MS_STANDARD_STORAGE) {
+	// If an exception occurs the cleanup operation will be called.
+	cleanup->setCleanUp(this, myWriteRepo, repo_size);
+
+	head_size = myWriteRepo->getDefaultHeaderSize(metadata_size);
+	if (getDB()->myBlobType == MS_STANDARD_STORAGE) {
+		pop_(stream);
+		repo_offset = myWriteRepo->receiveBlob(this, head_size, blob_size, checksum, stream);
+	} else {
+		ASSERT(getDB()->myBlobType == MS_CLOUD_STORAGE);
+		CloudDB *cloud = getDB()->myBlobCloud;
+		
+		if (!cloud)
+			CSException::throwException(CS_CONTEXT, CS_ERR_GENERIC_ERROR, "Creating cloud BLOB without cloud.");
+	
+		repo_offset = repo_size + head_size;
+		memset(checksum, 0, sizeof(Md5Digest)); // The checksum is only for local storage.
+		
+		// If there is a stream then the data has not been sent to the cloud yet.
+		if (stream) { 
+			cloud_key = &cloud_key_rec;
+			cloud->cl_getNewKey(cloud_key);
 			pop_(stream);
-			repo_offset = myWriteRepo->receiveBlob(this, head_size, blob_size, checksum, stream);
-		} else {
-			ASSERT(getDB()->myBlobType == MS_CLOUD_STORAGE);
-			CloudDB *cloud = getDB()->myBlobCloud;
-			
-			if (!cloud)
-				CSException::throwException(CS_CONTEXT, CS_ERR_GENERIC_ERROR, "Creating cloud BLOB without cloud.");
-		
-			repo_offset = repo_size + head_size;
-			memset(checksum, 0, sizeof(Md5Digest)); // The checksum is only for local storage.
-			
-			// If there is a stream then the data has not been sent to the cloud yet.
-			if (stream) { 
-				cloud_key = &cloud_key_rec;
-				cloud->cl_getNewKey(cloud_key);
-				pop_(stream);
-				cloud->cl_putData(cloud_key, stream, blob_size);
-			}
-			
-		}
-		
-		repo_id = myWriteRepo->myRepoID;
-		if (isNotATable) {	
-			getDB()->queueForDeletion(this, MS_TL_REPO_REF, repo_id, repo_offset, auth_code, &log_id, &log_offset, &temp_time);
-			formatRepoURL(bh, repo_id, repo_offset, auth_code, blob_size);
-		}
-		else {
-			blob_id = getDBTable()->createBlobHandle(this, myWriteRepo->myRepoID, repo_offset, blob_size, head_size, auth_code);
-			getDB()->queueForDeletion(this, MS_TL_BLOB_REF, getDBTable()->myTableID, blob_id, auth_code, &log_id, &log_offset, &temp_time);
-			formatBlobURL(bh, blob_id, auth_code, blob_size, 0);
-		}
-		
-		myWriteRepo->writeBlobHead(this, repo_offset, myWriteRepo->myRepoDefRefSize, head_size, blob_size, checksum, metadata, metadata_size, blob_id, auth_code, log_id, log_offset, getDB()->myBlobType, cloud_key);
-	}
-	
-	catch_(a);
-	// In the event of an error reset the repository size to its original size.
-	myWriteRepo->setRepoFileSize(this, repo_size);
-	throw_();
-	
-	cont_(a);
+			cloud->cl_putData(cloud_key, stream, blob_size);
+		}
+		
+	}
+	
+	repo_id = myWriteRepo->myRepoID;
+	if (isNotATable) {	
+		getDB()->queueForDeletion(this, MS_TL_REPO_REF, repo_id, repo_offset, auth_code, &log_id, &log_offset, &temp_time);
+		formatRepoURL(bh, repo_id, repo_offset, auth_code, blob_size);
+	}
+	else {
+		blob_id = getDBTable()->createBlobHandle(this, myWriteRepo->myRepoID, repo_offset, blob_size, head_size, auth_code);
+		getDB()->queueForDeletion(this, MS_TL_BLOB_REF, getDBTable()->myTableID, blob_id, auth_code, &log_id, &log_offset, &temp_time);
+		formatBlobURL(bh, blob_id, auth_code, blob_size, 0);
+	}
+	
+	myWriteRepo->writeBlobHead(this, repo_offset, myWriteRepo->myRepoDefRefSize, head_size, blob_size, checksum, metadata, metadata_size, blob_id, auth_code, log_id, log_offset, getDB()->myBlobType, cloud_key);
+	
+	cleanup->cancelCleanUp();
+	release_(cleanup);
+	
 	exit_();
 }
 
@@ -188,40 +221,42 @@
 	uint32_t	log_id;
 	uint32_t log_offset;
 	uint32_t temp_time;
+	CreateBlobCleanUp *cleanup;
 	enter_();
+	
+	new_(cleanup, CreateBlobCleanUp());
+	push_(cleanup);
 
 	openForWriting();
 	ASSERT(myWriteRepo);
 	auth_code = random();
 	
 	repo_size = myWriteRepo->getRepoFileSize();
-	try_(a) {
-		head_size = myWriteRepo->getDefaultHeaderSize(metadata_size);
-
-		repo_offset = myWriteRepo->receiveBlob(this, head_size, blob_size);
-		repo_id = myWriteRepo->myRepoID;
-		temp_time = myWriteRepo->myLastTempTime;
-		getDB()->queueForDeletion(this, MS_TL_REPO_REF, repo_id, repo_offset, auth_code, &log_id, &log_offset, &temp_time);
-		myWriteRepo->myLastTempTime = temp_time;
-		myWriteRepo->writeBlobHead(this, repo_offset, myWriteRepo->myRepoDefRefSize, head_size, blob_size, NULL, metadata, metadata_size, 0, auth_code, log_id, log_offset, MS_STANDARD_STORAGE, NULL);
-		// myWriteRepo->setRepoFileSize(this, repo_offset + head_size + blob_size);This is now set by writeBlobHead()
-		
-		blob_id->bi_db_id = getDB()->myDatabaseID;
-		blob_id->bi_blob_id = repo_offset;
-		blob_id->bi_tab_id = repo_id;
-		blob_id->bi_auth_code = auth_code;
-		blob_id->bi_blob_size = blob_size;
-		blob_id->bi_blob_type = MS_URL_TYPE_REPO;
-		blob_id->bi_blob_ref_id = 0;
-	
-	}
-	
-	catch_(a);
-	// BLOBs created with this method are always created as standard local BLOBs. (No cloud storage)
-	myWriteRepo->setRepoFileSize(this, repo_size);
-	throw_();
-	
-	cont_(a);
+	
+	// If an exception occurs the cleanup operation will be called.
+	cleanup->setCleanUp(this, myWriteRepo, repo_size);
+
+	head_size = myWriteRepo->getDefaultHeaderSize(metadata_size);
+
+	repo_offset = myWriteRepo->receiveBlob(this, head_size, blob_size);
+	repo_id = myWriteRepo->myRepoID;
+	temp_time = myWriteRepo->myLastTempTime;
+	getDB()->queueForDeletion(this, MS_TL_REPO_REF, repo_id, repo_offset, auth_code, &log_id, &log_offset, &temp_time);
+	myWriteRepo->myLastTempTime = temp_time;
+	myWriteRepo->writeBlobHead(this, repo_offset, myWriteRepo->myRepoDefRefSize, head_size, blob_size, NULL, metadata, metadata_size, 0, auth_code, log_id, log_offset, MS_STANDARD_STORAGE, NULL);
+	// myWriteRepo->setRepoFileSize(this, repo_offset + head_size + blob_size);This is now set by writeBlobHead()
+	
+	blob_id->bi_db_id = getDB()->myDatabaseID;
+	blob_id->bi_blob_id = repo_offset;
+	blob_id->bi_tab_id = repo_id;
+	blob_id->bi_auth_code = auth_code;
+	blob_id->bi_blob_size = blob_size;
+	blob_id->bi_blob_type = MS_URL_TYPE_REPO;
+	blob_id->bi_blob_ref_id = 0;
+	
+	cleanup->cancelCleanUp();
+	release_(cleanup);
+
 	exit_();
 }
 

=== modified file 'plugin/pbms/src/open_table_ms.h'
--- plugin/pbms/src/open_table_ms.h	2010-10-07 01:20:26 +0000
+++ plugin/pbms/src/open_table_ms.h	2010-10-18 21:21:20 +0000
@@ -30,13 +30,13 @@
 #ifndef __OPENTABLE_MS_H__
 #define __OPENTABLE_MS_H__
 
+#include "cslib/CSDefs.h"
+#include "cslib/CSFile.h"
 
 #include "engine_ms.h"
+
 #include "database_ms.h"
 
-#include "cslib/CSDefs.h"
-#include "cslib/CSFile.h"
-
 class MSOpenTablePool;
 
 // Keep the value of MS_OT_BUFFER_SIZE less than 0XFFFF to avoid compiler warnings.

=== modified file 'plugin/pbms/src/parameters_ms.cc'
--- plugin/pbms/src/parameters_ms.cc	2010-10-02 21:15:42 +0000
+++ plugin/pbms/src/parameters_ms.cc	2010-10-18 21:21:20 +0000
@@ -30,6 +30,7 @@
 #include <drizzled/common.h>
 #include <drizzled/plugin.h>
 #include <drizzled/session.h>
+using namespace std;
 using namespace drizzled;
 using namespace drizzled::plugin;
 
@@ -44,6 +45,7 @@
 #endif 
 
 #include <inttypes.h>
+#include <string.h>
 
 #include "cslib/CSDefs.h"
 #include "cslib/CSObject.h"
@@ -316,6 +318,25 @@
 }		
 
 //-----------------
+bool PBMSParameters::try_LocateDB(CSThread *self, const char *db, bool *found)
+{
+	volatile bool rtc = true;
+	try_(a) {
+		lock_(&my_table_list_lock);	
+		
+			
+		*found = (locate_db(my_table_list, db, strlen(db)) != NULL);
+			
+		unlock_(&my_table_list_lock);
+		rtc = false;
+	}
+	
+	catch_(a)
+	cont_(a);
+	return rtc;
+}
+
+//-----------------
 bool PBMSParameters::isBLOBDatabase(const char *db)
 {
 	CSThread *self;
@@ -338,19 +359,9 @@
 	if ((err = MSEngine::enterConnectionNoThd(&self, &result)) == 0) {
 
 		inner_();
-		try_(a) {
-			lock_(&my_table_list_lock);	
-			
-				
-			found = (locate_db(my_table_list, db, strlen(db)) != NULL);
-				
-			unlock_(&my_table_list_lock);
-		}
-		
-		catch_(a) {
+		if (try_LocateDB(self, db, &found)) {
 			err = MSEngine::exceptionToResult(&self->myException, &result);
-		}
-		cont_(a);
+		}		
 		outer_();
 	
 	}
@@ -364,14 +375,56 @@
 }
 	
 //-----------------
+bool PBMSParameters::try_LocateTable(CSThread *self, const char *db, const char *table, bool *found)
+{
+	volatile bool rtc = true;
+	try_(a) {
+		const char *ptr;
+		int db_len, table_len, match_len;
+		
+		lock_(&my_table_list_lock);	
+				
+		db_len = strlen(db);
+		table_len = strlen(table);
+		
+		ptr = my_table_list;
+		while (ptr) {
+			ptr = locate_db(ptr, db, db_len);
+			if (ptr) {
+				match_len = 0;
+				if (*ptr == '*')
+					match_len = 1;
+				else if (strncmp(ptr, table, table_len) == 0)
+					match_len = table_len;
+					
+				if (match_len) {
+					ptr += match_len;
+					if ((!*ptr) || (*ptr == ',') || isspace(*ptr)) {
+						*found = true;
+						break;
+					}
+				}
+				
+				NEXT_IN_TABLE_LIST(ptr);
+			}
+		}
+			
+		unlock_(&my_table_list_lock);
+		rtc = false;
+	}
+	
+	catch_(a)
+	cont_(a);
+	return rtc;
+}
+
+//-----------------
 bool PBMSParameters::isBLOBTable(const char *db, const char *table)
 {
 	CSThread *self;
 	int		err;
 	PBMSResultRec result;
 	bool found = false;
-	const char *ptr;
-	int db_len, table_len, match_len;
 	
 	if (isBlackListedDB(db))
 		return false;
@@ -388,40 +441,9 @@
 	if ((err = MSEngine::enterConnectionNoThd(&self, &result)) == 0) {
 
 		inner_();
-		try_(a) {
-			lock_(&my_table_list_lock);	
-					
-			db_len = strlen(db);
-			table_len = strlen(table);
-			
-			ptr = my_table_list;
-			while (ptr) {
-				ptr = locate_db(ptr, db, db_len);
-				if (ptr) {
-					match_len = 0;
-					if (*ptr == '*')
-						match_len = 1;
-					else if (strncmp(ptr, table, table_len) == 0)
-						match_len = table_len;
-						
-					if (match_len) {
-						ptr += match_len;
-						if ((!*ptr) || (*ptr == ',') || isspace(*ptr)) {
-							found = true;
-							break;
-						}
-					}
-					
-					NEXT_IN_TABLE_LIST(ptr);
-				}
-			}
-				
-			unlock_(&my_table_list_lock);
-		}
-		catch_(a) {
+		if (try_LocateTable(self, db, table, &found)) {
 			err = MSEngine::exceptionToResult(&self->myException, &result);
-		}
-		cont_(a);
+		}		
 		outer_();
 	
 	}

=== modified file 'plugin/pbms/src/parameters_ms.h'
--- plugin/pbms/src/parameters_ms.h	2010-10-02 21:15:42 +0000
+++ plugin/pbms/src/parameters_ms.h	2010-10-18 21:21:20 +0000
@@ -66,6 +66,11 @@
 	static int32_t getBeforeUptateEventPosition();
 	
 	static int32_t getBeforeInsertEventPosition();	
+	
+	private:
+	static bool try_LocateDB(CSThread *self, const char *db, bool *found);
+	static bool try_LocateTable(CSThread *self, const char *db, const char *table, bool *found);
+
 #endif
 };
 

=== modified file 'plugin/pbms/src/pbms_enabled.cc'
--- plugin/pbms/src/pbms_enabled.cc	2010-07-30 20:31:19 +0000
+++ plugin/pbms/src/pbms_enabled.cc	2010-10-18 21:21:20 +0000
@@ -199,7 +199,7 @@
 			
 			
 		// Get the blob record:
-		field_offset = field->offset(field->table->getInsertRecord());
+		field_offset = field->offset(field->table->record[0]);
 		packlength = field->pack_length() - field->table->s->blob_ptr_size;
 
 		if (new_null_blob) {
@@ -278,7 +278,7 @@
 
 		// Get the blob record:
 		packlength = field->pack_length() - field->table->s->blob_ptr_size;
-		blob_rec = row_buffer + field->offset(field->table->getInsertRecord());
+		blob_rec = row_buffer + field->offset(field->table->record[0]);
 		
 		length = field->get_length(blob_rec);
 		memcpy(&blob_url, blob_rec +packlength, sizeof(char*));
@@ -330,7 +330,7 @@
 		// Get the blob record:
 		packlength = field->pack_length() - field->table->s->blob_ptr_size;
 
-		blob_rec = row_buffer + field->offset(field->table->getInsertRecord());
+		blob_rec = row_buffer + field->offset(field->table->record[0]);
 		length = field->get_length(blob_rec);
 		memcpy(&blob, blob_rec +packlength, sizeof(char*));
 

=== modified file 'plugin/pbms/src/pbmsdaemon_ms.cc'
--- plugin/pbms/src/pbmsdaemon_ms.cc	2010-10-05 18:58:54 +0000
+++ plugin/pbms/src/pbmsdaemon_ms.cc	2010-10-18 21:21:20 +0000
@@ -59,6 +59,6 @@
 	pbd_state = state;
 	
 	if (pbd_state == DaemonStartUp) { // In theory we could allow this to be set with a commandline option.
-		cs_make_absolute_path(PATH_MAX, pbd_home_dir, "pbms", ms_my_get_mysql_home_path().c_str()); 
+		cs_make_absolute_path(PATH_MAX, pbd_home_dir, "pbms", ms_my_get_mysql_home_path()); 
 	}
 }

=== modified file 'plugin/pbms/src/repository_ms.cc'
--- plugin/pbms/src/repository_ms.cc	2010-10-06 18:22:02 +0000
+++ plugin/pbms/src/repository_ms.cc	2010-10-18 21:21:20 +0000
@@ -331,7 +331,7 @@
 		new_(buffer, CSStringBuffer());
 		push_(buffer);
 		buffer->setLength(MS_COMPACTOR_BUFFER_SIZE);
-		CSFile::transfer(otab->myWriteRepoFile, dst_offset + new_head_size, this, offset + head_size, blob_size, buffer->getBuffer(0), MS_COMPACTOR_BUFFER_SIZE);
+		CSFile::transfer(RETAIN(otab->myWriteRepoFile), dst_offset + new_head_size, RETAIN(this), offset + head_size, blob_size, buffer->getBuffer(0), MS_COMPACTOR_BUFFER_SIZE);
 		release_(buffer);
 
 #ifdef HAVE_ALIAS_SUPPORT
@@ -653,7 +653,7 @@
 
 void MSRepoFile::setBlobMetaData(MSOpenTable *otab, uint64_t offset, const char *meta_data, uint16_t meta_data_len, bool reset_alias, const char  *alias)
 {
-	CSMutex				*lock;
+	CSMutex				*mylock;
 	MSRepoPointersRec	ptr;
 	size_t				read_size;
 	uint16_t				new_head_size;
@@ -663,8 +663,8 @@
 	
 	enter_();
 	/* Lock the BLOB: */
-	lock = &myRepo->myRepoLock[offset % CS_REPO_REC_LOCK_COUNT];
-	lock_(lock);
+	mylock = &myRepo->myRepoLock[offset % CS_REPO_REC_LOCK_COUNT];
+	lock_(mylock);
 
 	/* Read the header: */
 	if (read(&blob, offset, sizeof(MSBlobHeadRec), 0) < sizeof(MSBlobHeadRec)) {
@@ -742,7 +742,7 @@
 		alias_offset = 0;
 	}
 #else
-	uint32_t alias_hash = UINT32_MAX;
+	uint32_t alias_hash = -1;
 	if (alias || reset_alias) {
 		CSException::throwException(CS_CONTEXT, MS_ERR_NOT_IMPLEMENTED, "No BLOB alias support.");
 	}
@@ -755,7 +755,7 @@
 		
 	update_blob_header(otab, offset, blob_size, head_size, new_head_size);
 		
-	unlock_(lock);
+	unlock_(mylock);
 	exit_();
 
 }
@@ -763,7 +763,7 @@
 
 void MSRepoFile::releaseBlob(MSOpenTable *otab, uint64_t offset, uint16_t head_size, uint32_t tab_id, uint64_t blob_id, uint64_t blob_ref_id, uint32_t auth_code)
 {
-	CSMutex				*lock;
+	CSMutex				*mylock;
 	MSRepoPointersRec	ptr;
 	uint32_t			table_ref_count = 0;
 	uint32_t			size;
@@ -776,8 +776,8 @@
 
 	enter_();
 	/* Lock the BLOB: */
-	lock = &myRepo->myRepoLock[offset % CS_REPO_REC_LOCK_COUNT];
-	lock_(lock);
+	mylock = &myRepo->myRepoLock[offset % CS_REPO_REC_LOCK_COUNT];
+	lock_(mylock);
 	/* Read the header: */
 	ASSERT(head_size <= MS_OT_BUFFER_SIZE);
 	read_size = read(otab->myOTBuffer, offset, head_size, 0);
@@ -907,13 +907,13 @@
 	}
 
 	exit:
-	unlock_(lock);
+	unlock_(mylock);
 	exit_();
 }
 
 void MSRepoFile::commitBlob(MSOpenTable *otab, uint64_t offset, uint16_t head_size, uint32_t tab_id, uint64_t blob_id, uint64_t blob_ref_id, uint32_t auth_code)
 {
-	CSMutex				*lock;
+	CSMutex				*mylock;
 	MSRepoPointersRec	ptr;
 	uint32_t				size;
 	size_t				ref_size, ref_count, read_size;
@@ -921,8 +921,8 @@
 
 	enter_();
 	/* Lock the BLOB: */
-	lock = &myRepo->myRepoLock[offset % CS_REPO_REC_LOCK_COUNT];
-	lock_(lock);
+	mylock = &myRepo->myRepoLock[offset % CS_REPO_REC_LOCK_COUNT];
+	lock_(mylock);
 	/* Read the header: */
 	ASSERT(head_size <= MS_OT_BUFFER_SIZE);
 	read_size = read(otab->myOTBuffer, offset, head_size, 0);
@@ -996,7 +996,7 @@
 	}
 	
 	exit:
-	unlock_(lock);
+	unlock_(mylock);
 	exit_();
 }
 
@@ -1051,7 +1051,7 @@
 /* This function will free the BLOB reference, if the record is invalid. */
 void MSRepoFile::freeTableReference(MSOpenTable *otab, uint64_t offset, uint16_t head_size, uint32_t tab_id, uint64_t blob_id, uint32_t auth_code)
 {
-	CSMutex				*lock;
+	CSMutex				*mylock;
 	MSRepoPointersRec	ptr;
 	uint32_t				blob_ref_count = 0;
 	uint32_t				table_ref_count = 0;
@@ -1063,8 +1063,8 @@
 
 	enter_();
 	/* Lock the BLOB: */
-	lock = &myRepo->myRepoLock[offset % CS_REPO_REC_LOCK_COUNT];
-	lock_(lock);
+	mylock = &myRepo->myRepoLock[offset % CS_REPO_REC_LOCK_COUNT];
+	lock_(mylock);
 	/* Read the header: */
 	ASSERT(head_size <= MS_OT_BUFFER_SIZE);
 	read_size = read(otab->myOTBuffer, offset, head_size, 0);
@@ -1155,7 +1155,7 @@
 		 */
 		write(otab->myOTBuffer + MS_BLOB_STAT_OFFS, offset + MS_BLOB_STAT_OFFS, head_size - MS_BLOB_STAT_OFFS);
 
-	unlock_(lock);
+	unlock_(mylock);
 
 	if (!table_ref_count || !tab_ref)
 		/* Free the table reference, if there are no more
@@ -1168,13 +1168,13 @@
 	exit_();
 
 	exit:
-	unlock_(lock);
+	unlock_(mylock);
 	exit_();
 }
 
 void MSRepoFile::checkBlob(CSStringBuffer *buffer, uint64_t offset, uint32_t auth_code, uint32_t temp_log_id, uint32_t temp_log_offset)
 {
-	CSMutex				*lock;
+	CSMutex				*mylock;
 	MSBlobHeadRec		blob;
 	MSRepoPointersRec	ptr;
 	uint32_t				blob_ref_count = 0;
@@ -1189,8 +1189,8 @@
 	enter_();
 	
 	/* Lock the BLOB: */
-	lock = &myRepo->myRepoLock[offset % CS_REPO_REC_LOCK_COUNT];
-	lock_(lock);
+	mylock = &myRepo->myRepoLock[offset % CS_REPO_REC_LOCK_COUNT];
+	lock_(mylock);
 
 	/* Read the head of the header: */
 	if (read(&blob, offset, sizeof(MSBlobHeadRec), 0) < sizeof(MSBlobHeadRec)) 
@@ -1276,7 +1276,7 @@
 	}
 	
 	exit:
-	unlock_(lock);
+	unlock_(mylock);
 	exit_();
 }
 
@@ -1399,7 +1399,7 @@
 			offset += (uint64_t) tfer;
 			blob_size -= (uint64_t) tfer;
 		}
-		if (checksum) md5.md5_digest(checksum);
+		if (checksum) md5.md5_get_digest(checksum);
 		release_(stream);
 	} else {
 		// Write 1 byte to the end to reserver the space.

=== modified file 'plugin/pbms/src/systab_backup_ms.cc'
--- plugin/pbms/src/systab_backup_ms.cc	2010-10-09 10:57:54 +0000
+++ plugin/pbms/src/systab_backup_ms.cc	2010-10-18 21:21:20 +0000
@@ -58,17 +58,17 @@
 
 DT_FIELD_INFO pbms_backup_info[]=
 {
-	{"Id",				NULL,	NULL, MYSQL_TYPE_LONG,		NULL,			NOT_NULL_FLAG,	"The backup reference ID"},
+	{"Id",				NOVAL,	NULL, MYSQL_TYPE_LONG,		NULL,			NOT_NULL_FLAG,	"The backup reference ID"},
 	{"Database_Name",	64,		NULL, MYSQL_TYPE_VARCHAR,	&UTF8_CHARSET,	NOT_NULL_FLAG,	"The database name"},
-	{"Database_Id",		NULL,	NULL, MYSQL_TYPE_LONG,		NULL,			0,	"The database ID"},
+	{"Database_Id",		NOVAL,	NULL, MYSQL_TYPE_LONG,		NULL,			0,	"The database ID"},
 	{"Started",			32,		NULL, MYSQL_TYPE_VARCHAR,	&UTF8_CHARSET,	0,	"The start time"},
 	{"Completed",		32,		NULL, MYSQL_TYPE_VARCHAR,	&UTF8_CHARSET,	0,	"The completion time"},
 	{"IsRunning",		3,		NULL, MYSQL_TYPE_VARCHAR,	&UTF8_CHARSET,	0,	"Is the backup still running"},
 	{"IsDump",			3,		NULL, MYSQL_TYPE_VARCHAR,	&UTF8_CHARSET,	0,	"Is the backup the result of a dump"},
 	{"Location",		1024,	NULL, MYSQL_TYPE_VARCHAR,	&UTF8_CHARSET,	0,	"The backup location"},
-	{"Cloud_Ref",		NULL,	NULL, MYSQL_TYPE_LONG,		NULL,			0,	"The S3 cloud reference number refering to the pbms.pbms_cloud table."},
-	{"Cloud_Backup_No",	NULL,	NULL, MYSQL_TYPE_LONG,		NULL,			0,	"The cloud backup number"},
-	{NULL,NULL, NULL, MYSQL_TYPE_STRING,NULL, 0, NULL}
+	{"Cloud_Ref",		NOVAL,	NULL, MYSQL_TYPE_LONG,		NULL,			0,	"The S3 cloud reference number refering to the pbms.pbms_cloud table."},
+	{"Cloud_Backup_No",	NOVAL,	NULL, MYSQL_TYPE_LONG,		NULL,			0,	"The cloud backup number"},
+	{NULL,NOVAL, NULL, MYSQL_TYPE_STRING,NULL, 0, NULL}
 };
 
 DT_KEY_INFO pbms_backup_keys[]=
@@ -268,14 +268,22 @@
 
 	new_(timeVal, CSTime());
 	push_(timeVal);
+#ifdef DRIZZLED
 	memset(buf, 0xFF, table->getNullBytes());
+#else
+	memset(buf, 0xFF, table->s->null_bytes);
+#endif
  	for (Field **field=GET_TABLE_FIELDS(table) ; *field ; field++) {
  		curr_field = *field;
 		save = curr_field->ptr;
 #if MYSQL_VERSION_ID < 50114
 		curr_field->ptr = (byte *) buf + curr_field->offset();
 #else
+#ifdef DRIZZLED
 		curr_field->ptr = (byte *) buf + curr_field->offset(curr_field->getTable()->getInsertRecord());
+#else
+		curr_field->ptr = (byte *) buf + curr_field->offset(curr_field->table->record[0]);
+#endif
 #endif
 		switch (curr_field->field_name[0]) {
 			case 'I':
@@ -458,62 +466,91 @@
 	exit_();
 }
 
+class InsertRowCleanUp : public CSRefObject {
+	bool do_cleanup;
+	CSThread *myself;
+	
+	uint32_t ref_id;
+
+	public:
+	
+	InsertRowCleanUp(CSThread *self): CSRefObject(),
+		do_cleanup(true), myself(self){}
+		
+	~InsertRowCleanUp() 
+	{
+		if (do_cleanup) {
+			myself->logException();
+			if (ref_id)
+				MSBackupInfo::gBackupInfo->remove(ref_id);
+
+		}
+	}
+	
+	void setCleanUp(uint32_t id)
+	{
+		ref_id = id;
+	}
+	
+	void cancelCleanUp()
+	{
+		do_cleanup = false;
+	}
+	
+};
+
 void MSBackupTable::insertRow(char *data) 
 {
 	uint32_t ref_id = 0, db_id, cloud_ref, cloud_backup_no;
 	String name, start, end, isRunning, isDump, location;
 	MSBackupInfo *info = NULL;
 	const char *db_name;
-	bool duplicate = true;
+	InsertRowCleanUp *cleanup;
 
 	enter_();
 
-	try_(a) {
-		getFieldValue(data, 0, &ref_id);
-			
-		// The id must be unique.
-		if (ref_id && MSBackupInfo::gBackupInfo->get(ref_id)) {
-			CSException::throwException(CS_CONTEXT, MS_ERR_DUPLICATE, "Attempt to insert a row with a duplicate key in the "BACKUP_TABLE_NAME" table.");
-		}
-		duplicate = false;
-		
-		// The 'Database_Id', 'Start', 'Completion' and "IsDump" fields are ignored.
-		// I still need to get the fields though to advance the field position pointer.
-		getFieldValue(data, 1, &name);
-		getFieldValue(data, 2, &db_id);
-		getFieldValue(data, 3, &start);
-		getFieldValue(data, 4, &end);
-		getFieldValue(data, 5, &isRunning);
-		getFieldValue(data, 6, &isDump);
-		getFieldValue(data, 7, &location);
-		getFieldValue(data, 8, &cloud_ref);
-		getFieldValue(data, 9, &cloud_backup_no);
-		
-		if (ref_id == 0)
-			ref_id = MSBackupInfo::gMaxInfoRef++;
-		else if (ref_id >= MSBackupInfo::gMaxInfoRef)
-			MSBackupInfo::gMaxInfoRef = ref_id +1;
-		
-		db_name	= name.c_ptr();
-		db_id = MSDatabase::getDatabaseID(db_name, false);
-		
-		new_(info, MSBackupInfo(ref_id, db_name, db_id, 0, 0, false, location.c_ptr(), cloud_ref, cloud_backup_no));
-		MSBackupInfo::gBackupInfo->set(ref_id, info);
-		
-		// There is no need to call this now, startBackup() will call it
-		// after the backup is started.
-		// saveTable(RETAIN(myShare->mySysDatabase)); 
-		info->startBackup(RETAIN(myShare->mySysDatabase));
+	new_(cleanup, InsertRowCleanUp(self));
+	push_(cleanup);
+	
+	getFieldValue(data, 0, &ref_id);
+		
+	// The id must be unique.
+	if (ref_id && MSBackupInfo::gBackupInfo->get(ref_id)) {
+		CSException::throwException(CS_CONTEXT, MS_ERR_DUPLICATE, "Attempt to insert a row with a duplicate key in the "BACKUP_TABLE_NAME" table.");
 	}
 	
-	catch_(a);
-	// It is good to know the details if the backup could not be started.
-	self->logException();
-	if (ref_id && ! duplicate)
-		MSBackupInfo::gBackupInfo->remove(ref_id);
-	throw_();
-	
-	cont_(a);
+	// The 'Database_Id', 'Start', 'Completion' and "IsDump" fields are ignored.
+	// I still need to get the fields though to advance the field position pointer.
+	getFieldValue(data, 1, &name);
+	getFieldValue(data, 2, &db_id);
+	getFieldValue(data, 3, &start);
+	getFieldValue(data, 4, &end);
+	getFieldValue(data, 5, &isRunning);
+	getFieldValue(data, 6, &isDump);
+	getFieldValue(data, 7, &location);
+	getFieldValue(data, 8, &cloud_ref);
+	getFieldValue(data, 9, &cloud_backup_no);
+	
+	if (ref_id == 0)
+		ref_id = MSBackupInfo::gMaxInfoRef++;
+	else if (ref_id >= MSBackupInfo::gMaxInfoRef)
+		MSBackupInfo::gMaxInfoRef = ref_id +1;
+	
+	db_name	= name.c_ptr();
+	db_id = MSDatabase::getDatabaseID(db_name, false);
+	
+	cleanup->setCleanUp(ref_id);
+	new_(info, MSBackupInfo(ref_id, db_name, db_id, 0, 0, false, location.c_ptr(), cloud_ref, cloud_backup_no));
+	MSBackupInfo::gBackupInfo->set(ref_id, info);
+	
+	// There is no need to call this now, startBackup() will call it
+	// after the backup is started.
+	// saveTable(RETAIN(myShare->mySysDatabase)); 
+	info->startBackup(RETAIN(myShare->mySysDatabase));
+
+	cleanup->cancelCleanUp();
+	release_(cleanup);
+	
 	exit_();
 }
 

=== modified file 'plugin/pbms/src/systab_cloud_ms.cc'
--- plugin/pbms/src/systab_cloud_ms.cc	2010-10-09 10:57:54 +0000
+++ plugin/pbms/src/systab_cloud_ms.cc	2010-10-18 21:21:20 +0000
@@ -57,12 +57,12 @@
 
 DT_FIELD_INFO pbms_cloud_info[]=
 {
-	{"Id",			NULL,	NULL, MYSQL_TYPE_LONG,		NULL,			NOT_NULL_FLAG,	"The Cloud storage reference ID"},
+	{"Id",			NOVAL,	NULL, MYSQL_TYPE_LONG,		NULL,			NOT_NULL_FLAG,	"The Cloud storage reference ID"},
 	{"Server",		1024,	NULL, MYSQL_TYPE_VARCHAR,	&UTF8_CHARSET,	NOT_NULL_FLAG,	"S3 server name"},
 	{"Bucket",		124,	NULL, MYSQL_TYPE_VARCHAR,	&UTF8_CHARSET,	NOT_NULL_FLAG,	"S3 bucket name"},
 	{"PublicKey",	124,	NULL, MYSQL_TYPE_VARCHAR,	&UTF8_CHARSET,	NOT_NULL_FLAG,	"S3 public key"},
 	{"PrivateKey",	124,	NULL, MYSQL_TYPE_VARCHAR,	&UTF8_CHARSET,	NOT_NULL_FLAG,	"S3 private key"},
-	{NULL,NULL, NULL, MYSQL_TYPE_STRING,NULL, 0, NULL}
+	{NULL,NOVAL, NULL, MYSQL_TYPE_STRING,NULL, 0, NULL}
 };
 
 DT_KEY_INFO pbms_cloud_keys[]=
@@ -250,14 +250,22 @@
 	save_write_set = table->write_set;
 	table->write_set = NULL;
 
+#ifdef DRIZZLED
 	memset(buf, 0xFF, table->getNullBytes());
+#else
+	memset(buf, 0xFF, table->s->null_bytes);
+#endif
  	for (Field **field=GET_TABLE_FIELDS(table) ; *field ; field++) {
  		curr_field = *field;
 		save = curr_field->ptr;
 #if MYSQL_VERSION_ID < 50114
 		curr_field->ptr = (byte *) buf + curr_field->offset();
 #else
+#ifdef DRIZZLED
 		curr_field->ptr = (byte *) buf + curr_field->offset(curr_field->getTable()->getInsertRecord());
+#else
+		curr_field->ptr = (byte *) buf + curr_field->offset(curr_field->table->record[0]);
+#endif
 #endif
 		switch (curr_field->field_name[0]) {
 			case 'I':

=== modified file 'plugin/pbms/src/systab_dump_ms.cc'
--- plugin/pbms/src/systab_dump_ms.cc	2010-10-09 10:57:54 +0000
+++ plugin/pbms/src/systab_dump_ms.cc	2010-10-18 21:21:20 +0000
@@ -59,8 +59,8 @@
 
 DT_FIELD_INFO pbms_dump_info[]=
 {
-	{"Data",			NULL, NULL, MYSQL_TYPE_LONG_BLOB,	&my_charset_bin,	NOT_NULL_FLAG,	"A BLOB repository record"},
-	{NULL,NULL, NULL, MYSQL_TYPE_STRING,NULL, 0, NULL}
+	{"Data",			NOVAL, NULL, MYSQL_TYPE_LONG_BLOB,	&my_charset_bin,	NOT_NULL_FLAG,	"A BLOB repository record"},
+	{NULL,NOVAL, NULL, MYSQL_TYPE_STRING,NULL, 0, NULL}
 };
 
 DT_KEY_INFO pbms_dump_keys[]=
@@ -157,7 +157,11 @@
 	 */
 	save_write_set = table->write_set;
 	table->write_set = NULL;
+#ifdef DRIZZLED
 	memset(buf, 0xFF, table->getNullBytes());
+#else
+	memset(buf, 0xFF, table->s->null_bytes);
+#endif
 	
  	for (Field **field=GET_TABLE_FIELDS(table) ; *field ; field++) {
  		curr_field = *field;
@@ -166,7 +170,11 @@
 #if MYSQL_VERSION_ID < 50114
 		curr_field->ptr = (byte *) buf + curr_field->offset();
 #else
+#ifdef DRIZZLED
 		curr_field->ptr = (byte *) buf + curr_field->offset(curr_field->getTable()->getInsertRecord());
+#else
+		curr_field->ptr = (byte *) buf + curr_field->offset(curr_field->table->record[0]);
+#endif
 #endif
 		switch (curr_field->field_name[0]) {
 			case 'D':
@@ -254,7 +262,7 @@
 	}
 
 
-	return returnDumpRow(iBlobBuffer->getBuffer(0), record_size, buf);
+	return_(returnDumpRow(iBlobBuffer->getBuffer(0), record_size, buf));
 }
 
 //-----------------------
@@ -329,7 +337,7 @@
 	INC_INFO_SPACE(4);
 	memcpy(ptr, sysTablesDump->getBuffer(0), sysTablesDump->length());
 	INC_INFO_SPACE(sysTablesDump->length());
-	sysTablesDump->release();
+	release_(sysTablesDump);
 	sysTablesDump = NULL;
 			
 	tab_info = (TabInfoPtr)ptr;
@@ -426,8 +434,13 @@
 	field = (Field_blob *)GET_FIELD(table, 0);
 	
     /* Get the blob record: */
+#ifdef DRIZZLED
     blob_rec= buf + field->offset(table->getInsertRecord());
     packlength= field->pack_length() - table->getShare()->blob_ptr_size;
+#else
+    blob_rec= buf + field->offset(table->record[0]);
+    packlength= field->pack_length() - table->s->blob_ptr_size;
+#endif
 
     memcpy(&blob_ptr, blob_rec +packlength, sizeof(char*));
     length= field->get_length();

=== modified file 'plugin/pbms/src/systab_enabled_ms.cc'
--- plugin/pbms/src/systab_enabled_ms.cc	2010-10-02 21:15:42 +0000
+++ plugin/pbms/src/systab_enabled_ms.cc	2010-10-18 21:21:20 +0000
@@ -56,8 +56,8 @@
 	{"Name",			32,		NULL, MYSQL_TYPE_VARCHAR,	&UTF8_CHARSET,	NOT_NULL_FLAG,	"PBMS enabled engine name"},
 	{"IsServer",		3,		NULL, MYSQL_TYPE_VARCHAR,	&UTF8_CHARSET,					NOT_NULL_FLAG,	"Enabled at server level."},
 	{"Transactional",	5,		NULL, MYSQL_TYPE_VARCHAR,	&UTF8_CHARSET,					NOT_NULL_FLAG,	"Does the engine support transactions."},
-	{"API-Version",		NULL,	NULL, MYSQL_TYPE_LONG,		NULL,							NOT_NULL_FLAG,	"The PBMS enabled api version used."},
-	{NULL,NULL, NULL, MYSQL_TYPE_STRING,NULL, 0, NULL}
+	{"API-Version",		NOVAL,	NULL, MYSQL_TYPE_LONG,		NULL,							NOT_NULL_FLAG,	"The PBMS enabled api version used."},
+	{NULL,NOVAL, NULL, MYSQL_TYPE_STRING,NULL, 0, NULL}
 };
 
 DT_KEY_INFO pbms_enabled_keys[]=
@@ -107,14 +107,23 @@
 	save_write_set = table->write_set;
 	table->write_set = NULL;
 
+#ifdef DRIZZLED
+	memset(buf, 0xFF, table->getNullBytes());
+#else
 	memset(buf, 0xFF, table->s->null_bytes);
+#endif
+
  	for (Field **field=GET_TABLE_FIELDS(table) ; *field ; field++) {
  		curr_field = *field;
 		save = curr_field->ptr;
 #if MYSQL_VERSION_ID < 50114
 		curr_field->ptr = (byte *) buf + curr_field->offset();
 #else
-		curr_field->ptr = (byte *) buf + curr_field->offset(curr_field->table->getInsertRecord());
+#ifdef DRIZZLED
+		curr_field->ptr = (byte *) buf + curr_field->offset(curr_field->getTable()->getInsertRecord());
+#else
+		curr_field->ptr = (byte *) buf + curr_field->offset(curr_field->table->record[0]);
+#endif
 #endif
 
 		switch (curr_field->field_name[0]) {

=== modified file 'plugin/pbms/src/systab_httpheader_ms.cc'
--- plugin/pbms/src/systab_httpheader_ms.cc	2010-10-09 10:57:54 +0000
+++ plugin/pbms/src/systab_httpheader_ms.cc	2010-10-18 21:21:20 +0000
@@ -62,7 +62,7 @@
 DT_FIELD_INFO pbms_metadata_headers_info[]=
 {
 	{"Name", 32,	NULL, MYSQL_TYPE_VARCHAR,	&UTF8_CHARSET,	NOT_NULL_FLAG,	"HTTP field name"},
-	{NULL,	NULL,				NULL, MYSQL_TYPE_STRING,	NULL, 0, NULL}
+	{NULL,	NOVAL,				NULL, MYSQL_TYPE_STRING,	NULL, 0, NULL}
 };
 
 DT_KEY_INFO pbms_metadata_headers_keys[]=
@@ -98,7 +98,7 @@
 				gDefaultMetaDataHeaders->setStringField(defaults, ptr - defaults);
 				gDefaultMetaDataHeaders->endRecord();
 			}
-			if (!ptr)
+			if (!*ptr)
 				break;
 				
 			defaults = ptr +1;
@@ -135,8 +135,7 @@
 		
 		
 	} else if (gDefaultMetaDataHeaders) { // Load the defaults if they exist.
-		headerData = RETAIN(gDefaultMetaDataHeaders);
-		push_(headerData);
+		headerData = gDefaultMetaDataHeaders;
 	}
 
 	if (headerData) {
@@ -148,7 +147,11 @@
 			}
 		}
 		
-		release_(headerData); 
+		if (headerData == gDefaultMetaDataHeaders)
+			gDefaultMetaDataHeaders->resetRecord();
+		else {
+			release_(headerData); 
+		}
 	}
 	
 	release_(path);	
@@ -344,14 +347,22 @@
 	save_write_set = table->write_set;
 	table->write_set = NULL;
 
+#ifdef DRIZZLED
 	memset(buf, 0xFF, table->getNullBytes());
+#else
+	memset(buf, 0xFF, table->s->null_bytes);
+#endif
  	for (Field **field=GET_TABLE_FIELDS(table); *field ; field++) {
  		curr_field = *field;
 		save = curr_field->ptr;
 #if MYSQL_VERSION_ID < 50114
 		curr_field->ptr = (byte *) buf + curr_field->offset();
 #else
+#ifdef DRIZZLED
 		curr_field->ptr = (byte *) buf + curr_field->offset(curr_field->getTable()->getInsertRecord());
+#else
+		curr_field->ptr = (byte *) buf + curr_field->offset(curr_field->table->record[0]);
+#endif
 #endif
 		switch (curr_field->field_name[0]) {
 			case 'N':

=== modified file 'plugin/pbms/src/systab_util_ms.cc'
--- plugin/pbms/src/systab_util_ms.cc	2010-10-02 21:15:42 +0000
+++ plugin/pbms/src/systab_util_ms.cc	2010-10-18 21:21:20 +0000
@@ -124,6 +124,12 @@
 }
 	
 //---------------------
+void SysTabRec::resetRecord()
+{
+	ptr= 0;
+}
+
+//---------------------
 uint8_t SysTabRec::getInt1Field()
 {
 	uint8_t val = 0;

=== modified file 'plugin/pbms/src/systab_util_ms.h'
--- plugin/pbms/src/systab_util_ms.h	2010-10-02 21:15:42 +0000
+++ plugin/pbms/src/systab_util_ms.h	2010-10-18 21:21:20 +0000
@@ -55,6 +55,7 @@
 	// Methods used to read records
 	bool firstRecord();
 	bool nextRecord();
+	void resetRecord();
 		
 	uint8_t getInt1Field();	
 	uint32_t getInt4Field();	

=== modified file 'plugin/pbms/src/systab_variable_ms.cc'
--- plugin/pbms/src/systab_variable_ms.cc	2010-10-09 10:57:54 +0000
+++ plugin/pbms/src/systab_variable_ms.cc	2010-10-18 21:21:20 +0000
@@ -60,11 +60,11 @@
 
 DT_FIELD_INFO pbms_variable_info[]=
 {
-	{"Id",			NULL,	NULL, MYSQL_TYPE_LONG,		NULL,							NOT_NULL_FLAG,	"The variable ID"},
+	{"Id",			NOVAL,	NULL, MYSQL_TYPE_LONG,		NULL,							NOT_NULL_FLAG,	"The variable ID"},
 	{"Name",		32,		NULL, MYSQL_TYPE_VARCHAR,	&UTF8_CHARSET,	NOT_NULL_FLAG,	"PBMS variable name"},
 	{"Value",	1024,	NULL, MYSQL_TYPE_VARCHAR,	&UTF8_CHARSET,	0,	"PBMS variable value."},
 	{"Description",	124,	NULL, MYSQL_TYPE_VARCHAR,	&UTF8_CHARSET,	NOT_NULL_FLAG,	"PBMS variable description."},
-	{NULL,NULL, NULL, MYSQL_TYPE_STRING,NULL, 0, NULL}
+	{NULL,NOVAL, NULL, MYSQL_TYPE_STRING,NULL, 0, NULL}
 };
 
 DT_KEY_INFO pbms_variable_keys[]=
@@ -327,6 +327,39 @@
 	return_(path);
 }
 
+class LoadTableCleanUp : public CSRefObject {
+	bool do_cleanup;
+	CSThread *myself;
+	
+	uint32_t ref_id;
+
+	public:
+	
+	LoadTableCleanUp(): CSRefObject(),
+		do_cleanup(false), myself(NULL){}
+		
+	~LoadTableCleanUp() 
+	{
+		if (do_cleanup) {
+			CSL.log(myself, CSLog::Protocol, "\nRestore failed!\n");
+			CSL.flush();
+			myself->logException();
+		}
+	}
+	
+	void setCleanUp(CSThread *self)
+	{
+		myself = self;
+		do_cleanup = true;
+	}
+	
+	void cancelCleanUp()
+	{
+		do_cleanup = false;
+	}
+	
+};
+
 void MSVariableTable::loadTable(MSDatabase *db)
 {
 	CSPath	*path;
@@ -392,16 +425,17 @@
 		CSL.log(self, CSLog::Protocol, db->myDatabaseName->getCString());
 		CSL.log(self, CSLog::Protocol, " ...");
 		CSL.flush();
-		try_(a) {
-			db->myBlobCloud->cl_restoreDB();
-		}
-		catch_(a) {
-			CSL.log(self, CSLog::Protocol, "\nRestore failed!\n");
-			CSL.flush();
-			self->logException();
-			throw_();
-		}
-		cont_(a);
+		LoadTableCleanUp *cleanup;
+		
+		new_(cleanup, LoadTableCleanUp());
+		push_(cleanup);
+		cleanup->setCleanUp(self);
+		
+		db->myBlobCloud->cl_restoreDB();
+		
+		cleanup->cancelCleanUp();
+		release_(cleanup);
+
 		CSL.log(self, CSLog::Protocol, "\nRestore done.\n");
 		CSL.flush();
 		set_BackupNo(RETAIN(db), "0");
@@ -509,14 +543,22 @@
 	save_write_set = table->write_set;
 	table->write_set = NULL;
 
+#ifdef DRIZZLED
 	memset(buf, 0xFF, table->getNullBytes());
+#else
+	memset(buf, 0xFF, table->s->null_bytes);
+#endif
  	for (Field **field=GET_TABLE_FIELDS(table) ; *field ; field++) {
  		curr_field = *field;
 		save = curr_field->ptr;
 #if MYSQL_VERSION_ID < 50114
 		curr_field->ptr = (byte *) buf + curr_field->offset();
 #else
+#ifdef DRIZZLED
 		curr_field->ptr = (byte *) buf + curr_field->offset(curr_field->getTable()->getInsertRecord());
+#else
+		curr_field->ptr = (byte *) buf + curr_field->offset(curr_field->table->record[0]);
+#endif
 #endif
 		switch (curr_field->field_name[0]) {
 			case 'I':

=== modified file 'plugin/pbms/src/system_table_ms.cc'
--- plugin/pbms/src/system_table_ms.cc	2010-10-09 10:57:54 +0000
+++ plugin/pbms/src/system_table_ms.cc	2010-10-18 21:21:20 +0000
@@ -93,17 +93,17 @@
 #ifdef HAVE_ALIAS_SUPPORT
 	{"Blob_alias",			BLOB_ALIAS_LENGTH, NULL, MYSQL_TYPE_VARCHAR,		&my_charset_utf8_bin,	0,								"The BLOB alias"},
 #endif
-	{"Repository_id",		NULL, NULL, MYSQL_TYPE_LONG,		NULL,					NOT_NULL_FLAG,					"The repository file number"},
-	{"Repo_blob_offset",	NULL, NULL, MYSQL_TYPE_LONGLONG,	NULL,					NOT_NULL_FLAG,					"The offset of the BLOB in the repository file"},
-	{"Blob_size",			NULL, NULL, MYSQL_TYPE_LONGLONG,	NULL,					NOT_NULL_FLAG,					"The size of the BLOB in bytes"},
+	{"Repository_id",		NOVAL, NULL, MYSQL_TYPE_LONG,		NULL,					NOT_NULL_FLAG,					"The repository file number"},
+	{"Repo_blob_offset",	NOVAL, NULL, MYSQL_TYPE_LONGLONG,	NULL,					NOT_NULL_FLAG,					"The offset of the BLOB in the repository file"},
+	{"Blob_size",			NOVAL, NULL, MYSQL_TYPE_LONGLONG,	NULL,					NOT_NULL_FLAG,					"The size of the BLOB in bytes"},
 	{"MD5_Checksum",		32,   NULL,	MYSQL_TYPE_VARCHAR,		system_charset_info,	0,								"The MD5 Digest of the BLOB data."},
-	{"Head_size",			NULL, NULL, MYSQL_TYPE_SHORT,		NULL,					NOT_NULL_FLAG | UNSIGNED_FLAG,	"The size of the BLOB header - proceeds the BLOB data"},
-	{"Access_code",			NULL, NULL, MYSQL_TYPE_LONG,		NULL,					NOT_NULL_FLAG,					"The 4-byte authorisation code required to access the BLOB - part of the BLOB URL"},
-	{"Creation_time",		NULL, NULL, MYSQL_TYPE_TIMESTAMP,	NULL,					NOT_NULL_FLAG,					"The time the BLOB was created"},
-	{"Last_ref_time",		NULL, NULL, MYSQL_TYPE_TIMESTAMP,	NULL,					0,								"The last time the BLOB was referenced"},
-	{"Last_access_time",	NULL, NULL, MYSQL_TYPE_TIMESTAMP,	NULL,					0,								"The last time the BLOB was accessed (read)"},
-	{"Access_count",		NULL, NULL, MYSQL_TYPE_LONG,		NULL,					NOT_NULL_FLAG,					"The count of the number of times the BLOB has been read"},
-	{NULL,NULL, NULL, MYSQL_TYPE_STRING,NULL, 0, NULL}
+	{"Head_size",			NOVAL, NULL, MYSQL_TYPE_SHORT,		NULL,					NOT_NULL_FLAG | UNSIGNED_FLAG,	"The size of the BLOB header - proceeds the BLOB data"},
+	{"Access_code",			NOVAL, NULL, MYSQL_TYPE_LONG,		NULL,					NOT_NULL_FLAG,					"The 4-byte authorisation code required to access the BLOB - part of the BLOB URL"},
+	{"Creation_time",		NOVAL, NULL, MYSQL_TYPE_TIMESTAMP,	NULL,					NOT_NULL_FLAG,					"The time the BLOB was created"},
+	{"Last_ref_time",		NOVAL, NULL, MYSQL_TYPE_TIMESTAMP,	NULL,					0,								"The last time the BLOB was referenced"},
+	{"Last_access_time",	NOVAL, NULL, MYSQL_TYPE_TIMESTAMP,	NULL,					0,								"The last time the BLOB was accessed (read)"},
+	{"Access_count",		NOVAL, NULL, MYSQL_TYPE_LONG,		NULL,					NOT_NULL_FLAG,					"The count of the number of times the BLOB has been read"},
+	{NULL,NOVAL, NULL, MYSQL_TYPE_STRING,NULL, 0, NULL}
 };
 
 #ifdef PBMS_HAS_KEYS
@@ -116,11 +116,11 @@
 
 static DT_FIELD_INFO pbms_metadata_info[]=
 {
-	{"Repository_id",		NULL,					NULL, MYSQL_TYPE_LONG,		NULL,							NOT_NULL_FLAG,	"The repository file number"},
-	{"Repo_blob_offset",	NULL,					NULL, MYSQL_TYPE_LONGLONG,	NULL,							NOT_NULL_FLAG,	"The offset of the BLOB in the repository file"},
+	{"Repository_id",		NOVAL,					NULL, MYSQL_TYPE_LONG,		NULL,							NOT_NULL_FLAG,	"The repository file number"},
+	{"Repo_blob_offset",	NOVAL,					NULL, MYSQL_TYPE_LONGLONG,	NULL,							NOT_NULL_FLAG,	"The offset of the BLOB in the repository file"},
 	{"Name",				MS_META_NAME_SIZE,	NULL, MYSQL_TYPE_VARCHAR,	&UTF8_CHARSET,	NOT_NULL_FLAG,	"Metadata name"},
 	{"Value",				MS_META_VALUE_SIZE,	NULL, MYSQL_TYPE_VARCHAR,	&UTF8_CHARSET,			NOT_NULL_FLAG,	"Metadata value"},
-	{NULL,					NULL,					NULL, MYSQL_TYPE_STRING,	NULL, 0, NULL}
+	{NULL,					NOVAL,					NULL, MYSQL_TYPE_STRING,	NULL, 0, NULL}
 };
 
 #ifdef PBMS_HAS_KEYS
@@ -135,10 +135,10 @@
 #ifdef HAVE_ALIAS_SUPPORT
 static DT_FIELD_INFO pbms_alias_info[]=
 {
-	{"Repository_id",		NULL, NULL, MYSQL_TYPE_LONG,		NULL,				NOT_NULL_FLAG,	"The repository file number"},
-	{"Repo_blob_offset",	NULL, NULL, MYSQL_TYPE_LONGLONG,	NULL,				NOT_NULL_FLAG,	"The offset of the BLOB in the repository file"},
+	{"Repository_id",		NOVAL, NULL, MYSQL_TYPE_LONG,		NULL,				NOT_NULL_FLAG,	"The repository file number"},
+	{"Repo_blob_offset",	NOVAL, NULL, MYSQL_TYPE_LONGLONG,	NULL,				NOT_NULL_FLAG,	"The offset of the BLOB in the repository file"},
 	{"Blob_alias",			BLOB_ALIAS_LENGTH, NULL, MYSQL_TYPE_VARCHAR,		&my_charset_utf8_bin,	NOT_NULL_FLAG,			"The BLOB alias"},
-	{NULL,NULL, NULL, MYSQL_TYPE_STRING,NULL, 0, NULL}
+	{NULL,NOVAL, NULL, MYSQL_TYPE_STRING,NULL, 0, NULL}
 };
 
 static DT_KEY_INFO pbms_alias_keys[]=
@@ -150,10 +150,10 @@
 
 static DT_FIELD_INFO pbms_blobs_info[]=
 {
-	{"Repository_id",		NULL, NULL, MYSQL_TYPE_LONG,		NULL,				NOT_NULL_FLAG,	"The repository file number"},
-	{"Repo_blob_offset",	NULL, NULL, MYSQL_TYPE_LONGLONG,	NULL,				NOT_NULL_FLAG,	"The offset of the BLOB in the repository file"},
-	{"Blob_data",			NULL, NULL, MYSQL_TYPE_LONG_BLOB,	&my_charset_bin,	NOT_NULL_FLAG,	"The data of this BLOB"},
-	{NULL,NULL, NULL, MYSQL_TYPE_STRING,NULL, 0, NULL}
+	{"Repository_id",		NOVAL, NULL, MYSQL_TYPE_LONG,		NULL,				NOT_NULL_FLAG,	"The repository file number"},
+	{"Repo_blob_offset",	NOVAL, NULL, MYSQL_TYPE_LONGLONG,	NULL,				NOT_NULL_FLAG,	"The offset of the BLOB in the repository file"},
+	{"Blob_data",			NOVAL, NULL, MYSQL_TYPE_LONG_BLOB,	&my_charset_bin,	NOT_NULL_FLAG,	"The data of this BLOB"},
+	{NULL,NOVAL, NULL, MYSQL_TYPE_STRING,NULL, 0, NULL}
 };
 
 #ifdef PBMS_HAS_KEYS
@@ -167,17 +167,17 @@
 static DT_FIELD_INFO pbms_reference_info[]=
 {
 	{"Table_name",		MS_TABLE_NAME_SIZE,		NULL, MYSQL_TYPE_STRING,	system_charset_info,	0,	"The name of the referencing table"},
-	{"Column_ordinal",	NULL,					NULL, MYSQL_TYPE_LONG,		NULL,					0,	"The column ordinal of the referencing field"},
-	{"Blob_id",			NULL,					NULL, MYSQL_TYPE_LONGLONG,	NULL,					NOT_NULL_FLAG,	"The BLOB reference number - part of the BLOB URL"},
+	{"Column_ordinal",	NOVAL,					NULL, MYSQL_TYPE_LONG,		NULL,					0,	"The column ordinal of the referencing field"},
+	{"Blob_id",			NOVAL,					NULL, MYSQL_TYPE_LONGLONG,	NULL,					NOT_NULL_FLAG,	"The BLOB reference number - part of the BLOB URL"},
 	{"Blob_url",		PBMS_BLOB_URL_SIZE,		NULL, MYSQL_TYPE_VARCHAR,	system_charset_info,	0,	"The BLOB URL for HTTP GET access"},
-	{"Repository_id",	NULL,					NULL, MYSQL_TYPE_LONG,		NULL,					NOT_NULL_FLAG,	"The repository file number of the BLOB"},
-	{"Repo_blob_offset",NULL,					NULL, MYSQL_TYPE_LONGLONG,	NULL,					NOT_NULL_FLAG,	"The offset in the repository file"},
-	{"Blob_size",		NULL,					NULL, MYSQL_TYPE_LONGLONG,	NULL,					NOT_NULL_FLAG,	"The size of the BLOB in bytes"},
-	{"Deletion_time",	NULL,					NULL, MYSQL_TYPE_TIMESTAMP,	NULL,					0,				"The time the BLOB was deleted"},
-	{"Remove_in",		NULL,					NULL, MYSQL_TYPE_LONG,		NULL,					0,				"The number of seconds before the reference/BLOB is removed perminently"},
-	{"Temp_log_id",		NULL,					NULL, MYSQL_TYPE_LONG,		NULL,					0,				"Temporary log number of the referencing deletion entry"},
-	{"Temp_log_offset",	NULL,					NULL, MYSQL_TYPE_LONGLONG,	NULL,					0,				"Temporary log offset of the referencing deletion entry"},
-	{NULL,NULL, NULL, MYSQL_TYPE_STRING,NULL, 0, NULL}
+	{"Repository_id",	NOVAL,					NULL, MYSQL_TYPE_LONG,		NULL,					NOT_NULL_FLAG,	"The repository file number of the BLOB"},
+	{"Repo_blob_offset",NOVAL,					NULL, MYSQL_TYPE_LONGLONG,	NULL,					NOT_NULL_FLAG,	"The offset in the repository file"},
+	{"Blob_size",		NOVAL,					NULL, MYSQL_TYPE_LONGLONG,	NULL,					NOT_NULL_FLAG,	"The size of the BLOB in bytes"},
+	{"Deletion_time",	NOVAL,					NULL, MYSQL_TYPE_TIMESTAMP,	NULL,					0,				"The time the BLOB was deleted"},
+	{"Remove_in",		NOVAL,					NULL, MYSQL_TYPE_LONG,		NULL,					0,				"The number of seconds before the reference/BLOB is removed perminently"},
+	{"Temp_log_id",		NOVAL,					NULL, MYSQL_TYPE_LONG,		NULL,					0,				"Temporary log number of the referencing deletion entry"},
+	{"Temp_log_offset",	NOVAL,					NULL, MYSQL_TYPE_LONGLONG,	NULL,					0,				"Temporary log offset of the referencing deletion entry"},
+	{NULL,NOVAL, NULL, MYSQL_TYPE_STRING,NULL, 0, NULL}
 };
 
 #ifdef PBMS_HAS_KEYS
@@ -423,7 +423,7 @@
 	if (!db)
 		return false;
 		
-	cs_strcpy(PATH_MAX, path, ms_my_get_mysql_home_path().c_str());
+	cs_strcpy(PATH_MAX, path, ms_my_get_mysql_home_path());
 	cs_add_name_to_path(PATH_MAX, path, db);
 	
 	if (stat(path, &stat_info) == 0)
@@ -489,40 +489,44 @@
 }
 
 //----------------
+bool PBMSSystemTables::try_loadSystemTables(CSThread *self, int i, MSDatabase *db)
+{
+	volatile bool rtc = true;
+	try_(a) {
+		switch (i) {
+			case 0:
+				MSHTTPHeaderTable::loadTable(RETAIN(db));
+				break;
+			case 1:
+				MSCloudTable::loadTable(RETAIN(db));
+				break;
+			case 2:
+				MSBackupTable::loadTable(RETAIN(db));
+				break;
+			case 3:
+				// Variable must be loaded after cloud and backup info
+				// incase BLOB recovery is required.
+				MSVariableTable::loadTable(RETAIN(db)); 
+				break;
+				
+			default:
+				ASSERT(false);
+		}
+		rtc = false;
+	}
+	catch_(a);
+	cont_(a);
+	return rtc;
+}
+//----------------
 void PBMSSystemTables::loadSystemTables(MSDatabase *db)
 {
-	int i = 0;
-	
-	CLOBBER_PROTECT(i);
 	enter_();
 	push_(db);
 	
-	for ( i = 0; i < 4; i++) {
-		try_(a) {
-			switch (i) {
-				case 0:
-					MSHTTPHeaderTable::loadTable(RETAIN(db));
-					break;
-				case 1:
-					MSCloudTable::loadTable(RETAIN(db));
-					break;
-				case 2:
-					MSBackupTable::loadTable(RETAIN(db));
-					break;
-				case 3:
-					// Variable must be loaded after cloud and backup info
-					// incase BLOB recovery is required.
-					MSVariableTable::loadTable(RETAIN(db)); 
-					break;
-					
-				default:
-					ASSERT(false);
-			}
-		}
-		catch_(a) {
-			self->logException();
-		}
-		cont_(a);
+	for ( int i = 0; i < 4; i++) {
+		if (try_loadSystemTables(self, i, db))
+			self->logException();			
 	}
 	
 	release_(db);
@@ -675,8 +679,13 @@
 
 void MSOpenSystemTable::setNotNullInRecord(Field *field, char *record)
 {
+#ifdef DRIZZLED
 	if (field->null_ptr)
 		record[(uint) (field->null_ptr - (uchar *) field->getTable()->getInsertRecord())] &= (uchar) ~field->null_bit;
+#else
+	if (field->null_ptr)
+		record[(uint) (field->null_ptr - (uchar *) field->table->record[0])] &= (uchar) ~field->null_bit;
+#endif
 }
 
 /*
@@ -722,6 +731,22 @@
 //-----------------------
 void MSRepositoryTable::use()
 {
+	MSDatabase *db;
+	enter_();
+
+	if (!iBlobBuffer)
+		new_(iBlobBuffer, CSStringBuffer(20));
+		
+	db = myShare->mySysDatabase;
+	if ((iCompactor = db->getCompactorThread())) {
+		if (iCompactor->isMe(self))
+			iCompactor = NULL;
+		else {
+			iCompactor->retain();
+			iCompactor->suspend();
+		}
+	}
+	exit_();
 }
 
 //-----------------------
@@ -744,8 +769,6 @@
 //-----------------------
 void MSRepositoryTable::seqScanInit()
 {
-	MSDatabase *db;
-
 	enter_();
 	
 	// Flush all committed transactions to the repository file.
@@ -753,17 +776,6 @@
 
 	iRepoIndex = 0;
 	iRepoOffset = 0;
-	if (!iBlobBuffer)
-		new_(iBlobBuffer, CSStringBuffer(20));
-	db = myShare->mySysDatabase;
-	if ((iCompactor = db->getCompactorThread())) {
-		if (iCompactor->isMe(self))
-			iCompactor = NULL;
-		else {
-			iCompactor->retain();
-			iCompactor->suspend();
-		}
-	}
 		
 	exit_();
 }
@@ -959,15 +971,22 @@
 	save_write_set = table->write_set;
 	table->write_set = NULL;
 
+#ifdef DRIZZLED
 	memset(buf, 0xFF, table->getNullBytes());
+#else
+	memset(buf, 0xFF, table->s->null_bytes);
+#endif
  	for (Field **field=GET_TABLE_FIELDS(table) ; *field ; field++) {
  		curr_field = *field;
-
 		save = curr_field->ptr;
 #if MYSQL_VERSION_ID < 50114
 		curr_field->ptr = (byte *) buf + curr_field->offset();
 #else
+#ifdef DRIZZLED
 		curr_field->ptr = (byte *) buf + curr_field->offset(curr_field->getTable()->getInsertRecord());
+#else
+		curr_field->ptr = (byte *) buf + curr_field->offset(curr_field->table->record[0]);
+#endif
 #endif
 		switch (curr_field->field_name[0]) {
 			case 'A':
@@ -1121,15 +1140,22 @@
 	save_write_set = table->write_set;
 	table->write_set = NULL;
 
+#ifdef DRIZZLED
 	memset(buf, 0xFF, table->getNullBytes());
+#else
+	memset(buf, 0xFF, table->s->null_bytes);
+#endif
  	for (Field **field=GET_TABLE_FIELDS(table) ; *field ; field++) {
  		curr_field = *field;
-
 		save = curr_field->ptr;
 #if MYSQL_VERSION_ID < 50114
 		curr_field->ptr = (byte *) buf + curr_field->offset();
 #else
+#ifdef DRIZZLED
 		curr_field->ptr = (byte *) buf + curr_field->offset(curr_field->getTable()->getInsertRecord());
+#else
+		curr_field->ptr = (byte *) buf + curr_field->offset(curr_field->table->record[0]);
+#endif
 #endif
 		switch (curr_field->field_name[0]) {
 			case 'R':
@@ -1211,15 +1237,22 @@
 	save_write_set = table->write_set;
 	table->write_set = NULL;
 
+#ifdef DRIZZLED
+	memset(buf, 0xFF, table->getNullBytes());
+#else
 	memset(buf, 0xFF, table->s->null_bytes);
+#endif
  	for (Field **field=GET_TABLE_FIELDS(table) ; *field ; field++) {
  		curr_field = *field;
-
 		save = curr_field->ptr;
 #if MYSQL_VERSION_ID < 50114
 		curr_field->ptr = (byte *) buf + curr_field->offset();
 #else
-		curr_field->ptr = (byte *) buf + curr_field->offset(curr_field->table->getInsertRecord());
+#ifdef DRIZZLED
+		curr_field->ptr = (byte *) buf + curr_field->offset(curr_field->getTable()->getInsertRecord());
+#else
+		curr_field->ptr = (byte *) buf + curr_field->offset(curr_field->table->record[0]);
+#endif
 #endif
 		switch (curr_field->field_name[0]) {
 			case 'R':
@@ -1564,15 +1597,22 @@
 	table->write_set = NULL;
 	table->read_set = NULL;
 
+#ifdef DRIZZLED
 	memset(buf, 0xFF, table->getNullBytes());
+#else
+	memset(buf, 0xFF, table->s->null_bytes);
+#endif
  	for (Field **field=GET_TABLE_FIELDS(table) ; *field ; field++) {
  		curr_field = *field;
-
 		save = curr_field->ptr;
 #if MYSQL_VERSION_ID < 50114
 		curr_field->ptr = (byte *) buf + curr_field->offset();
 #else
+#ifdef DRIZZLED
 		curr_field->ptr = (byte *) buf + curr_field->offset(curr_field->getTable()->getInsertRecord());
+#else
+		curr_field->ptr = (byte *) buf + curr_field->offset(curr_field->table->record[0]);
+#endif
 #endif
 		switch (curr_field->field_name[0]) {
 			case 'B':
@@ -1730,6 +1770,13 @@
 	return  (MSMetaDataTable*) MSSystemTableShare::openSystemTable(path, NULL);
 }
 
+void MSMetaDataTable::use()
+{
+	MSRepositoryTable::use();
+	new_(iMetData, CSStringBuffer(80));
+	iMetDataSize = 0;
+}
+
 void MSMetaDataTable::unuse()
 {
 	MSRepositoryTable::unuse();
@@ -1748,7 +1795,6 @@
 	iMetDataPos = 0;
 	iMetBlobRepo = 0;
 	iMetBlobOffset = 0;
-	new_(iMetData, CSStringBuffer(80));
 	iMetStateSaved = false;
 }
 
@@ -1921,15 +1967,22 @@
 	save_write_set = table->write_set;
 	table->write_set = NULL;
 
+#ifdef DRIZZLED
 	memset(buf, 0xFF, table->getNullBytes());
+#else
+	memset(buf, 0xFF, table->s->null_bytes);
+#endif
  	for (Field **field=GET_TABLE_FIELDS(table) ; *field ; field++) {
  		curr_field = *field;
-
 		save = curr_field->ptr;
 #if MYSQL_VERSION_ID < 50114
 		curr_field->ptr = (byte *) buf + curr_field->offset();
 #else
+#ifdef DRIZZLED
 		curr_field->ptr = (byte *) buf + curr_field->offset(curr_field->getTable()->getInsertRecord());
+#else
+		curr_field->ptr = (byte *) buf + curr_field->offset(curr_field->table->record[0]);
+#endif
 #endif
 		switch (curr_field->field_name[0]) {
 			case 'R':
@@ -2162,6 +2215,39 @@
 	exit_();
 }
 
+class UpdateRowCleanUp : public CSRefObject {
+	bool do_cleanup;
+	MSMetaDataTable *tab;
+	char *row_data;
+	
+	uint32_t ref_id;
+
+	public:
+	
+	UpdateRowCleanUp(): CSRefObject(),
+		do_cleanup(false), tab(NULL), row_data(NULL){}
+		
+	~UpdateRowCleanUp() 
+	{
+		if (do_cleanup) {
+			tab->deleteRow(row_data);
+		}
+	}
+	
+	void setCleanUp(MSMetaDataTable *table, char *data)
+	{
+		row_data = data;
+		tab = table;
+		do_cleanup = true;
+	}
+	
+	void cancelCleanUp()
+	{
+		do_cleanup = false;
+	}
+	
+};
+
 void MSMetaDataTable::updateRow(char *old_data, char *new_data)
 {
 	uint32_t o_repo_index, n_repo_index;
@@ -2191,15 +2277,19 @@
 	
 	// If the meta data is not for the same BLOB then we do an insert and delete.
 	if ((n_repo_index != o_repo_index) || (n_repo_offset != o_repo_offset)) {
+		UpdateRowCleanUp *cleanup;
+		new_(cleanup, UpdateRowCleanUp());
+		push_(cleanup);
+
 		insertRow(new_data);
-		try_(a) {
-			deleteRow(old_data);
-		}
-		catch_(a) {
-			deleteRow(new_data);
-			throw_();
-		}
-		cont_(a);
+		
+		cleanup->setCleanUp(this, new_data);
+		
+		deleteRow(old_data);
+		
+		cleanup->cancelCleanUp();
+		release_(cleanup);
+		
 		exit_();
 	}
 	
@@ -2300,7 +2390,11 @@
 
 MSSystemTableShare::~MSSystemTableShare()
 {
+#ifdef DRIZZLED
 	myThrLock.unlock();
+#else
+	thr_lock_delete(&myThrLock);
+#endif
 	if (myTablePath) {
 		myTablePath->release();
 		myTablePath = NULL;

=== modified file 'plugin/pbms/src/system_table_ms.h'
--- plugin/pbms/src/system_table_ms.h	2010-10-02 21:15:42 +0000
+++ plugin/pbms/src/system_table_ms.h	2010-10-18 21:21:20 +0000
@@ -100,6 +100,8 @@
 	static void systemTablesStartUp();
 	static void systemTableShutDown();
 
+	private:
+	static bool try_loadSystemTables(CSThread *self, int i, MSDatabase *db);
 };
 
 class MSSystemTableShare;
@@ -133,11 +135,12 @@
 		Field *assumed_str_field = GET_FIELD(mySQLTable, column_index);
 		unsigned char *old_ptr = assumed_str_field->ptr;
 		
+
+#ifdef DRIZZLED
 		assumed_str_field->ptr = (unsigned char *)row + assumed_str_field->offset(mySQLTable->getInsertRecord());
-
-#ifdef DRIZZLED
 		assumed_str_field->setReadSet();
 #else
+		assumed_str_field->ptr = (unsigned char *)row + assumed_str_field->offset(mySQLTable->record[0]);
 		assumed_str_field->table->use_all_columns();
 #endif
 		value = assumed_str_field->val_str(value);
@@ -150,11 +153,12 @@
 		Field *assumed_int_field = GET_FIELD(mySQLTable, column_index);
 		unsigned char *old_ptr = assumed_int_field->ptr;
 
+
+#ifdef DRIZZLED
 		assumed_int_field->ptr = (unsigned char *)row + assumed_int_field->offset(mySQLTable->getInsertRecord());
-
-#ifdef DRIZZLED
 		assumed_int_field->setReadSet();
 #else
+		assumed_int_field->ptr = (unsigned char *)row + assumed_int_field->offset(mySQLTable->record[0]);
 		assumed_int_field->table->use_all_columns();
 #endif
 		*value = assumed_int_field->val_int();
@@ -310,6 +314,7 @@
 	MSMetaDataTable(MSSystemTableShare *share, TABLE *table);
 	virtual ~MSMetaDataTable();
 
+	void use();
 	void unuse();
 	void seqScanInit();
 	int	getRefLen();
@@ -347,7 +352,7 @@
 	virtual bool returnSubRecord(char *buf);
 	virtual bool returnRow(MSBlobHeadPtr blob, char *buf) { return MSRepositoryTable::returnRow(blob, buf);}
 	virtual void returnRow(char *name, char *value, char *buf);
-	virtual bool resetScan(bool positioned, uint32_t index = 0) { return MSRepositoryTable::resetScan(positioned, index);}
+	virtual bool resetScan(bool positioned, uint32_t index = 0) {bool have_data= false; return resetScan(positioned, &have_data, index);}
 	virtual bool resetScan(bool positioned, bool *have_data, uint32_t iRepoIndex = 0);
 	
 	static MSMetaDataTable *newMSMetaDataTable(MSDatabase *db);

=== modified file 'plugin/pbms/src/table_ms.cc'
--- plugin/pbms/src/table_ms.cc	2010-10-07 01:20:26 +0000
+++ plugin/pbms/src/table_ms.cc	2010-10-18 21:21:20 +0000
@@ -32,14 +32,16 @@
 #include <stdlib.h>
 #include <stddef.h>
 
-#include "open_table_ms.h"
-#include "database_ms.h"
+#include "defs_ms.h"
 
 #include "cslib/CSGlobal.h"
 #include "cslib/CSLog.h"
 #include "cslib/CSPath.h"
 #include "cslib/CSStrUtil.h"
 
+#include "database_ms.h"
+#include "open_table_ms.h"
+
 MSTable::MSTable():
 CSSharedRefObject(),
 myTableName(NULL),

=== modified file 'plugin/pbms/src/temp_log_ms.cc'
--- plugin/pbms/src/temp_log_ms.cc	2010-10-07 01:20:26 +0000
+++ plugin/pbms/src/temp_log_ms.cc	2010-10-18 21:21:20 +0000
@@ -31,16 +31,18 @@
 
 #include <stddef.h>
 
+#include "defs_ms.h"
+
+#include "cslib/CSGlobal.h"
+#include "cslib/CSStrUtil.h"
+#include "cslib/CSStorage.h"
+
 #include "temp_log_ms.h"
 #include "open_table_ms.h"
 #include "trans_log_ms.h"
 #include "transaction_ms.h"
 #include "parameters_ms.h"
 
-#include "cslib/CSGlobal.h"
-#include "cslib/CSStrUtil.h"
-#include "cslib/CSStorage.h"
-
 
 // Search the transaction log for a MS_ReferenceTxn record for the given BLOB.
 // Just search the log file and not the cache. Seaching the cache may be faster but
@@ -111,18 +113,22 @@
 MSTempLogFile *MSTempLogFile::newTempLogFile(uint32_t id, MSTempLog *temp_log, CSFile *file)
 {
 	MSTempLogFile *f;
-	
-	if (!(f = new MSTempLogFile())) {
-		temp_log->release();
-		file->release();
+	enter_();
+	
+	push_(temp_log);
+	push_(file);
+	
+	if (!(f = new MSTempLogFile())) 
 		CSException::throwOSError(CS_CONTEXT, ENOMEM);
-	}
+
 	f->myTempLogID = id;
+	
+	pop_(file);
+	f->setFile(file);
+	
+	pop_(temp_log);
 	f->myTempLog = temp_log;
-	f->myFile = file;
-	f->myFilePath = file->myFilePath;
-	f->myFilePath->retain();
-	return f;
+	return_(f);
 }
 
 MSTempLog::MSTempLog(uint32_t id, MSDatabase *db, off64_t file_size):
@@ -261,6 +267,70 @@
 	}
 }
 
+bool MSTempLogThread::try_ReleaseBLOBReference(CSThread *self, CSStringBuffer *buffer, uint32_t tab_id, int type, uint64_t blob_id, uint32_t auth_code)
+{
+	volatile bool rtc = true;
+	try_(a) {
+		/* Release the BLOB reference. */
+		MSOpenTable *otab;
+
+		if (type == MS_TL_REPO_REF) {
+			MSRepoFile	*repo_file;
+
+			if ((repo_file = iTempLogDatabase->getRepoFileFromPool(tab_id, true))) {
+				frompool_(repo_file);
+				repo_file->checkBlob(buffer, blob_id, auth_code, iTempLogFile->myTempLogID, iLogOffset);
+				backtopool_(repo_file);
+			}
+		}
+		else {
+			if ((otab = MSTableList::getOpenTableByID(iTempLogDatabase->myDatabaseID, tab_id))) {
+				frompool_(otab);
+				if (type == MS_TL_BLOB_REF) {
+					otab->checkBlob(buffer, blob_id, auth_code, iTempLogFile->myTempLogID, iLogOffset);
+					backtopool_(otab);
+				}
+				else {
+					ASSERT(type == MS_TL_TABLE_REF);
+					if ((type == MS_TL_TABLE_REF) && otab->deleteReferences(iTempLogFile->myTempLogID, iLogOffset, &myMustQuit)) {
+						/* Delete the file now... */
+						MSTable			*tab;
+						CSPath			*from_path;
+						MSOpenTablePool *tab_pool;
+
+						tab = otab->getDBTable();
+						from_path = otab->getDBTable()->getTableFile();
+
+						pop_(otab);
+
+						push_(from_path);
+						tab->retain();
+						push_(tab);
+
+						tab_pool = MSTableList::lockTablePoolForDeletion(otab); // This returns otab to the pool.
+						frompool_(tab_pool);
+
+						from_path->removeFile();
+						tab->myDatabase->removeTable(tab);
+
+						backtopool_(tab_pool); // The will unlock and close the table pool freeing all tables in it.
+						pop_(tab);				// Returning the pool will have released this. (YUK!)
+						release_(from_path);
+					}
+					else 
+						backtopool_(otab);
+				}
+			}
+		}
+		
+		rtc = false;
+	}
+	
+	catch_(a);
+	cont_(a);
+	return rtc;
+}
+
 bool MSTempLogThread::doWork()
 {
 	size_t				tfer;
@@ -308,7 +378,6 @@
 			uint32_t then;
 			time_t	now;
 
-			CLOBBER_PROTECT(blob_id);
 			/*
 			 * Items in the temp log are never updated.
 			 * If a temp operation is canceled then the object 
@@ -330,61 +399,8 @@
 				myWaitTime = MSTempLog::adjustWaitTime(then, now);
 				break;
 			}
-
-			try_(a) {
-				/* Release the BLOB reference. */
-				MSOpenTable *otab;
-
-				if (type == MS_TL_REPO_REF) {
-					MSRepoFile	*repo_file;
-
-					if ((repo_file = iTempLogDatabase->getRepoFileFromPool(tab_id, true))) {
-						frompool_(repo_file);
-						repo_file->checkBlob(buffer, blob_id, auth_code, iTempLogFile->myTempLogID, iLogOffset);
-						backtopool_(repo_file);
-					}
-				}
-				else {
-					if ((otab = MSTableList::getOpenTableByID(iTempLogDatabase->myDatabaseID, tab_id))) {
-						frompool_(otab);
-						if (type == MS_TL_BLOB_REF) {
-							otab->checkBlob(buffer, blob_id, auth_code, iTempLogFile->myTempLogID, iLogOffset);
-							backtopool_(otab);
-						}
-						else {
-							ASSERT(type == MS_TL_TABLE_REF);
-							if ((type == MS_TL_TABLE_REF) && otab->deleteReferences(iTempLogFile->myTempLogID, iLogOffset, &myMustQuit)) {
-								/* Delete the file now... */
-								MSTable			*tab;
-								CSPath			*from_path;
-								MSOpenTablePool *tab_pool;
-
-								tab = otab->getDBTable();
-								from_path = otab->getDBTable()->getTableFile();
-
-								pop_(otab);
-
-								push_(from_path);
-								tab->retain();
-								push_(tab);
-
-								tab_pool = MSTableList::lockTablePoolForDeletion(otab); // This returns otab to the pool.
-								frompool_(tab_pool);
-
-								from_path->removeFile();
-								tab->myDatabase->removeTable(tab);
-
-								backtopool_(tab_pool); // The will unlock and close the table pool freeing all tables in it.
-								pop_(tab);				// Returning the pool will have released this. (YUK!)
-								release_(from_path);
-							}
-							else 
-								backtopool_(otab);
-						}
-					}
-				}
-			}
-			catch_(a) {
+		
+			if (try_ReleaseBLOBReference(self, buffer, tab_id, type, blob_id, auth_code)) {
 				int err = self->myException.getErrorCode();
 				
 				if (err == MS_ERR_TABLE_LOCKED) {
@@ -401,7 +417,7 @@
 				else
 					self->myException.log(NULL);
 			}
-			cont_(a);
+
 		}
 		else {
 			// Only part of the data read, don't wait very long to try again:

=== modified file 'plugin/pbms/src/temp_log_ms.h'
--- plugin/pbms/src/temp_log_ms.h	2010-10-07 01:20:26 +0000
+++ plugin/pbms/src/temp_log_ms.h	2010-10-18 21:21:20 +0000
@@ -37,13 +37,11 @@
 #ifndef __TEMPLOG_MS_H__
 #define __TEMPLOG_MS_H__
 
-#include "defs_ms.h"
-
 #include "cslib/CSDefs.h"
 #include "cslib/CSFile.h"
 #include "cslib/CSStream.h"
-#include "cslib/CSThread.h"
 
+#include "defs_ms.h"
 
 class MSOpenTable;
 class MSDatabase;
@@ -142,6 +140,8 @@
 	virtual void *completeWork();
 
 private:
+	bool try_ReleaseBLOBReference(CSThread *self, CSStringBuffer *buffer, uint32_t tab_id, int type, uint64_t blob_id, uint32_t auth_code);
+
 	MSDatabase			*iTempLogDatabase;
 	MSTempLogFile		*iTempLogFile;
 	size_t				iLogRecSize;

=== modified file 'plugin/pbms/src/trans_cache_ms.cc'
--- plugin/pbms/src/trans_cache_ms.cc	2010-10-06 18:22:02 +0000
+++ plugin/pbms/src/trans_cache_ms.cc	2010-10-18 21:21:20 +0000
@@ -47,6 +47,7 @@
 	uint64_t	tc_position;	// The log position of the record.
 } myTransRec, *myTransPtr;
 
+#define BAD_LOG_POSITION ((uint64_t) -1)
 typedef struct TransList {
 #ifdef DEBUG
 	uint32_t		old_tid;
@@ -285,7 +286,7 @@
 		// The cache is full.
 		tc_OverFlowTID = 0;
 		tc_OverFlow->tid = tid; // save the tid of the first transaction to overflow.
-		tc_OverFlow->log_position = UINT64_MAX;
+		tc_OverFlow->log_position = BAD_LOG_POSITION;
 		tc_OverFlow->len = 0; 
 		tc_OverFlow->terminated = MS_Running; 
 		ref = OVERFLOW_TREF;
@@ -320,7 +321,7 @@
 		
 	tc_List[ref].tid = tid;
 	tc_List[ref].len = 0;
-	tc_List[ref].log_position = UINT64_MAX;
+	tc_List[ref].log_position = BAD_LOG_POSITION;
 	tc_List[ref].terminated = MS_Running;
 
 	// Update these after initializing the structure because

=== modified file 'plugin/pbms/src/transaction_ms.cc'
--- plugin/pbms/src/transaction_ms.cc	2010-10-08 14:22:02 +0000
+++ plugin/pbms/src/transaction_ms.cc	2010-10-18 21:21:20 +0000
@@ -28,21 +28,21 @@
  */
 
 #include "cslib/CSConfig.h"
+
 #include <inttypes.h>
-#include <string>
 
 #include "defs_ms.h"
+
+#include "cslib/CSGlobal.h"
+#include "cslib/CSStrUtil.h"
+#include "cslib/CSLog.h"
+
 #include "mysql_ms.h"
 #include "open_table_ms.h"
 #include "trans_log_ms.h"
 #include "transaction_ms.h"
 #include "pbmsdaemon_ms.h"
 
-#include "cslib/CSGlobal.h"
-#include "cslib/CSStrUtil.h"
-#include "cslib/CSLog.h"
-
-
 /*
  * The pbms_ functions are utility functions supplied by ha_pbms.cc
  */
@@ -179,10 +179,10 @@
 
 void MSTransactionThread::dereference(MSTransPtr rec, MS_TxnState state)
 {
-	MSOpenTable		*otab;
 	enter_();
 	
 	try_(a) {
+		MSOpenTable		*otab;
 		otab = MSTableList::getOpenTableByID(rec->tr_db_id, rec->tr_tab_id);
 		frompool_(otab);
 		otab->freeReference(rec->tr_blob_id, rec->tr_blob_ref_id);
@@ -199,10 +199,10 @@
 
 void MSTransactionThread::commitReference(MSTransPtr rec, MS_TxnState state)
 {
-	MSOpenTable		*otab;
 	enter_();
 	
 	try_(a) {
+		MSOpenTable		*otab;
 		otab = MSTableList::getOpenTableByID(rec->tr_db_id, rec->tr_tab_id);
 		frompool_(otab);
 		otab->commitReference(rec->tr_blob_id, rec->tr_blob_ref_id);
@@ -237,11 +237,11 @@
 
 bool MSTransactionThread::doWork()
 {
-	MSTransRec rec = {0,0,0,0,0,0,0};
-	MS_TxnState state;
 	enter_();
 	
 	try_(a) {
+		MSTransRec rec = {0,0,0,0,0,0,0};
+		MS_TxnState state;
 		while (!myMustQuit) {
 			// This will sleep while waiting for the next 
 			// completed transaction.
@@ -412,14 +412,11 @@
 	exit_();
 }
 
-class MSTransactionCheckPoint: public CSCString
+class MSTransactionCheckPoint: public CSString
 {
 	public:
-	MSTransactionCheckPoint(const char *name, uint32_t stmtCount ):CSCString()
+	MSTransactionCheckPoint(const char *name, uint32_t stmtCount ):CSString(name)
 	{
-		myCString = cs_strdup(name);
-		myStrLen = strlen(name);
-
 		position = stmtCount;
 	}
 	
@@ -444,10 +441,10 @@
 void MSTransactionManager::releaseSavepoint(const char *savePoint)
 {
 	MSTransactionCheckPoint *checkPoint;
-	CSCString *name;
+	CSString *name;
 	enter_();
 	
-	name = CSCString::newString(savePoint);
+	name = CSString::newString(savePoint);
 	push_(name);
 
 	checkPoint = (MSTransactionCheckPoint*) self->mySavePoints.find(name);
@@ -462,10 +459,10 @@
 void MSTransactionManager::rollbackTo(const char *savePoint)
 {
 	MSTransactionCheckPoint *checkPoint;
-	CSCString *name;
+	CSString *name;
 	enter_();
 	
-	name = CSCString::newString(savePoint);
+	name = CSString::newString(savePoint);
 	push_(name);
 
 	checkPoint = (MSTransactionCheckPoint*) self->mySavePoints.find(name);

=== modified file 'plugin/pbxt/plugin.ini'
--- plugin/pbxt/plugin.ini	2010-10-06 17:22:20 +0000
+++ plugin/pbxt/plugin.ini	2010-10-18 21:21:20 +0000
@@ -11,7 +11,7 @@
 # 
 # You should have received a copy of the GNU General Public License along with
 # this program; if not, write to the Free Software Foundation, Inc., 59 Temple
-# Place, Suite 330, Boston, MA 02111-1307 USA
+# Place, Suite 330, Boston, MA 02110-1301 USA
 #
 
 [plugin]

